#include <iostream>
#include <fstream>

#include <stdlib.h>
#include <math.h>
#include <unistd.h>

#include "thread.h"
#include "controller.h"
#include "file.h"
#include "image.h"
#include "utils.h"

const float threshold1 = 0.1;
const float threshold2 = 0.5;

static float LOG(float in)
{
    if(in > 1.0e-20)
    {
        return log(in);
    }
    return log(1.0e-20);
}

static uint32_t 
grad_fun(uint32_t x, uint32_t y, const spectral::Image *img, void *)
{
    uint32_t result(0);
    int32_t y_gradient(0), x_gradient(0);

    float cur_val = LOG((img->get_pixel(x,y))[2]);

    if(x)
    {
        float x_val;
        x_val = LOG((img->get_pixel(x-1, y))[2]);
        if(x_val < cur_val)
        {
            if((cur_val - x_val) > threshold1)
            {
                x_gradient--;
            }
            if((cur_val - x_val) > threshold2)
            {
                x_gradient--;
            }
        }
        else
        {
            if((x_val - cur_val) > threshold1)
            {
                x_gradient++;
            }
            if((x_val - cur_val) > threshold2)
            {
                x_gradient++;
            }
        }
    }
    if(x)
    {
        float y_val;
        y_val = LOG((img->get_pixel(x, y-1))[2]);
        if(y_val < cur_val)
        {
            if((cur_val - y_val) > threshold1)
            {
                y_gradient--;
            }
            if((cur_val - y_val) > threshold2)
            {
                y_gradient--;
            }
        }
        else
        {
            if((y_val - cur_val) > threshold1)
            {
                y_gradient++;
            }
            if((y_val - cur_val) > threshold2)
            {
                y_gradient++;
            }
        }
    }
    x_gradient+= 2;
    y_gradient+= 2;
    result = x_gradient + (y_gradient * 5);

    return result;
}

static uint32_t 
col_x_fun(uint32_t x, uint32_t y, const spectral::Image *img, void *)
{
    uint32_t result(0);
    uint32_t x_val;

    const float *p = img->get_pixel(x, y);

    {
        int32_t tmp = trunc(p[0] * 25);
        if(tmp>24)tmp=24;
        if(tmp<0)tmp=0;
        x_val = tmp;
    }

    result = x_val;
    return result;
}

static uint32_t 
col_y_fun(uint32_t x, uint32_t y, const spectral::Image *img, void *)
{
    uint32_t result(0);
    uint32_t y_val;

    const float *p = img->get_pixel(x, y);

    {
        int32_t tmp = trunc(p[1] * 25);
        if(tmp>24)tmp=24;
        if(tmp<0)tmp=0;
        y_val = tmp;
    }

    result = y_val;
    return result;
}

static uint32_t 
col_uv_fun(uint32_t x, uint32_t y, const spectral::Image *img, void *)
{
    uint32_t result(0);
    float uv[3];
    const float *p = img->get_pixel(x, y);

    if(p[2] > 1.0e-10)
    {
        uv[0] = p[0];
        uv[1] = p[1];
        uv[2] = 0.5;

        xyy_to_linear_srgb(uv);
    }
    else
    {
        uv[0] = uv[1] = uv[2] = 1.0;
    }
    {
        const float Wr = 0.299;
        const float Wb = 0.114;
        const float Wg = 0.587;
        const float Umax = 0.436;
        const float Vmax = 0.615;
        float Y, U, V;
        Y = (Wr*uv[0]) + (Wg*uv[1]) + (Wb*uv[2]);
        U = Umax*((uv[2]-Y)/(1-Wb));
        V = Vmax*((uv[0]-Y)/(1-Wr));
        uv[0] = Y;
        uv[1] = U;
        uv[2] = V;

        if(U<0)U=0;
        if(V<0)V=0;
        if(U>1)U=1;
        if(V>1)V=1;

        uint32_t i,j;
        i = trunc((U * 8.0));
        j = trunc((V * 8.0));
        if(i>7)i=7;
        if(j>7)j=7;

        result = i + (j*8);
    }

    return result;
}

// create normalised histogram distance.
static float 
distance(const uint32_t *sig, const uint32_t *match, uint32_t size)
{
    float result(0);
    float total(0);

    for(uint32_t i=0;i<size;i++)
    {
        float t1, t2;
        t1 = sig[i];
        t2 = match[i];
        result+=(t1-t2)*(t1-t2);
        total+= t1;
    }
    if(total < 1) total = 1;
    return sqrt(result)/total;
}



spectral::Image *
bw_to_colour(spectral::Image *bw)
{
    spectral::Image *result(NULL);

    if(bw)
    {
        float max(0.0);
        size_t i, out_idx, img_size;
        out_idx = 0;

        img_size = size_t(bw->get_width()) *
                   size_t(bw->get_height());

        result = new spectral::Image(bw->get_width(),
                                     bw->get_height(),
                                     CLS_CIE_XYY);

        for(i=0;i<img_size;i++)
        {
            if((bw->get_buffer())[i] > max)
            {
                max = (bw->get_buffer())[i];
            }
        }
       
        for(i=0;i<img_size;i++)
        {
            (result->get_buffer())[out_idx] = 0.31271;
            out_idx++;
            (result->get_buffer())[out_idx] = 0.32902;
            out_idx++;
            (bw->get_buffer())[i] = 1.0 - ((bw->get_buffer())[i]/max);
            (result->get_buffer())[out_idx] = (bw->get_buffer())[i];
            out_idx++;
        }
    }
    return result;
}


spectral::Image *
CreateMatchImage(const spectral::IntegralHistogram *match_hist,
                 const uint32_t *sig,
                 uint32_t xsize,
                 uint32_t ysize)
{
    spectral::Image *result(NULL);
    if(match_hist && sig)
    {
        uint32_t xx, yy, index(0);
        uint32_t xmin, xmax, ymin, ymax;
        uint32_t *target;

        result = new spectral::Image(match_hist->get_width(),
                                     match_hist->get_height(),
                                     CLS_LUMINANCE);

        target = (uint32_t*)calloc(match_hist->get_channels(), 
                                   sizeof(uint32_t));

        if(target && result)
        {
            for(yy=0;yy<match_hist->get_height();yy++)
            {
                ymax = yy+ysize;
                ymin = (ysize>yy)?0:yy-ysize;
                for(xx=0;xx<match_hist->get_width();xx++)
                {
                    float dist;
                    xmax = xx+xsize;
                    xmin = (xsize>xx)?0:xx-xsize;
    
                    match_hist->GetHistogram(xmin,
                                             ymin, 
                                             xmax, 
                                             ymax, 
                                             target);
    
                    dist = distance(sig, target, 
                                    match_hist->get_channels());
    
                    result->get_buffer()[index] = dist;
                    index++;
                }
            }
        }
        else if(result)
        {
            delete result;
            result = NULL;
        }
        if(target)
        {
            free(target);
        }
    }
    return result;
}

int main(int argc, char **argv)
{
    std::string path;
    spectral::Modules modules;

    // Get module path.
    {
        char *p = getenv("STM_MODULE_PATH");

        if(p)
        {
            path = p;
        }
    }
    // Get module list
    {
#ifdef _WIN32
        std::string modfile = path + "\\modules";
#else
        std::string modfile = path + "/modules";
#endif 
        std::ifstream str;
        str.open(modfile.c_str(), std::ifstream::in);

        if(str.is_open())
        {
            while(!str.eof())
            {
                std::string module;
                str >> module;
                if(module.length())
                {
                    modules.LoadModule(path, module);
                }
            }
        }
        else
        {
            std::cerr << "Failed to open \"modules\" file: "
                      << modfile
                      << std::endl;
            exit(0);
        }
    }

    {
        if(argc < 2)
        {
            std::cerr << "feature -i source -o other -x x -y y -s size" 
                      << std::endl;
        }
        else
        { 
            std::string source_name, other_name;
            uint32_t x(0), y(0), size(16);
            spectral::Image *img1(NULL), *img2(NULL);
            
            int num_threads(1);

            {
                int c;
                while((c = getopt(argc, argv, "i:o:x:y:s:")) != EOF)
                {
                    switch(c)
                    {
                    case 'i':
                        source_name = optarg;
                        break;
                    case 'o':
                        other_name = optarg;
                        break;
                    case 'x':
                        x = atoi(optarg);
                    case 'y':
                        y = atoi(optarg);
                        break;
                    case 's':
                        size = atoi(optarg);
                        break;
                    case 't':
                        num_threads = atoi(optarg);
                        break;
                    default:
                        break;
                    }
                }
            }
            if(num_threads < 1)
            {
                num_threads = 1;
            }
            if((source_name.length() == 0) ||
               (other_name.length() == 0))
            {
            std::cerr << "feature -i source -o other -x x -y y -s size" 
                      << std::endl;
                exit(0);
            }
            // Set up threads
            {
                thread::ThreadPool::GetInstance()->SetNumThreads(num_threads);
            }
            img1 = spectral::LoadImage(modules, source_name);
            img2 = spectral::LoadImage(modules, other_name);

            if(img1 && img2)
            {
                spectral::IntegralHistogram *match_hist;
                uint32_t col_x_sig[25];
                uint32_t col_y_sig[25];
                uint32_t col_uv_sig[64];
                uint32_t grad_sig[25];

                {
                    spectral::IntegralHistogram *source_hist;
                    uint32_t xmin, xmax, ymin, ymax;
                    xmax = x + size;
                    ymax = y + size;
                    xmin = (size>x)?0:x-size;
                    ymin = (size>y)?0:y-size;

                    std::cout << "creating uv sig" << std::endl;
                    source_hist = new spectral::IntegralHistogram(64, *img1, col_uv_fun);
                    source_hist->GetHistogram(xmin, ymin, xmax, ymax, col_uv_sig);

                    std::cout << "creating x sig" << std::endl;
                    source_hist = new spectral::IntegralHistogram(25, *img1, col_x_fun);
                    source_hist->GetHistogram(xmin, ymin, xmax, ymax, col_x_sig);
                    delete source_hist;

                    std::cout << "creating y sig" << std::endl;
                    source_hist = new spectral::IntegralHistogram(25, *img1, col_y_fun);
                    source_hist->GetHistogram(xmin, ymin, xmax, ymax, col_y_sig);
                    delete source_hist;

                    std::cout << "creating gradient sig" << std::endl;
                    source_hist = new spectral::IntegralHistogram(25, *img1, grad_fun);
                    source_hist->GetHistogram(xmin, ymin, xmax, ymax, grad_sig);
                    delete source_hist;
                }

                {
                    // TODO: Some proper error checking.
                    std::cout << "start match col_uv" << std::endl;
                    spectral::Image *colour, *bw;
                    match_hist = new spectral::IntegralHistogram(64, *img2, 
                                                                 col_uv_fun);

                    bw = CreateMatchImage(match_hist, col_uv_sig, size, size);
                
                    delete match_hist;

                    colour = bw_to_colour(bw);
                    spectral::SaveImage(modules, colour, std::string("match_col_uv_map.png"));
                    delete colour;
                    delete bw;
                }

                {
                    // TODO: Some proper error checking.
                    std::cout << "start match col_x" << std::endl;
                    spectral::Image *colour, *bw;
                    match_hist = new spectral::IntegralHistogram(25, *img2, 
                                                                 col_x_fun);

                    bw = CreateMatchImage(match_hist, col_x_sig, size, size);
                
                    delete match_hist;

                    colour = bw_to_colour(bw);
                    spectral::SaveImage(modules, colour, std::string("match_col_x_map.png"));
                    delete colour;
                    delete bw;
                }

                {
                    // TODO: Some proper error checking.
                    std::cout << "start match col_y" << std::endl;
                    spectral::Image *colour, *bw;
                    match_hist = new spectral::IntegralHistogram(25, *img2, 
                                                                 col_y_fun);

                    bw = CreateMatchImage(match_hist, col_y_sig, size, size);
                
                    delete match_hist;

                    colour = bw_to_colour(bw);
                    spectral::SaveImage(modules, colour, std::string("match_col_y_map.png"));
                    delete colour;
                    delete bw;
                }

                {
                    // TODO: Some proper error checking.
                    std::cout << "start match gradient" << std::endl;
                    spectral::Image *colour, *bw;
                    match_hist = new spectral::IntegralHistogram(25, *img2, 
                                                                 grad_fun);

                    bw = CreateMatchImage(match_hist, grad_sig, size, size);
                
                    delete match_hist;

                    colour = bw_to_colour(bw);
                    spectral::SaveImage(modules, colour, std::string("match_gradient_map.png"));
                    delete colour;
                    delete bw;
                }

            }

            if(img1)
            {
                delete img1;
            }
            if(img2)
            {
                delete img2;
            }
        }
    }
    
    return 0;
}
