// A simple, non thread safe observer mechanism.  For single threaded guis only!
// Use at your own risk.

#include "SimpleObserver.h"

////////////////////////////////////////////////////////////////////////////////
// Observer class
SimpleObserver::~SimpleObserver()
{
    std::set<SimpleObservable*>::iterator i = m_observing.begin();

    while(i != m_observing.end())
    {
        (*i)->UnRegister(this, true);
        i++;
    }
}

void
SimpleObserver::Register(SimpleObservable *thing)
{
    if(thing)
    {
        // connect to an observable, and remember that we have done so.

        std::set<SimpleObservable*>::iterator i =
            m_observing.find(thing);

        if(i == m_observing.end())
        {
            thing->Register(this);
            m_observing.insert(thing);
        }
    }
}

void
SimpleObserver::UnRegister(SimpleObservable *thing)
{
    if(thing)
    {
        std::set<SimpleObservable*>::iterator i =
            m_observing.find(thing);

        if(i != m_observing.end())
        {
            thing->UnRegister(this);

            m_observing.erase(i);
        }
    }
}

void
SimpleObserver::ReceiveMessage(const SimpleObserverMessage *msg)
{
    msg->Deliver(this);
}

////////////////////////////////////////////////////////////////////////////////
// Observable class
void 
SimpleObservable::Register(SimpleObserver *observer)
{
    if(observer)
    {
        std::map<SimpleObserver*, unsigned int>::iterator i = 
            m_observers.find(observer);
    
        if(i == m_observers.end())
        {
            // A brand new observer.
            m_observers[observer] = 1;
        }
        else
        {
            // Increase reference count.
            m_observers[observer] = (i->second) + 1;
        }
    }
}

void 
SimpleObservable::UnRegister(SimpleObserver *observer, bool all_instances)
{
    if(observer)
    {
        std::map<SimpleObserver*, unsigned int>::iterator i = 
            m_observers.find(observer);
    
        if(i == m_observers.end())
        {
            // An unknown observer.
            //assert(0);
        }
        else
        {
            // Increase reference count.
            if(i->second == 1)
            {
                // ref count will drop to zero, so remove.
                m_observers.erase(i);
            }
            else
            {
                if(all_instances)
                {
                    // We have probably destroyed the observer.
                    m_observers.erase(i);
                }
                else
                {
                    // reduce ref count.
                    m_observers[observer] = (i->second) - 1;
                }
            }
        }
    }
}

unsigned int
SimpleObservable::SendMessage(const SimpleObserverMessage *msg) const
{
    unsigned int result(0);

    if(msg)
    {
        std::map<SimpleObserver*, unsigned int>::const_iterator i = m_observers.begin();

        while(i != m_observers.end())
        {
            i->first->ReceiveMessage(msg);
            i++;
        }
        delete msg;

        result = m_observers.size(); // Should really use some sort of success/failure 
    }
    return result;

}

