#include <iostream>
#include <stdio.h>
#include <stdlib.h>

#include "utils.h"

#include <math.h>

#include <vector>

#include <QApplication>
#include <QWidget>

// Layout the doalog
#include <QGridLayout>


// supported widget types
#include <QLabel>
#include <QDoubleSpinBox>
#include <QSpinBox>
#include <QCheckBox>
#include <QComboBox>

#include <QPainter>
#include <QMouseEvent>

#include "StmChromaticityWidget.h"

const int image_size(512);

StmXYPane::GammaObserver::GammaObserver(StmXYPane *pane)
    : QtGuiConfigObserver()
    , m_pane(pane)
{
    QtGuiConfig *config = QtGuiConfig::GetInstance();

    if(config)
    {
        Register((SimpleObservable *)config);
    }
}

StmXYPane::StmXYPane(QWidget *parent)
    : QWidget(parent)
    , m_x(0)
    , m_y(0)
    , m_gamma(1.0)
    , m_image(image_size, image_size, QImage::Format_RGB32)
    , m_observer(this)
{
    setMinimumSize(image_size, image_size);
    
    QtGuiConfig *config = QtGuiConfig::GetInstance();
    if(config)
    { 
        m_gamma = config->GetScreenGamma();
    }
    setGamma(m_gamma);
}

void 
StmXYPane::setGamma(float gamma)
{
    float inv_gamma;
    m_gamma = gamma;

    if(m_gamma > 0.25)
    {
        inv_gamma = 1.0 / m_gamma;
    }
    else 
    {
        inv_gamma = 4;
    }

    for(int x=0;x<image_size;x++)
    {
        float fx;
        fx = x;
        fx/= image_size;
        for(int y=0;y<image_size;y++)
        {
            float fy;
            float mapx, mapy;
            uint pixel;
            fy = y;
            fy/= image_size;

            pixel = 0;

            mapx = fx;
            mapy = fy;

            MapFrom(mapx, mapy);

            if(Validate(mapx, mapy))
            {
                pixel = IntColour(mapx, mapy, inv_gamma);
            }
            m_image.setPixel(x, y, pixel);
        }
    }
    repaint();
}

unsigned int 
StmXYPane::IntColour(float x, float y, float gamma)
{
    unsigned int pixel(0);

    // We have a valid x,y value.
    float col[3];
    col[0]=x;
    col[1]=y;
    col[2]=1;

    xyy_to_linear_srgb(col);

    // Scale.
    {
        float scale=0;
        int i;

        for(i=0;i<3;i++)
        {
            scale += (col[i] * col[i]);
            if(col[i] > scale)
            {
          //      scale = col[i];
            }
        }
        scale = sqrt(scale);
        //scale*= 1.6; // Reduce intensities slightly.

        for(i=0;i<3;i++)
        {
            col[i]/= scale;
            if(col[i]>1.0) col[i] = 1.0;
            if(col[i]<0.0) col[i] = 0.0;
            col[i] = pow(col[i], gamma);
        }
    }
    for(int i=0;i<3;i++)
    {
        unsigned int channel;
        float tmp;
        if(col[i] > 1.0)
        {
            col[i] = 1.0;
        }
        if(col[i] < 0.0)
        {
            col[i] = 0.0;
        }
        tmp = col[i] * 256;
        channel = floor(tmp);
        if(channel > 0xff)
        {
            channel = 0xff;
        }
        pixel<<= 8;
        pixel |= channel;
    }
    return pixel;
}

bool
StmXYPane::Validate(float x, float y)
{
    bool result(true);
    // First ensure we are in triangle (0,0) (1,0) (0,1)

    if((y < 0.0) || (x < 0.0) || (y > (1.0 - x)))
    {
        result = false;
    }
    
    if(result)
    {
        // TODO: More expensive check?
        result = (m_validator.Valid(x, y));
    }

    return result;
}

// Map colour xy to image xy
void 
StmXYPane::MapTo(float &x, float &y)
{
    float tx;
    tx = x + (y * 0.5);
    x = tx;
}

// Map from image xy to colour xy
void 
StmXYPane::MapFrom(float &x, float &y)
{
    float tx;
    tx = x - (y * 0.5);
    x = tx;
}

void
StmXYPane::paintEvent(QPaintEvent *)
{
    QPainter painter(this);
    painter.drawImage(QPoint(0,0), m_image);

    {
        // draw crosshairs
        float x, y;
        int xx, yy;
        x = m_x;
        y = m_y;
        MapTo(x, y);

        x*= image_size;
        y*= image_size;

        xx = floor(x);
        yy = floor(y);

        if(xx < 0)
           xx = 0;
        if(yy < 0)
           yy = 0;
        if(xx >= image_size)
          xx = image_size - 1;
        if(yy >= image_size)
          yy = image_size - 1;
        painter.setPen(QColor(0,0,0));
        painter.drawLine(xx, 0, xx, image_size - 1);
        painter.drawLine(0, yy, image_size - 1, yy);
    }
    {
        unsigned int intColour;
        float inv_gamma;
        int y;

        y = image_size-(image_size/8);

        if(m_gamma > 0.25)
        {
            inv_gamma = 1.0 / m_gamma;
        }
        else 
        {
            inv_gamma = 4;
        }
        intColour = IntColour(m_x, m_y, inv_gamma);
        painter.fillRect(1,y,image_size-2,image_size-2, QColor(intColour));
    }
}

void
StmXYPane::mousePressEvent(QMouseEvent *event)
{
    if(event->buttons() == Qt::LeftButton)
    {
        float x = event->x();
        float y = event->y();
        x/= image_size;
        y/= image_size;
        MapFrom(x, y);

        if(Validate(x, y))
        {
            m_x = x;
            m_y = y;
            repaint();
            chromaticityChanged(m_x, m_y);
        }
    }
}


void
StmXYPane::setXY(float x, float y)
{
    m_x = x;
    m_y = y;
    repaint();
}

////////////////////////////////////////////////////////////////////////////////
StmChromaticityDialog::StmChromaticityDialog(QWidget *parent)
    : QDialog(parent)
    , m_x(0)
    , m_y(0)
    , m_buttonHolder(this)
    , m_buttonLayout(&m_buttonHolder)
    , m_widgetHolder(this)
    , m_widgetLayout(&m_widgetHolder)
    , m_mainLayout(this)
    , m_ok(&m_buttonHolder)
    , m_cancel(&m_buttonHolder)
    , m_pane(&m_widgetHolder)
    , m_textWidgets(&m_widgetHolder)
    , m_textWidgetLayout(&m_textWidgets)
    , m_selectCombo(&m_textWidgets)
    , m_label(&m_textWidgets)
{
    m_mainLayout.addWidget(&m_widgetHolder);
    m_mainLayout.addWidget(&m_buttonHolder);
    m_widgetLayout.addWidget(&m_pane);
    m_widgetLayout.addWidget(&m_textWidgets);

    m_textWidgetLayout.addWidget(&m_selectCombo);
    m_textWidgetLayout.addWidget(&m_label);

    m_buttonLayout.addWidget(&m_ok);
    m_buttonLayout.addWidget(&m_cancel);
    m_ok.setText("Ok");
    m_cancel.setText("Cancel");

    m_pane.setXY(m_x, m_y);
    UpdateText();

    m_whitePoints.push_back(WhitePoint("Custom", 0, 0));
    m_whitePoints.push_back(WhitePoint("Noon Daylight/sRGB (D65)", 0.31271, 0.32902));
    m_whitePoints.push_back(WhitePoint("Overcast (D55)", 0.33242, 0.34743));
    m_whitePoints.push_back(WhitePoint("Daylight Flourescent (F1)", 0.31310, 0.33727));
    m_whitePoints.push_back(WhitePoint("Daylight Flourescent (F1)", 0.31310, 0.33727));
    m_whitePoints.push_back(WhitePoint("White Flourescent (F3)", 0.40910, 0.39430));
    m_whitePoints.push_back(WhitePoint("Tungsten (A)", 0.44747, 0.40745));
    m_whitePoints.push_back(WhitePoint("Equal Energy (E)", (1.0/3.0), (1.0/3.0)));

    for(size_t i=0;i<m_whitePoints.size();i++)
    {
        m_selectCombo.addItem(m_whitePoints[i].m_name.c_str());
    }

    connect(&m_pane, SIGNAL(chromaticityChanged(float, float)), 
            SLOT(onChromaticityChanged(float, float)));

    connect(&m_selectCombo, SIGNAL(currentIndexChanged(int)), 
            SLOT(onComboChanged(int)));

    connect(&m_ok, SIGNAL(pressed(void)), SLOT(onOk(void)));

    connect(&m_cancel, SIGNAL(pressed(void)), SLOT(onCancel(void)));
}

void
StmChromaticityDialog::setXY(float x, float y)
{
    m_x = x;
    m_y = y;
    m_pane.setXY(x, y);
    UpdateText();
}


void
StmChromaticityDialog::onChromaticityChanged(float x, float y)
{
    m_x = x;
    m_y = y;
    m_selectCombo.setCurrentIndex(0);
    UpdateText();    
}

void
StmChromaticityDialog::onComboChanged(int index)
{
    // ignore if custom.
    if(index && (index < m_whitePoints.size()))
    {
        m_x = m_whitePoints[index].m_x;
        m_y = m_whitePoints[index].m_y;
        m_pane.setXY(m_x, m_y);
        UpdateText();
    }
}

void
StmChromaticityDialog::onOk(void)
{
    chromaticityChanged(m_x, m_y);
    close();
}

void
StmChromaticityDialog::onCancel(void)
{
    close();
}

void
StmChromaticityDialog::UpdateText(void)
{
    char str[64];
    sprintf(str, "%f, %f", m_x, m_y);

    m_label.setText(str);
    
}

////////////////////////////////////////////////////////////////////////////////

StmChromaticityWidget::StmChromaticityWidget(QWidget *parent)
    : QPushButton(parent)
    , m_x(0)
    , m_y(0)
{
    connect(this, SIGNAL(pressed(void)), SLOT(onPressed(void)));
    connect(&m_dialog, SIGNAL(chromaticityChanged(float, float)), 
            SLOT(onChromaticityChanged(float, float)));

    UpdateText();
}

StmChromaticityWidget::~StmChromaticityWidget()
{
}

void
StmChromaticityWidget::setXY(float x, float y)
{
    m_x = x;
    m_y = y;
    UpdateText();
}

void
StmChromaticityWidget::UpdateText(void)
{
    char str[64];
    sprintf(str, "%f, %f", m_x, m_y);

    setText(str);
}

void
StmChromaticityWidget::onPressed(void)
{
    m_dialog.setXY(m_x, m_y);
    m_dialog.show();
}

void 
StmChromaticityWidget::onChromaticityChanged(float x, float y)
{
    setXY(x, y);
    chromaticityChanged(x,y);
}
