#include <iostream>
#include <stdio.h>
#include <stdlib.h>

#include "utils.h"
#include "modules.h"
#include "processor.h"
#include "parameters.h"

#include <vector>

#include <QApplication>
#include <QWidget>

// Layout the doalog
#include <QGridLayout>

// supported widget types
#include <QLabel>
#include <QDoubleSpinBox>
#include <QSpinBox>
#include <QCheckBox>
#include <QComboBox>

#include <QColorDialog>

#include "StmProcessorDialog.h"

////////////////////////////////////////////////////////////////////////////////

StmParameterForm::StmParameterForm(QWidget *parent, spectral::Parameters *params)
    : QWidget(parent)
    , m_params(params)
    , m_layout(NULL)
    , m_dirty(NULL)
{
    m_layout = new QGridLayout(this);

    if(params)
    {
        std::list<std::string> names;
        params->GetParameterNames(names);
        std::list<std::string>::iterator j = names.begin();

        while(j != names.end())
        {
            AddParameter((*j), params);
            j++;
        }
    }
}

StmParameterForm::~StmParameterForm()
{
    //TODO
    if(m_params)
    {
        delete m_params;
    }
}

void 
StmParameterForm::setDirty(void)
{
    m_dirty = true;
    isDirty();
}


void 
StmParameterForm::AddParameter(std::string name, spectral::Parameters *params)
{
    if(params && m_layout)
    {
        QLabel *label;
        QWidget *widget;

        label = new QLabel(this);
        label->setText(name.c_str());
        widget = NULL;

        switch(params->GetParameterType(name))
        {
        case spectral::PARAM_TYPE_FLOAT:
            {
                float min, max, value;
                params->GetRange(name, min, max);
                params->GetParameter(name, value);

                QDoubleSpinBox *box = new QDoubleSpinBox(this);
                box->setMinimum(min);
                box->setMaximum(max);
                box->setValue(value);
                connect(box, SIGNAL(valueChanged(double)), SLOT(setDirty(void)));

                widget = box;
            }
            break;
        case spectral::PARAM_TYPE_INT:
            {
                int32_t min, max, value;
                params->GetRange(name, min, max);
                params->GetParameter(name, value);

                QSpinBox *box = new QSpinBox(this);
                box->setMinimum(min);
                box->setMaximum(max);
                box->setValue(value);
                connect(box, SIGNAL(valueChanged(int)), SLOT(setDirty(void)));

                widget = box;
            }
            break;
        case spectral::PARAM_TYPE_ENUM:
            {
                std::string s_value;
                uint32_t value;
                int index(0), n(0);
                std::list<std::pair<uint32_t, std::string> > range;

                params->GetParameter(name, value, s_value);
                params->GetRange(name, range);
               
                QComboBox *box = new QComboBox(this);
                std::list<std::pair<uint32_t, std::string> >::iterator i =
                    range.begin();

                while(i != range.end())
                {
                    if(i->first == value)
                    {
                        index = n;
                    }
                    box->addItem(i->second.c_str(), QVariant(i->first));
                    i++;
                    n++;
                }
                box->setCurrentIndex(index);
                connect(box, SIGNAL(currentIndexChanged(int)), SLOT(setDirty(void)));
                widget = box;
            }
            break;
        case spectral::PARAM_TYPE_CHROMATICITY:
            {
                float x, y;
                spectral::Chromaticity chrom;
                params->GetParameter(name, chrom);
                chrom.GetXY(x,y);
                char foo[50];
                sprintf(foo, "%f, %f", x, y);
                StmChromaticityWidget *button = new StmChromaticityWidget(this);
                button->setXY(x, y);
                connect(button, SIGNAL(chromaticityChanged(float, float)), SLOT(setDirty(void)));
                widget = button;
            }
            break;
        case spectral::PARAM_TYPE_BOOL:
            {
                bool value;
                params->GetParameter(name, value);

                QCheckBox *box = new QCheckBox(this);

                if(value)
                {
                    box->setCheckState(Qt::Checked);
                }
                else
                {
                    box->setCheckState(Qt::Unchecked);
                }
                connect(box, SIGNAL(stateChanged(int)), SLOT(setDirty(void)));

                widget = box;
            }
            break;
        }
        
        if(!widget)
        {
            QLabel *u_label = new QLabel(this);
            u_label->setText("Unsupported");
            widget = u_label;
        }

        {
            int row = m_layout->rowCount();
            m_layout->addWidget(label, row, 0);
            m_layout->addWidget(widget, row, 1);

            m_lines.push_back(UILine(name, label, widget));
        }
    }
}

void
StmParameterForm::Apply(void)
{
    if(m_params)
    {
        for(size_t i=0;i<m_lines.size();i++)
        {
            std::string name;
            name = m_lines[i].GetName(); 
            switch(m_params->GetParameterType(name))
            {
            case spectral::PARAM_TYPE_FLOAT:
                {
                    QDoubleSpinBox *box = dynamic_cast<QDoubleSpinBox*>(m_lines[i].GetWidget());
                    if(box)
                    {
                        float value = box->value();
                        m_params->SetParameter(name, value);
                    }
                }
                break;
            case spectral::PARAM_TYPE_INT:
                {
                    QSpinBox *box = dynamic_cast<QSpinBox*>(m_lines[i].GetWidget());
                    if(box)
                    {
                        int value = box->value();
                        m_params->SetParameter(name, value);
                    }
                }
                break;
            case spectral::PARAM_TYPE_ENUM:
                {
                    QComboBox *box = dynamic_cast<QComboBox*>(m_lines[i].GetWidget());
                    if(box)
                    {
                        QVariant var = box->itemData(box->currentIndex());
                        uint32_t value = var.toUInt();
                        m_params->SetParameter(name, value);
                    }
                }
                break;/*
            case spectral::PARAM_TYPE_CHROMATICITY:
                {
                    float x, y;
                    spectral::Chromaticity chrom;
                    params->GetParameter(name, chrom);
                    chrom.GetXY(x,y);
                }
                break;*/
            case spectral::PARAM_TYPE_BOOL:
                {
                    QCheckBox *box = dynamic_cast<QCheckBox*>(m_lines[i].GetWidget());
                    if(box)
                    {
                        m_params->SetParameter(name, (box->checkState() == Qt::Checked));
                    }
                }
                break;
            }
        }
    }
}

////////////////////////////////////////////////////////////////////////////////

StmProcessorDialog::StmProcessorDialog(QWidget *parent,
                                       spectral::Processor *processor)
    : QDialog(parent)
    , m_processor(processor)
    , m_layout(NULL)
    , m_buttons(NULL)
    , m_buttonLayout(NULL)
    , m_ok(NULL)
    , m_cancel(NULL)
    , m_form(NULL)
{
    spectral::Parameters *params(NULL);

    m_layout = new QVBoxLayout(this);

    m_buttons = new QWidget(this);
    m_buttonLayout = new QHBoxLayout(m_buttons);

    m_ok = new QPushButton(m_buttons);
    m_ok->setText("Ok");
    m_ok->setEnabled(false);

    m_cancel = new QPushButton(m_buttons);
    m_cancel->setText("Cancel");

    m_reset = new QPushButton(m_buttons);
    m_reset->setText("Reset");

    connect(m_ok, SIGNAL(pressed(void)), SLOT(onOk(void)));
    connect(m_cancel, SIGNAL(pressed(void)), SLOT(onCancel(void)));
    connect(m_reset, SIGNAL(pressed(void)), SLOT(onReset(void)));


    m_buttonLayout->addWidget(m_ok);
    m_buttonLayout->addWidget(m_cancel);
    m_buttonLayout->addWidget(m_reset);

    if(processor)
    {
        params = processor->GetParameters();
    }

    m_form = new StmParameterForm(this, params);
   
    connect(m_form, SIGNAL(isDirty(void)), SLOT(onDirty(void)));

    m_layout->addWidget(m_form); 
    m_layout->addWidget(m_buttons); 
}

void
StmProcessorDialog::onDirty(void)
{
    m_ok->setEnabled(true);
}

void
StmProcessorDialog::onOk(void)
{
    m_form->Apply();
    m_processor->SetParameters(m_form->GetParameters());
    close();
}

void
StmProcessorDialog::onCancel(void)
{
    close();
}

void
StmProcessorDialog::onReset(void)
{
    spectral::Parameters *params(NULL);

    if(m_processor)
    {
        params = m_processor->GetParameters();
    }

    m_ok->setEnabled(false);

    StmParameterForm *new_form = new StmParameterForm(this, params);

    m_layout->removeWidget(m_form);
    m_layout->removeWidget(m_buttons);
    
    m_layout->addWidget(new_form);
    m_layout->addWidget(m_buttons);

    delete m_form;
    m_form = new_form;
    connect(m_form, SIGNAL(isDirty(void)), SLOT(onDirty(void)));
}

