#include <iostream>
#include <stdio.h>
#include <stdlib.h>

#include "utils.h"
#include "image.h"
#include "bufutils.h"

#include <math.h>

#include <vector>

#include <QApplication>
#include <QWidget>

// supported widget types
#include <QPainter>
#include <QMouseEvent>
#include <QImage>

#include "StmView.h"

////////////////////////////////////////////////////////////////////////////////

spectral::Image *
StmViewWindow::CreateRGBImage(spectral::Image *img)
{
    spectral::Image *result(NULL);
    
    if(img)
    {
	int width = img->get_width();
        int height = img->get_height();
        size_t size = size_t(width) * size_t(height);

        result = new spectral::Image(width, height, CLS_SRGB_LINEAR);
        if(result)
        {
            switch(img->get_colour_space())
            {
            case CLS_SRGB_LINEAR:
            case CLS_SRGB: // TODO: Fix.
                fprintf(stderr, "Width = %i, Height = %i, CS = RGB\n", 
                        width, height);
                memcpy(result->get_buffer(), img->get_buffer(),
                       size * sizeof(float) * 3);
                break;
            case CLS_CIE_XYY:
                fprintf(stderr, "Width = %i, Height = %i, CS = XYY\n", 
                        width, height);
                memcpy(result->get_buffer(), img->get_buffer(),
                       size * sizeof(float) * 3);
                buf_xyy_to_linear_srgb(result->get_buffer(), width * height);
                break;
            case CLS_CIE_XYZ:
                fprintf(stderr, "Width = %i, Height = %i, CS = XYZ\n", 
                        width, height);
                memcpy(result->get_buffer(), img->get_buffer(),
                       size * sizeof(float) * 3);
                buf_xyz_to_xyy(result->get_buffer(), width * height);
                buf_xyy_to_linear_srgb(result->get_buffer(), width * height);
                break;
            case CLS_LUMINANCE:
                fprintf(stderr, "Width = %i, Height = %i, CS = Luminance\n", 
                        width, height);
                {
                    float *out, *in;
                    out = result->get_buffer();
                    in = img->get_buffer();
                    for(int i=0;i<size;i++)
                    {
                        *out++ = *in;
                        *out++ = *in;
                        *out++ = *in;
                        in++;
                    }
                }
                break;
            case CLS_DENSITY:
                fprintf(stderr, "Width = %i, Height = %i, CS = Density\n", 
                        width, height);
                {
                    float *out, *in;
                    out = result->get_buffer();
                    in = img->get_buffer();
                    for(int i=0;i<size;i++)
                    {
                        float val;
                        val = exp(*in);
                        *out++ = val;
                        *out++ = val;
                        *out++ = val;
                        in++;
                    }
                }
            default:
                break;
            }
        }
    }
    return result;
}

////////////////////////////////////////////////////////////////////////////////

StmViewWindow::GammaObserver::GammaObserver(StmViewWindow *pane)
    : QtGuiConfigObserver()
    , m_pane(pane)
{
    QtGuiConfig *config = QtGuiConfig::GetInstance();

    if(config)
    {
        Register((SimpleObservable *)config);
    }
}

////////////////////////////////////////////////////////////////////////////////

StmViewWindow::StmViewWindow(QWidget *parent)
    : QWidget(parent)
    , m_gamma(1.0)
    , m_image(NULL)
    , m_qt_image(NULL)
    , m_observer(this)
    , m_dirty(true)
    , m_scale(1)
    , m_normalise(true)
{
    QtGuiConfig *config = QtGuiConfig::GetInstance();
    if(config)
    { 
        m_gamma = config->GetScreenGamma();
    }
    setGamma(m_gamma);
}

////////////////////////////////////////////////////////////////////////////////

void 
StmViewWindow::setGamma(float gamma)
{
    if(gamma != m_gamma)
    {
        m_gamma = gamma;

        m_dirty = true;

        repaint();
    }
}

////////////////////////////////////////////////////////////////////////////////

void
StmViewWindow::SetImage(spectral::Image *p)
{
    int width, height; 
    
    if(m_image)
    {
        delete m_image;
    }

    if(p)
    {
        fprintf(stderr, "Generating RGB image\n");
        m_image = CreateRGBImage(p);

        delete p;
    }

    CalculateSize(width, height);

    setFixedSize(width, height);

    m_dirty = true;
   
    repaint();
}

////////////////////////////////////////////////////////////////////////////////

void
StmViewWindow::onScaleChanged(float scale)
{
    int width, height; 
    m_scale = scale;

    CalculateSize(width, height);
    setFixedSize(width, height);
    
    repaint();
}

////////////////////////////////////////////////////////////////////////////////

void
StmViewWindow::onNormaliseChanged(bool normalise)
{
    if(normalise != m_normalise)
    {
        m_normalise = normalise;
        m_dirty = true;
        repaint();
    }
}

////////////////////////////////////////////////////////////////////////////////

void
StmViewWindow::CalculateSize(int &width, int &height)
{
    if(m_image)
    {
        float fw, fh;
        fw = m_image->get_width();
        fh = m_image->get_height();
        fw*= m_scale;
        fh*= m_scale;
        width = floor(fw);
        height = floor(fh);
    }
    else
    {
        width = height = 16;
    }
}

////////////////////////////////////////////////////////////////////////////////

void
StmViewWindow::paintEvent(QPaintEvent *)
{
    QPainter painter(this);

    if(m_dirty)
    {
        fprintf(stderr, "Generating display image\n");
        if(m_qt_image)
        {
            delete m_qt_image;
            m_qt_image = NULL;
        }

        if(m_image)
        {
            m_qt_image = new QImage(m_image->get_width(),
                                    m_image->get_height(),
                                    QImage::Format_RGB888);
            float scale_factor = 1.0;

            // Scale image
            if(m_normalise)
            {
                float *p = m_image->get_buffer();
                size_t size = size_t(m_image->get_width()) * size_t(m_image->get_height()) * 3;
                scale_factor = 0.0;

                for(;size;size--,p++)
                {
                    if(scale_factor < *p)
                    {
                        scale_factor = *p;
                    }
                }                
            }

            fprintf(stderr, "Luminance scale factor = %f\n", scale_factor); 

            {
                float *in = m_image->get_buffer();
                uint8_t *out = m_qt_image->bits();
                size_t size = size_t(m_image->get_width()) * size_t(m_image->get_height()) * 3;
                float inv_gamma;

                if(m_gamma > 0.25)
                {
                    inv_gamma = 1.0 / m_gamma;
                }
                else 
                {
                    inv_gamma = 4;
                }

                for(;size;size--,in++,out++)
                {
                    float val;
                    int ival;
                    val = *in;
                    val/= scale_factor;

                    val = pow(val, inv_gamma);
                    
                    if(val < 0.0) val = 0.0;
                    if(val > 1.0) val = 1.0;

                    ival = floor(val * 256.0);
                    if(ival > 255) ival = 255;
                    if(ival < 0) ival = 0;
 
                    *out = ival;
                }                
            }
        }
        m_dirty = false;
    }

    if(m_qt_image)
    {
        int width, height;
        CalculateSize(width, height);
        painter.drawImage(QRect(0,0, width, height), *m_qt_image);
    }
}

////////////////////////////////////////////////////////////////////////////////

StmViewControls::StmViewControls(QWidget *parent)
    : QWidget(parent)
    , m_layout(this)
    , m_scaleSelector(this)
    , m_imageSelector(this)
    , m_normalise(this)
{
    m_layout.addWidget(&m_imageSelector, 0, 0);
    m_layout.addWidget(&m_scaleSelector, 0, 1);
    m_layout.addWidget(&m_normalise, 0, 2);

    // Set up mormalisation.
    m_normalise.setText("Normalise");
    m_normalise.setCheckState(Qt::Checked);

    // Set up scales.
    m_scaleSelector.addItem("800%",   QVariant(8.0));
    m_scaleSelector.addItem("400%",   QVariant(4.0));
    m_scaleSelector.addItem("200%",   QVariant(2.0));
    m_scaleSelector.addItem("100%",   QVariant(1.0));
    m_scaleSelector.addItem("50%",    QVariant(0.5));
    m_scaleSelector.addItem("25%",    QVariant(0.25));
    m_scaleSelector.addItem("12.5%",  QVariant(0.125));
    m_scaleSelector.addItem("6.25%",  QVariant(0.0625));
    m_scaleSelector.addItem("3.125%", QVariant(0.03125));
    m_scaleSelector.setCurrentIndex(3);

    // Set up slots.
    m_imageSelector.addItem("0 - source", QVariant(0));

    connect(&m_scaleSelector, SIGNAL(currentIndexChanged(int)), SLOT(onScaleSelected(int)));
    connect(&m_normalise, SIGNAL(stateChanged(int)), SLOT(onNormaliseChanged(int)));
}

StmViewControls::~StmViewControls()
{
}

void
StmViewControls::onScaleSelected(int idx)
{
    float scale = 1.0;

    scale = m_scaleSelector.itemData(idx).toFloat();

    scaleChanged(scale);
}

void
StmViewControls::onNormaliseChanged(int state)
{
    normaliseChanged(state == Qt::Checked);
}

////////////////////////////////////////////////////////////////////////////////

StmView::StmView(QWidget *parent)
    : QWidget(parent)
    , m_requestedView(0)
    , m_mainLayout(this)
    , m_scrollArea(this)
    , m_viewWindow(&m_scrollArea)
    , m_controlWindow(this)
{
 
    m_scrollArea.setBackgroundRole(QPalette::Dark);
    m_scrollArea.setWidget(&m_viewWindow);

    m_mainLayout.addWidget(&m_scrollArea, 0, 0);
    m_mainLayout.addWidget(&m_controlWindow, 1, 0);
    //setMinimumSize(512, 512);
    //m_sizeHint = QSize(512,512);
    m_viewWindow.connect(&m_controlWindow, SIGNAL(scaleChanged(float)), SLOT(onScaleChanged(float)));
    m_viewWindow.connect(&m_controlWindow, SIGNAL(normaliseChanged(bool)), SLOT(onNormaliseChanged(bool)));
}


////////////////////////////////////////////////////////////////////////////////

StmView::~StmView()
{
}

void
StmView::onImageData(int id, void *p)
{
    fprintf(stderr, "StmView::onImageData\n");

    if(p)
    {
        spectral::Image *img = reinterpret_cast<spectral::Image*>(p);

        if(id == m_requestedView)
        {
            m_viewWindow.SetImage(img);
        }
        else
        {
            delete img;
        }
    }
}
