#include <stdio.h>
#include "controller.h"
#include "thread.h"
#include "StmWrapper.h"

StmWrapper::StmWrapper(const std::string &modulePath,
                       const std::list<std::string> &modules)
    : QThread()
    , m_controller(NULL)
{
    int num_threads = thread::GetProcessorCount();
    thread::ThreadPool::GetInstance()->SetNumThreads(num_threads);

    m_controller = new control::Controller(modulePath, modules);
    start();
}

StmWrapper::~StmWrapper()
{
    if(m_controller)
    {
        fprintf(stderr, "Shutting down STM controller\n");
        delete m_controller;
    }
}

void
StmWrapper::run()
{
    if(m_controller)
    {
        bool terminate(false);
        do
        {
            control::Message *msg;
            msg = (m_controller->GetOutputQueue()).Pop();
            if(msg)
            {
                switch(msg->GetType())
                {
                case control::Message::MSG_START:          // Start processing.
                    printf("Unhandled message %x\n", msg->GetType());
                    break;
                case control::Message::MSG_STOP:           // Stop processing.
                    printf("Unhandled message %x\n", msg->GetType());
                    break;
                case control::Message::MSG_SHUTDOWN:       // Terminate thread.
                    printf("Unhandled message %x\n", msg->GetType());
                    break;
                case control::Message::MSG_EXIT:       // Terminate.
                    printf("Shut down received\n");
                    terminate = true;
                    break;
                case control::Message::MSG_ACTOR_INSTALLED:      // Add a processor actor
                    printf("Installed processor %s in slot %i\n", 
                    msg->GetString().c_str(), msg->GetId());
                    processorAdded(msg->GetId(), msg->GetString().c_str());
                    break;
                case control::Message::MSG_ACTOR_REMOVED:      // Add a processor actor
                    printf("Removed processor from slot %i\n", msg->GetId());
                    processorRemoved(msg->GetId());
                    break;
                case control::Message::MSG_INPUT_ADDED:      // Add a link
                    printf("Added link from slot %i to %i:%i\n", 
                           msg->GetId(), msg->GetLink().second, msg->GetLink().first);
                    break;
                case control::Message::MSG_INPUT_REMOVED:      // removed a link
                    printf("Removed link from slot %i to %i:%i\n", 
                           msg->GetId(), msg->GetLink().second, msg->GetLink().first);
                    break;
                case control::Message::MSG_SET_INPUT:      // Set an input.
                    printf("Unhandled message %x\n", msg->GetType());
                    break;
                case control::Message::MSG_PARAMETERS:     // Set parameters
                    printf("Unhandled message %x\n", msg->GetType());
                    break;
                case control::Message::MSG_SET_IMAGE:      // Set an image.
                    printf("Received image in slot %i\n", msg->GetId());
                    imageData(msg->GetId(), msg->GetImage());
                    break;
                case control::Message::MSG_IMAGE_LOAD_SUCCEEDED:     // Load an image.
                    printf("Loaded image in slot %i\n", msg->GetId());
                    imageLoaded(msg->GetId());
                    break;
                case control::Message::MSG_IMAGE_LOAD_FAILED:     // Load an image.
                    printf("failed to load image in slot %i\n", msg->GetId());
                    break;
                case control::Message::MSG_PROGRESS:       // Report progress
                    printf("Unhandled message %x\n", msg->GetType());
                    break;
                case control::Message::MSG_ERROR:          // Report an error
                    printf("ERROR: %s (%i)\n", msg->GetString().c_str(), msg->GetId());
                    break;
                case control::Message::MSG_STARTED:        // Begin working.
                    printf("Unhandled message %x\n", msg->GetType());
                    break;
                case control::Message::MSG_STOPPED:        // Finish working.
                    break;
                case control::Message::MSG_RESULT:         // Pass a result.
                    printf("Unhandled message %x\n", msg->GetType());
                    break;
                case control::Message::MSG_REQUEST_PARAMETERS: // Get parameters.
                    printf("Unhandled message %x\n", msg->GetType());
                    break;
                default:
                    printf("Unknown message %x\n", msg->GetType());
                    break;
                }
                delete msg;
            }
        } while(!terminate);
    }
    else
    {
        fprintf(stderr, "Error in StmWrapper::run() - Invalid controller.\n");
    }
}
