#include <iostream>
#include <stdio.h>
#include <stdlib.h>

#include "utils.h"
#include "modules.h"
#include "processor.h"
#include "parameters.h"

void output_parameter(std::string name, spectral::Parameters *params)
{
    std::cout << "    " << name << ": ";

    if(params)
    {
        switch(params->GetParameterType(name))
        {
        case spectral::PARAM_TYPE_FLOAT:
            {
                float min, max, value;
                params->GetRange(name, min, max);
                params->GetParameter(name, value);
                std::cout << "float = " << value 
                          << " (" << min << "," << max << ")" << std::endl;
            }
            break;
        case spectral::PARAM_TYPE_INT:
            {
                int32_t min, max, value;
                params->GetRange(name, min, max);
                params->GetParameter(name, value);
                std::cout << "int = " << value 
                          << " (" << min << "," << max << ")" << std::endl;
            }
            break;
        case spectral::PARAM_TYPE_ENUM:
            {
                std::string s_value;
                uint32_t value;
                std::list<std::pair<uint32_t, std::string> > range;

                params->GetParameter(name, value, s_value);
                params->GetRange(name, range);

                std::cout << "enum = "
                          << "[" << value << "," << s_value << "]"
                          << "(";
                std::list<std::pair<uint32_t, std::string> >::iterator i =
                    range.begin();

                while(i != range.end())
                {
                    std::cout << "[" << i->first << "," << i->second << "], ";
                    i++;
                }
                std::cout << ")" << std::endl;
            }
            break;
        case spectral::PARAM_TYPE_CHROMATICITY:
            {
                float x, y;
                spectral::Chromaticity chrom;
                params->GetParameter(name, chrom);
                chrom.GetXY(x,y);

                std::cout << "xy colour = "
                          << "<" << x << "," << y << ">"
                          << std::endl;
            }
            break;
        case spectral::PARAM_TYPE_BOOL:
            {
                bool value;
                params->GetParameter(name, value);
                std::cout << "boolean = " << (value?"true":"false")
                          << std::endl;
            }
            break;
        default:
            std::cout << "unrecognised type." << std::endl; 
        }
    }
}

std::string translate_cs(uint32_t type)
{
    std::string result;

    switch(type)
    {
    case CLS_SRGB_LINEAR:
        result = "Linear sRGB";
        break;
    case CLS_SRGB:
        result = "sRGB";
        break;
    case CLS_CIE_XYZ:
        result = "XYZ";
        break;
    case CLS_CIE_XYY:
        result = "xyY";
        break;
    case CLS_LUMINANCE:
        result = "Luminance";
        break;
    case CLS_DENSITY:
        result = "Density";
        break;
    default:
        result = "Unsupported";
    }

    return result;
}

int main(int argc, char **argv)
{
    spectral::Modules modules;

    std::string path;

    {
        char *p = getenv("STM_MODULE_PATH");
        if(p)
        {
            path = p;
        }
    }

    if(argc > 1)
    {
        for(;argc>1;argc--)
        {
            std::string name(argv[argc-1]);

            bool b;
            b = modules.LoadModule(path, name);
            if(b)
            {
                std::cout << "Loaded module " << name << std::endl;
            }
            else
            {
                std::cout << "Module " << name << "not loaded" << std::endl;
            }
        }
        {
            std::map<std::string, std::string> exts;

            modules.GetLoaders(exts);
            std::map<std::string, std::string>::iterator i = exts.begin();


            std::cout << "*** FILE LOADERS ***" << std::endl;
            while(i != exts.end())
            {
                std::cout << "Extension: \"" << i->first 
                          << "\" handled by module " << i->second
                          << std::endl;
                i++;
            }
        }
        {
            std::map<std::string, std::string> exts;

            modules.GetSavers(exts);
            std::map<std::string, std::string>::iterator i = exts.begin();

            std::cout << "*** FILE SAVERS ***" << std::endl;
            while(i != exts.end())
            {
                std::cout << "Extension: \"" << i->first 
                          << "\" handled by module " << i->second
                          << std::endl;
                i++;
            }
        }
        {
            std::set<std::string> processors;

            modules.GetProcessors(processors);
            std::set<std::string>::iterator i = processors.begin();

            std::cout << "*** PROCESSORS ***" << std::endl;

            while(i != processors.end())
            {
                spectral::Processor *processor;

                std::string name, desc;
                name = *i;
                modules.GetModuleDesc(name, desc);
                std::cout << std::endl
                          <<"Processor \"" << desc << "\" (" << name << ")" 
                          << std::endl;

                processor = (spectral::Processor *)modules.CreateObject(name);

                if(processor)
                {
                    // Dump input & output types.
                    {
                        std::list<std::pair<uint32_t, uint32_t> > mandatory,
                                                                  optional;
                        uint32_t type;
                        type = processor->GetInterface(mandatory, optional);
                        std::cout << "Output: " << translate_cs(type) 
                                  << std::endl;

                        std::list<std::pair<uint32_t, uint32_t> >::iterator i;

                        std::cout << "Inputs (mandatory): ";

                        i = mandatory.begin(); 
                        while(i != mandatory.end())
                        {
                            std::cout << i->first << ": " 
                                      << translate_cs(i->second) << ", ";
                            i++;
                        }
                        std::cout << std::endl;
                        std::cout << "Inputs (optional):  ";

                        i = optional.begin(); 
                        while(i != optional.end())
                        {
                            std::cout << i->first << ": " 
                                      << translate_cs(i->second) << ", ";
                            i++;
                        }
                        std::cout << std::endl;
                    }

                    spectral::Parameters *parameters =
                        processor->GetParameters();

                    if(parameters)
                    {
                        std::cout << "Parameters:" << std::endl;

                        std::list<std::string> names;
                        parameters->GetParameterNames(names);
                        std::list<std::string>::iterator j = names.begin();

                        while(j != names.end())
                        {
                            output_parameter((*j), parameters);
                            j++;
                        }
                        delete parameters;
                    }
                    delete processor;
                }
                else
                {
                    std::cout << "unable to create object" << std::endl;
                }

                i++;
            }
        }
    }
    else
    {
        std::cout << "usage: test_module <module1> <module2> <...>" 
                  << std::endl;
    }
    return 0;
}
