#include <iostream>
#include <fstream>
#include <sstream>

#include <stdlib.h>

#include "parameters.h"
#include "graphloader.h"
#include "controller.h"

#undef yyFlexLexer
#define yyFlexLexer stmFlexLexer
//#include "FlexLexer.h"
#include <FlexLexer.h>
#undef yyFlexLexer

namespace graphloader {

bool 
InstallProcessor(int id, stmFlexLexer &scanner, control::Controller *cnt)
{
    bool result(true);

    if(scanner.yylex() == PROCESSOR_NAME)
    {
        std::string name = scanner.YYText();

        if(id)
        {
           std::string prcname = name.substr(1, name.length() - 2);
 
           cnt->AddActor(id, prcname);
        }
        else
        {
            std::cerr << "Warning: Unable to load " << name 
                      << " into process 0" << std::endl;
        }
    }
    else
    {
        result = false;
        std::cerr << "Parse error." << std::endl;
    }
    return result;
}

bool 
AddDependency(int dest_id, stmFlexLexer &scanner, control::Controller *cnt)
{
    bool result(true);
    uint32_t dest_slot = 0;
    int32_t source = 0;
    int tok;

    tok = scanner.yylex();
    if(tok == DEP_SLOT)
    {
        dest_slot = atoi(scanner.YYText());
    }
    else
    {
        result = false;
        std::cerr << "Parse error: " << tok << std::endl;
    }

    if(result)
    {
        if(scanner.yylex() == DEP_SOURCE)
        {
            source = atoi(scanner.YYText());
        }
        else
        {
            result = false;
            std::cerr << "Parse error: " << tok << std::endl;
        }
    }

    if(result)
    {
        cnt->AddLink(dest_id, dest_slot, source);
    }

    return result;
}

bool 
SetParameter(int dest_id, stmFlexLexer &scanner, control::Controller *cnt)
{
    bool result(true);
    spectral::Parameters *params = NULL;
    std::string name;

    // Get parameters.
    cnt->RequestParameters(dest_id);

    {
        bool received(false);
        do
        {
            control::Message *msg;
            msg = cnt->GetOutputQueue().Pop();

            if(msg)
            {
                if(msg->GetType() == control::Message::MSG_PARAMETERS)
                {
                    params = msg->GetParameters();
                    received = true;
    	        }
            }
            delete msg;
        } while(!received);
    }
    
    std::cerr << dest_id << ": ";

    // Get parameter name.
    if(scanner.yylex() == PARAM_NAME)
    {
 	std::string str = scanner.YYText();

        name = str.substr(1, str.length() - 2);
        std::cerr << "param name = \"" << name << "\", ";
    }
    else
    {
        result = false;
        //std::cerr << "Parse error. " << tok << std::endl;
    }

    if(result)
    {
        switch(scanner.yylex())
        {
        case PARAM_VALUE_INT:
            {
                int32_t value = atoi(scanner.YYText());
                std::cerr << "int value = " << value << std::endl;

                if(params)
                {
                    params->SetParameter(name, value);
                }
            }
            break;
        case PARAM_VALUE_FLOAT:
            {
                float value = atof(scanner.YYText());
                std::cerr << "float value = " << value << std::endl;

                if(params)
                {
                    params->SetParameter(name, value);
                }
            }
            break;
        case PARAM_VALUE_BOOL:
            {
                bool value(false);

                switch(*scanner.YYText())
                {
                case 't':
                case 'T':
                    value = true;
                    break;
                case 'f':
                case 'F':
                    value = false;
                    break;
                default:
                    std::cerr << "Warning.  Invalid bool.  Defaulting to false."
                              << std::endl;
                }
                std::cerr << "bool value = " 
                          << (value?"true":"false") << std::endl;

                if(params)
                {
                    params->SetParameter(name, value);
                }
            }
            break;
        case PARAM_VALUE_CHROMATICITY:
            {
                // chromaticity looks like <float,float>
                float x, y;
 	        std::string str = scanner.YYText();
                std::string str2= str.substr(1, str.length() - 2);

                size_t loc = str2.find(',');
                std::string strx=str2.substr(0, loc);
                std::string stry=str2.substr(loc + 1, str2.npos - loc);

                x = atof(strx.c_str());
                y = atof(stry.c_str());
                std::cerr << "strx = " << strx << " stry = " << stry << std::endl;

                std::cerr << "colour value = <" << x << ", "
                          << y << ">" << std::endl;
                if(params)
                {
                    params->SetParameter(name, spectral::Chromaticity(x,y));
                }
            }
            break;
        case PARAM_VALUE_STRING:
            {
                std::string value;
 	        std::string str = scanner.YYText();

                value = str.substr(1, str.length() - 2);
                std::cerr << "string value = " << value << std::endl;
                if(params)
                {
                    //params->SetParameter(name, value);
                }
            }
            break;
        default:
            //std::cerr << "Parse error. " << tok << std::endl;
            result = false;
        }
    }

    if(result && params)
    {
        cnt->SetParameters(dest_id, params);
    }

    return result;
}

////////////////////////////////////////////////////////////////////////////////
// External interface
////////////////////////////////////////////////////////////////////////////////

// Half arsed parser for process graph files.
// Can't use bison in c++ mode, because I enjoy life too much and I aint 
// getting paid.
bool 
LoadGraph(const std::string &fn, control::Controller *cnt, int32_t &output_id)
{
    bool result(false);

    output_id = 0;

    if(cnt && fn.length())
    {
        std::ifstream in(fn.c_str());

        if(in.is_open())
        {
            std::ostringstream out;
            stmFlexLexer scanner(&in, &out);
            result = true;

            int token;

            do
            {
                token = scanner.yylex();

                switch(token)
                {
                case PROCESSOR_ID:
                    {
                        int id = atoi(scanner.YYText());
                        result = InstallProcessor(id, scanner, cnt);
                    }
                    break;
                case DEP_DEST:
                    {
                        int id = atoi(scanner.YYText());
                        result = AddDependency(id, scanner, cnt); 
                    }
                    break;
                case PARAM_DEST:
                    {
                        int id = atoi(scanner.YYText());
                        result = SetParameter(id, scanner, cnt); 
                    }
                    break;
                case OUTPUT_ID:
                    output_id = atoi(scanner.YYText());
                    token = END;
                case END:
                    break;
                default: // Parse error.
                    result = false;
                    break;
                }
            }
            while((token != END) && result);
        }
    }
    return result;
}

}
