#include <iostream>
#include <fstream>

#include <stdlib.h>

#include "thread.h"
#include "controller.h"
#include "graphloader.h"

void AwaitCompletion(control::Controller *stm);

int main(int argc, char **argv)
{
    std::string            path;
    std::list<std::string> modules;

    // Get module path.
    {
        char *p = getenv("STM_MODULE_PATH");

        if(p)
        {
            path = p;
        }
    }
    // Get module list
    {
#ifdef _WIN32
        std::string modfile = path + "\\modules";
#else
        std::string modfile = path + "/modules";
#endif 
        std::ifstream str;
        str.open(modfile.c_str(), std::ifstream::in);

        if(str.is_open())
        {
            while(!str.eof())
            {
                std::string module;
                str >> module;
                if(module.length())
                {
                    std::cerr << "Identified module \"" << module << "\"." 
                              << std::endl;
                    modules.push_back(module);
                }
            }
        }
        else
        {
            std::cerr << "Failed to open \"modules\" file: "
                      << modfile
                      << std::endl;
        }
    }
    if(modules.size() == 0)
    {
        std::cerr << "No modules found.  "
                  << "Check STM_MODULE_PATH environment variable." 
                  << std::endl;
    }
    else
    {
        control::Controller *stm(NULL);

        if(argc < 2)
        {
            std::cerr << "tonemap -c config -i input -o output" << std::endl;
        }
        else
        { 
            std::string input_name, output_name, config_file;
            int num_threads = thread::GetProcessorCount();

            {
                int c;
                while((c = getopt(argc, argv, "c:i:o:t:")) != EOF)
                {
                    switch(c)
                    {
                    case 'c':
                        config_file = optarg;
                        break;
                    case 'i':
                        input_name = optarg;
                        break;
                    case 'o':
                        output_name = optarg;
                        break;
                    case 't':
                        num_threads = atoi(optarg);
                        break;
                    default:
                        break;
                    }
                }
            }
            if(num_threads < 1)
            {
                num_threads = 1;
            }
            if(input_name.length() == 0)
            {
                std::cerr<< "tonemap -c config -i input -o output [-t threads]" 
                         << std::endl;
                exit(0);
            }
            if(output_name.length() == 0)
            {
                output_name = input_name + ".png";
            }
            if(config_file.length() == 0)
            {
                std::cerr<< "tonemap -c config -i input -o output [-t threads]" 
                         << std::endl;
                exit(0);
            }
            {
                thread::ThreadPool::GetInstance()->SetNumThreads(num_threads);
            }
            

            stm = new control::Controller(path, modules);

            if(stm)
            {
                int32_t output_id;
                std::cerr << "Started STM controller." << std::endl;
   
                stm->LoadImage(input_name); 


                // TODO:
                if(graphloader::LoadGraph(config_file, stm, output_id))
                {
                    stm->Process();
                    AwaitCompletion(stm);
                }
                else
                {
                    std::cerr << "Graph loading failed" << std::endl;
                }

                stm->SaveImage(output_id, output_name);

                stm->Shutdown();

                delete stm;
            }
            else
            {
                std::cerr << "Unable to start STM controller." << std::endl;
            }
        }
    }
    
    return 0;
}

void AwaitCompletion(control::Controller *stm)
{
    bool cont(true);
    while(cont)
    {
        control::Message *msg;
        msg = stm->GetOutputQueue().Pop();
        if(msg)
        {
            switch(msg->GetType())
            {
            case control::Message::MSG_STOPPED:
                cont = false;
                break;
            case control::Message::MSG_PROGRESS:
                {
                    int id = msg->GetId();
                    float progress = msg->GetFloat();
                    fprintf(stderr, "Job %i %f%% complete.\n", id, progress);
                }
                break;
            case control::Message::MSG_ACTOR_INSTALLED:      // Add a processor actor
                fprintf(stderr, "Installed processor %s in slot %i\n",
                msg->GetString().c_str(), msg->GetId());
                break;
            case control::Message::MSG_ACTOR_REMOVED:      // Add a processor actor
                fprintf(stderr, "Removed processor from slot %i\n", msg->GetId());
                break;
            case control::Message::MSG_ERROR:
                fprintf(stderr, "Error: %s\n", msg->GetString().c_str()); 
                break;
            default:
                break;
            }
            delete msg;
        }
    }
}

