#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "utils.h"
#include "thread.h"
#include "image.h"
#include "parameters.h"

#include "rays.h"

namespace generators {

const float LOG2 = 0.6931471805599453;

struct line_params
{
    float total;
    float half_life;
};

static bool
DrawPixel(spectral::Image *image, void *p, 
                int x, int y, float start, float end)
{
    bool result(false);
    if(image && p)
    {
        struct line_params *params; 
        float contrib;
        if((x > 0) && 
           (y > 0) && 
           (uint32_t(x) < image->get_width()) && 
           (uint32_t(y) < image->get_height()))
        {
            params = (struct line_params *)p;
            {
                float k = -LOG2/params->half_life;
                float tmp1, tmp2;
                tmp1 = exp(k * start);
                tmp2 = exp(k * end);
                contrib = tmp2 - tmp1 / k;
            }
            {
                float value = *(image->get_pixel(x, y));
                value+=contrib;
                image->set_pixel(x, y, &value);
            }
            params->total+= contrib;
    
            result = true;
        }
    }
    return result;
}



Rays::Rays()
    : Generator()
    , m_halflife(20)
    , m_minimum(1e-5)
    , m_centre_weight(0.8)
    , m_angle(0)
    , m_num_rays(12)
{
}

Rays::~Rays()
{
}

// Get slots & output type.  Returns output colour space.
// pairs are <slot id, colour space>
uint32_t
Rays::GetInterface(
    std::list<std::pair<uint32_t, uint32_t> > &mandatory,
    std::list<std::pair<uint32_t, uint32_t> > &optional) const
{   
    mandatory.clear();
    optional.clear();
        
    return CLS_LUMINANCE;
}  

bool
Rays::SetImage(uint32_t, const spectral::Image *)
{
    bool result = false;

    return result;
}

void
Rays::ClearImage(uint32_t)
{
}

void 
Rays::GetParameters(spectral::Parameters *params) const 
{
    if(params)
    {
        params->AddParameter(std::string("half life"),
                             0.5, 2000.0, m_halflife);
        params->AddParameter(std::string("minimum"),
                             1e-6, 0.5, m_minimum);
        params->AddParameter(std::string("unscattered"),
                             1e-6, 1.0, m_centre_weight);
        params->AddParameter(std::string("number of rays"),
                             3, 64, m_num_rays);
        params->AddParameter(std::string("angle"),
                             -360.0, 360.0, m_angle);

    }
}

// Apply parameters to a processor.
void
Rays::SetParameters(const spectral::Parameters *params)
{
    if(params)
    {
        params->GetParameter(std::string("half life"), m_halflife);
        params->GetParameter(std::string("minimum"), m_minimum);
        params->GetParameter(std::string("unscattered"), m_centre_weight);
        params->GetParameter(std::string("number of rays"), m_num_rays);
        params->GetParameter(std::string("angle"), m_angle);
    }

}

spectral::Image *
Rays::Process(void)
{
    spectral::Image *result = NULL;

    int32_t diameter;
    float radius;

    {
        radius = fabs(log(m_minimum)/log(2.0) * m_halflife);
        diameter = floor(radius);
        diameter+=4;
        diameter*= 2;
    }

    result = new spectral::Image(diameter, diameter, CLS_LUMINANCE);

    if(result)
    {
        Progress(0);
        int32_t x, y;
        int32_t centre = diameter/2;

        struct line_params params;
        params.total = 0.0;
        params.half_life = m_halflife;
     
        for(y=0;y<diameter;y++)
        {
            for(x=0;x<diameter;x++)
            {
                float val(0);
                result->set_pixel(x,y,&val);
            }
        }

        for(int32_t i = 0;i<m_num_rays;i++)
        {
            float angle = i;
            angle*=360;
            angle/= m_num_rays;
            angle+= m_angle;

            angle*=3.14159; // TODO: clean this up.
            angle/=180;
            float fx, fy;
            
            fx = (float)centre + 0.5 + (cos(angle) * radius);
            fy = (float)centre + 0.5 + (sin(angle) * radius);

            DrawLine(result, 
                     &params, 
                     0.5 + (float) centre, 
                     0.5 + (float) centre, 
                     fx, 
                     fy,
                     DrawPixel);
        }
        {
           
            float scale(1.0);
            float midpoint;
            if(params.total > 1e-6)
            {
                scale = (1.0 - m_centre_weight) / params.total;
            }

            for(size_t i=0;i<result->get_width()*result->get_height();i++)
            {
                (result->get_buffer())[i]*= scale;
            }

            midpoint = *(result->get_pixel(centre, centre));
            midpoint+= m_centre_weight;
            result->set_pixel(centre, centre, &midpoint);
        }
    }
    Progress(100);

    return result;
}

}
