#include "bufutils.h"
#include "thread.h"
#include "utils.h"
#include <math.h>

#define JOB_SIZE 16000

////////////////////////////////////////////////////////////////////////////////
class xyYToXYZJob : public thread::Job
{
public:
    xyYToXYZJob(float *buf, size_t size)
        : thread::Job()
        , m_buf(buf)
        , m_size(size)
    {}

    virtual void Process(void)
    {
        if(m_buf)
        {
            size_t i;
            float *p = m_buf;

            for(i=0;i<m_size;i++)
            {
                xyy_to_xyz(p);
                p+=3;
            }
        }
    }

private:
    float *m_buf;
    size_t m_size;
};

////////////////////////////////////////////////////////////////////////////////
class XYZToxyYJob : public thread::Job
{
public:
    XYZToxyYJob(float *buf, size_t size)
        : thread::Job()
        , m_buf(buf)
        , m_size(size)
    {}

    virtual void Process(void)
    {
        if(m_buf)
        {
            size_t i;
            float *p = m_buf;

            for(i=0;i<m_size;i++)
            {
                xyz_to_xyy(p);
                p+=3;
            }
        }
    }

private:
    float *m_buf;
    size_t m_size;
};

////////////////////////////////////////////////////////////////////////////////
class xyYToSRGBJob : public thread::Job
{
public:
    xyYToSRGBJob(float *buf, size_t size)
        : thread::Job()
        , m_buf(buf)
        , m_size(size)
    {}

    virtual void Process(void)
    {
        if(m_buf)
        {
            size_t i;
            float *p = m_buf;

            for(i=0;i<m_size;i++)
            {
                xyy_to_linear_srgb(p);
                p+=3;
            }
        }
    }

private:
    float *m_buf;
    size_t m_size;
};
////////////////////////////////////////////////////////////////////////////////
class SRGBToxyYJob : public thread::Job
{
public:
    SRGBToxyYJob(float *buf, size_t size)
        : thread::Job()
        , m_buf(buf)
        , m_size(size)
    {}

    virtual void Process(void)
    {
        if(m_buf)
        {
            size_t i;
            float *p = m_buf;

            for(i=0;i<m_size;i++)
            {
                linear_srgb_to_xyy(p);
                p+=3;
            }
        }
    }

private:
    float *m_buf;
    size_t m_size;
};
////////////////////////////////////////////////////////////////////////////////
class ScaleYJob: public thread::Job
{
public:
    ScaleYJob(float *buf, size_t size, float scale)
        : thread::Job()
        , m_buf(buf)
        , m_size(size)
        , m_scale(scale)
    {}

    virtual void Process(void)
    {
        if(m_buf)
        {
            size_t i;
            float *p = m_buf;

            for(i=0;i<m_size;i++)
            {
                p[2]*=m_scale;
                p+=3;
            }
        }
    }

private:
    float *m_buf;
    size_t m_size;
    float m_scale;
};
////////////////////////////////////////////////////////////////////////////////
class ScaleJob: public thread::Job
{
public:
    ScaleJob(float *buf, size_t size, float scale)
        : thread::Job()
        , m_buf(buf)
        , m_size(size)
        , m_scale(scale)
    {}

    virtual void Process(void)
    {
        if(m_buf)
        {
            size_t i;
            float *p = m_buf;

            for(i=0;i<m_size;i++)
            {
                (*p)*=m_scale;
                p++;
            }
        }
    }

private:
    float *m_buf;
    size_t m_size;
    float m_scale;
};

////////////////////////////////////////////////////////////////////////////////
class LuminanceToDensityJob: public thread::Job
{
public:
    LuminanceToDensityJob(float *buf, size_t size, float min)
        : thread::Job()
        , m_buf(buf)
        , m_size(size)
        , m_min(min)
    {}

    virtual void Process(void)
    {
        if(m_buf)
        {
            size_t i;
            float *p = m_buf;

            for(i=0;i<m_size;i++)
            {
                if((*p) < m_min)
                {
                    *p = log(m_min);
                }
                else
                {
                    (*p) = log(*p);
                }
                p++;
            }
        }
    }

private:
    float *m_buf;
    size_t m_size;
    float m_min;
};
////////////////////////////////////////////////////////////////////////////////
class DensityToLuminanceJob: public thread::Job
{
public:
    DensityToLuminanceJob(float *buf, size_t size)
        : thread::Job()
        , m_buf(buf)
        , m_size(size)
    {}

    virtual void Process(void)
    {
        if(m_buf)
        {
            size_t i;
            float *p = m_buf;

            for(i=0;i<m_size;i++)
            {
                *p = exp(*p);
                p++;
            }
        }
    }

private:
    float *m_buf;
    size_t m_size;
};
////////////////////////////////////////////////////////////////////////////////
void buf_xyy_to_xyz(float *buf, size_t size)
{
    if(buf && size)
    {
        if(thread::ThreadPool::CanUsePool())
        {
            thread::Farmer farmer;

            size_t start, end;
            start = end = size;

            while(end)
            {
                if(end >= JOB_SIZE)
                {
                    start-= JOB_SIZE;
                }
                else
                {
                    start = 0;
                }
                farmer.Push(new xyYToXYZJob(&(buf[start*3]), (end-start)));
    
                end = start;
            }
            farmer.Process();
        }
        else
        {
            size_t i;
            float *p = buf;

            for(i=0;i<size;i++)
            {
                xyy_to_xyz(p);
                p+=3;
            }
        }
    }
}

void buf_xyz_to_xyy(float *buf, size_t size)
{
    if(buf && size)
    {
        if(thread::ThreadPool::CanUsePool())
        {
            thread::Farmer farmer;

            size_t start, end;
            start = end = size;

            while(end)
            {
                if(end >= JOB_SIZE)
                {
                    start-= JOB_SIZE;
                }
                else
                {
                    start = 0;
                }
                farmer.Push(new XYZToxyYJob(&(buf[start*3]), (end-start)));
    
                end = start;
            }
            farmer.Process();
        }
        else
        {
            size_t i;
            float *p = buf;

            for(i=0;i<size;i++)
            {
                xyz_to_xyy(p);
                p+=3;
            }
        }
    }
}

void buf_xyy_to_linear_srgb(float *buf, size_t size)
{
    if(buf && size)
    {
        if(thread::ThreadPool::CanUsePool())
        {
            thread::Farmer farmer;

            size_t start, end;
            start = end = size;

            while(end)
            {
                if(end >= JOB_SIZE)
                {
                    start-= JOB_SIZE;
                }
                else
                {
                    start = 0;
                }
                farmer.Push(new xyYToSRGBJob(&(buf[start*3]), (end-start)));
    
                end = start;
            }
            farmer.Process();
        }
        else
        {
            size_t i;
            float *p = buf;

            for(i=0;i<size;i++)
            {
                xyy_to_linear_srgb(p);
                p+=3;
            }
        }
    }
}

void buf_linear_srgb_to_xyy(float *buf, size_t size)
{
    if(buf && size)
    {
        if(thread::ThreadPool::CanUsePool())
        {
            thread::Farmer farmer;

            size_t start, end;
            start = end = size;

            while(end)
            {
                if(end >= JOB_SIZE)
                {
                    start-= JOB_SIZE;
                }
                else
                {
                    start = 0;
                }
                farmer.Push(new SRGBToxyYJob(&(buf[start*3]), (end-start)));
    
                end = start;
            }
            farmer.Process();
        }
        else
        {
            size_t i;
            float *p = buf;

            for(i=0;i<size;i++)
            {
                linear_srgb_to_xyy(p);
                p+=3;
            }
        }
    }
}

void buf_scale_xyy_y(float *buf, size_t size, float scale)
{
    if(buf && size)
    {
        if(thread::ThreadPool::CanUsePool())
        {
            thread::Farmer farmer;

            size_t start, end;
            start = end = size;

            while(end)
            {
                if(end >= JOB_SIZE)
                {
                    start-= JOB_SIZE;
                }
                else
                {
                    start = 0;
                }
                farmer.Push(new ScaleYJob(&(buf[start*3]), (end-start), scale));
    
                end = start;
            }
            farmer.Process();
        }
        else
        {
            size_t i;
            float *p = buf;

            for(i=0;i<size;i++)
            {
                p[2]*= scale;
                p+=3;
            }
        }
    }
}

void buf_scale(float *buf, size_t size, float scale)
{
    if(buf && size)
    {
        if(thread::ThreadPool::CanUsePool())
        {
            thread::Farmer farmer;

            size_t start, end;
            start = end = size;

            while(end)
            {
                if(end >= JOB_SIZE)
                {
                    start-= JOB_SIZE;
                }
                else
                {
                    start = 0;
                }
                farmer.Push(new ScaleJob(&(buf[start]), (end-start), scale));
    
                end = start;
            }
            farmer.Process();
        }
        else
        {
            size_t i;
            float *p = buf;

            for(i=0;i<size;i++)
            {
                (*p)*= scale;
                p++;
            }
        }
    }
}

void buf_luminance_to_density(float *buf, size_t size, float min) 
{
    if(buf && size)
    {
        if(thread::ThreadPool::CanUsePool())
        {
            thread::Farmer farmer;

            size_t start, end;
            start = end = size;

            while(end)
            {
                if(end >= JOB_SIZE)
                {
                    start-= JOB_SIZE;
                }
                else
                {
                    start = 0;
                }
                farmer.Push(new LuminanceToDensityJob(&(buf[start]), (end-start), min));
    
                end = start;
            }
            farmer.Process();
        }
        else
        {
            size_t i;
            float *p = buf;

            for(i=0;i<size;i++)
            {
                float tmp = *p;
                if(tmp < min) tmp = min;
                *p = log(tmp);
                p++;
            }
        }
    }
}

void buf_density_to_luminance(float *buf, size_t size)
{
    if(buf && size)
    {
        if(thread::ThreadPool::CanUsePool())
        {
            thread::Farmer farmer;

            size_t start, end;
            start = end = size;

            while(end)
            {
                if(end >= JOB_SIZE)
                {
                    start-= JOB_SIZE;
                }
                else
                {
                    start = 0;
                }
                farmer.Push(new DensityToLuminanceJob(&(buf[start]), (end-start)));
    
                end = start;
            }
            farmer.Process();
        }
        else
        {
            size_t i;
            float *p = buf;

            for(i=0;i<size;i++)
            {
                *p = exp(*p);
                p++;
            }
        }
    }
}

