
#include <iostream>
#include "image.h"
#include "controller.h"
#include "processor.h"
#include "parameters.h"
#include "utils.h"
#include "file.h"


namespace control {

NodeData::NodeData()
    : m_outputType(0)
    , m_parameters(NULL)
{
}

NodeData::NodeData(const NodeData &other)
    : m_outputType(0)
    , m_parameters(NULL)
{
    m_parameters = other.GetParameters();

    m_outputType = other.GetIO(m_mandatory, m_optional);

    m_inputs = other.GetInputs();
}

NodeData::NodeData(spectral::Processor *processor)
    : m_outputType(0)
    , m_parameters(NULL)
{
    if(processor)
    {
        m_outputType = processor->GetInterface(m_mandatory, m_optional);
        m_parameters = processor->GetParameters();
    }
    else
    {
        std::cerr << "NodeData constructor invalid" << std::endl;
    }
}

NodeData::~NodeData()
{
    if(m_parameters)
    {
        delete m_parameters;
        m_parameters = NULL;
    }
}

const NodeData &
NodeData::operator=(const NodeData &other)
{
    if(m_parameters)
    {
        delete m_parameters;
    }
    m_parameters = NULL;

    m_parameters = other.GetParameters();

    m_inputs.clear();
    m_mandatory.clear();
    m_optional.clear();

    m_outputType = other.GetIO(m_mandatory, m_optional);

    m_inputs = other.GetInputs();

    return *this;
}

uint32_t
NodeData::GetIO(std::list<std::pair<uint32_t, uint32_t> >&m,
                std::list<std::pair<uint32_t, uint32_t> >&o) const
{
    m.clear();
    o.clear();

    m = m_mandatory;
    o = m_optional;

    return m_outputType;
}

spectral::Parameters *
NodeData::GetParameters(void) const
{
    spectral::Parameters *result(NULL);
 
    if(m_parameters)
    {
        result = new spectral::Parameters(*m_parameters);
    }
    return result;
}

void 
NodeData::SetParameters(spectral::Parameters *p)
{
    if(p)
    {
        if(m_parameters)
        {
            delete m_parameters;
        }
        m_parameters = p;
    }
}

const std::map<uint32_t, int32_t>&
NodeData::GetInputs(void) const
{
    return m_inputs;
}

uint32_t
NodeData::GetInputType(uint32_t slot)
{
    uint32_t result = CLS_INVALID;
    std::list<std::pair<uint32_t, uint32_t> >::iterator i;

    i = m_mandatory.begin();
    while((i != m_mandatory.end()) && !result)
    {
        if(i->first == slot)
        {
            result = i->second;
        }
        i++;
    }

    i = m_optional.begin();
    while((i != m_optional.end()) && !result)
    {
        if(i->first == slot)
        {
            result = i->second;
        }

        i++;
    }
    return result;
}

int32_t 
NodeData::GetInputID(uint32_t slot)
{
    int32_t result = -1;

    std::map<uint32_t, int32_t>::const_iterator i =
        m_inputs.find(slot);

    if(i != m_inputs.end())
    {
        result = i->second;
    }
    return result;
}



void
NodeData::LinkInput(uint32_t slot, int32_t processor)
{
    bool add_link(false);

    std::list<std::pair<uint32_t, uint32_t> >::iterator i =
        m_mandatory.begin();
    while(i != m_mandatory.end())
    {
        if(i->first == slot)
            add_link = true;
        i++;
    }
    i = m_optional.begin();
    while(i != m_optional.end())
    {
        if(i->first == slot)
            add_link = true;
        i++;
    }
    if(add_link)
    {
        m_inputs[slot] = processor;
    }
}

void 
NodeData::UnlinkInput(int32_t slot)
{
    std::map<uint32_t, int32_t>::iterator i = m_inputs.find(slot);
    if(i != m_inputs.end())
    {
        m_inputs.erase(i);
    }
}

bool
NodeData::HasLink(int32_t processor)
{
    std::map<uint32_t, int32_t>::iterator i = m_inputs.begin();
    while(i != m_inputs.end())
    {
        if(i->second == processor)
            return true;
        i++;
    }
    return false;
}

////////////////////////////////////////////////////////////////////////////////

Controller::Controller(const std::string &modulePath,
                       const std::list<std::string> &modules)
    : m_modulePath(modulePath)
    , m_sequenceNumber(0)
{
    LoadModules(modules);

    std::cerr << "Modules loaded" << std::endl;

    // Reserve node 0 for loaded image.
    m_scheduler.AddNode(0);
    m_scheduler.SetComplete(0);
    m_actors[0] = NULL;

    Start();
}

Controller::~Controller()
{
    // Cleanup handled by Stop() function
    Stop();
}


////////////////////////////////////////////////////////////////////////////////
// External interface.
////////////////////////////////////////////////////////////////////////////////

void
Controller::LoadImage(const std::string &name)
{
    LoadImage(0, name);
}

void
Controller::LoadImage(int32_t id, const std::string &name)
{
    spectral::Image *img = spectral::SpLoadImage(m_moduleLoader, name);

    if(img)
    {
        m_queue.Push(new Message(Message::MSG_LOAD_IMAGE, id, img));
    }
}

void
Controller::SaveImage(int32_t id, const std::string &fn)
{
    m_queue.Push(new Message(Message::MSG_SAVE_IMAGE, id, fn));
}

void
Controller::Process(void)
{
    m_queue.Push(new Message(Message::MSG_START));
}

void
Controller::Shutdown(void)
{
    m_queue.Push(new Message(Message::MSG_SHUTDOWN));
    Join();
}

void
Controller::AddActor(int32_t id, std::string name)
{
    m_queue.Push(new Message(Message::MSG_SET_ACTOR, id, name));
}

void
Controller::RemoveActor(int32_t id)
{
    m_queue.Push(new Message(Message::MSG_SET_ACTOR, id, ""));
}

void
Controller::AddLink(int32_t id, uint32_t slot, int32_t input_id)
{
    std::pair<uint32_t, int32_t> link(slot, input_id);
    m_queue.Push(new Message(Message::MSG_SET_INPUT, id, link));
}

void 
Controller::RequestParameters(int32_t id)
{
    m_queue.Push(new Message(Message::MSG_REQUEST_PARAMETERS, id));
}

void 
Controller::RequestImage(int32_t id)
{
    m_queue.Push(new Message(Message::MSG_REQUEST_IMAGE, id));
}

void
Controller::SetParameters(int32_t id, spectral::Parameters *params)
{
    m_queue.Push(new Message(Message::MSG_PARAMETERS, id, params));
}

////////////////////////////////////////////////////////////////////////////////
// Event loops.
////////////////////////////////////////////////////////////////////////////////

void *
Controller::MainLoop(void)
{
    bool terminate(false);
    
    while(!terminate)
    {
        Message *msg;
        msg = m_queue.Pop();

        if(msg)
        {
            switch(msg->GetType())
            {
            case Message::MSG_START:
                // Begin processing.
                m_output.Push(new Message(Message::MSG_STARTED));
                ProcessImage();
                m_output.Push(new Message(Message::MSG_STOPPED));
                break;
            case Message::MSG_SHUTDOWN:
                // Terminate thread.
                terminate = true;
                m_output.Push(new Message(Message::MSG_EXIT));
                Stop();
                break;
            case Message::MSG_SET_ACTOR:
                // Install a processor in a given slot.
                {
                    int32_t id = msg->GetId();
                    std::string name = msg->GetString();
                    OnMsgSetActor(id, name);
                }
                break;
            case Message::MSG_SET_INPUT:
                // Link a slot on a processor to a given input.
                {
                    int32_t id = msg->GetId();
                    std::pair<uint32_t, int32_t> link = msg->GetLink();

                    OnMsgSetInput(id, link);
                }
                break;
            case Message::MSG_LOAD_IMAGE:
                // Load an image.
                {
                    int32_t id = msg->GetId();
                    spectral::Image *img = msg->GetImage();
   
                    OnMsgLoadImage(id, img); 
                }
                break;
            case Message::MSG_SAVE_IMAGE:
                // Save an image.
                {
                    int32_t id = msg->GetId();
                    std::string fn = msg->GetString();
   
                    OnMsgSaveImage(id, fn); 
                }
                break;
            case Message::MSG_REQUEST_PARAMETERS:
                {
                    int32_t id = msg->GetId();
                    spectral::Parameters *params(NULL);

                    std::map<int32_t, NodeData>::iterator i =
                        m_graph.find(id);

                    if(i != m_graph.end())
                    {
                        params = i->second.GetParameters();    
                    }

                    m_output.Push(new Message(Message::MSG_PARAMETERS,
                                              id,
                                              params)); 
                }
            case Message::MSG_REQUEST_IMAGE:
                {
                    int32_t id = msg->GetId();
                    spectral::Image *img(NULL);
                    std::map<int32_t, std::pair<uint32_t, spectral::Image*> >::iterator i =
                        m_images.find(id);

                    if(i != m_images.end())
                    {
                        if(i->second.second) 
                        {
                            img = new spectral::Image(*(i->second.second));
                        }   
                    }

                    m_output.Push(new Message(Message::MSG_SET_IMAGE,
                                              id,
                                              img)); 
                }


                break;
            case Message::MSG_PARAMETERS:
                {
                    int32_t id = msg->GetId();
                    spectral::Parameters *params = msg->GetParameters();

                    std::map<int32_t, NodeData>::iterator i =
                        m_graph.find(id);

                    m_scheduler.Invalidate(id);

                    if(i != m_graph.end())
                    {
                        if(params)
                        {
                            i->second.SetParameters(params);    
                        }
                    }
                }
                break;
            case Message::MSG_ERROR:
                // Report an error.
                {
                    m_output.Push(new Message(Message::MSG_ERROR,
                                              msg->GetId(),
                                              msg->GetString()));
                }
                break;
            default:
                // No default term required.
                break;
            }            
            delete msg;
        }
    }

    return NULL;
}

void
Controller::ProcessImage(void)
{
    bool terminate(false);

    terminate = !StartNextProcess();    

    while(!terminate)
    {
        Message *msg;
        msg = m_queue.Pop();

        if(msg)
        {
            switch(msg->GetType())
            {
            case Message::MSG_STOP:
                // Stop processing.
                terminate = true;
                break;
            case Message::MSG_SHUTDOWN:
                // Pass SHUTDOWN message to main loop.
                terminate = true;
                m_queue.Push(new Message(Message::MSG_SHUTDOWN));
                break;
            case Message::MSG_LOAD_IMAGE:
                {
                    spectral::Image *img = msg->GetImage();
                    // Can't load while running! 
                    if(img)
                    {
                        delete img;
                        m_queue.Push(new Message(Message::MSG_ERROR, 
                                     0,
                                     std::string("Can't load while processing!")));
                    }
                }
                break;
            case Message::MSG_RESULT:
                {
                    int32_t id = msg->GetId();
                    spectral::Image *img = msg->GetImage();

                    OnMsgResult(id, img, terminate);
                }
                break;
            case Message::MSG_PROGRESS:
                {
                    m_scheduler.SetProgress(msg->GetId(),
                                            msg->GetFloat());
                    m_output.Push(new Message(Message::MSG_PROGRESS,
                                              msg->GetId(),
                                              msg->GetFloat()));
                }
                break;
            case Message::MSG_ERROR:
                {
                    m_output.Push(new Message(Message::MSG_ERROR,
                                              msg->GetId(),
                                              msg->GetString()));
                    terminate = true;
                }
                break;
            default:
                // Ignore everything else.
                break;
            }            
            delete msg;
        }
    }

}

bool 
Controller::StartNextProcess(void)
{
    bool result(false);

    std::set<int> pending;
    m_scheduler.GetNextScheduledJobs(pending);

    if(pending.size())
    {
        int id = *(pending.begin());

        std::map<int32_t, ProcessThread*>::iterator i = m_actors.find(id);

        if(i != m_actors.end())
        {
            if(i->second)
            {
                spectral::Parameters *params(NULL);

                // We have a processor.  First see if it has any dependencies
                // that are newer than its current result.
                {
                    std::map<int32_t, NodeData>::iterator graph_it = 
                        m_graph.find(id);

                    if(graph_it != m_graph.end())
                    {
                        // Get parameters.
                        params = graph_it->second.GetParameters();

                        // If the input image is newer than the output, or
                        // the input is null we want to make sure we set this.
                        uint32_t cur_seq = _GetImageSeq(id);
                        std::map<uint32_t, int32_t>::const_iterator input_it =
                            graph_it->second.GetInputs().begin();

                        while(input_it != graph_it->second.GetInputs().end()) 
                        {
                            if((_GetImageSeq(input_it->second) > cur_seq) ||
                               (!_GetImage(input_it->second)))
                            {
                                // Update inputs.
                                i->second->SetImage(input_it->first,
                                                    _GetImage(input_it->second));
                            }
                            input_it++;
                        }
                    }
                }

                // Set parameters
                if(params)
                {
                    i->second->SetParameters(params);
                }
                // Start processing.
                i->second->Process();
                result = true;
            }
            else
            {
                // Processor is invalid.
                m_queue.Push(new Message(Message::MSG_ERROR,
                                         id, 
                                         std::string("NULL processor")));
            }
        }
        else
        {
            // Can't find process.
            m_queue.Push(new Message(Message::MSG_ERROR,
                                     id, 
                                     std::string("Unable to find processor")));
        }
    }
    return result;
}

////////////////////////////////////////////////////////////////////////////////
// Message handlers.

// Create a processor and load it into a slot.
void
Controller::OnMsgSetActor(int32_t id, std::string &name)
{
    //
    // Clean up
    //
 
    // First get rid of any images.
    _EraseImage(id);
    _RemoveActor(id);

    // Now see if we can load a suitable module.
    // Blank name means remove actor.

    if(name.length())
    {
        std::set<std::string> processors;
        
        {
            spectral::Processor *p(NULL);

            m_moduleLoader.GetProcessors(processors);
            if(processors.find(name) != processors.end())
            {
                p = (spectral::Processor*)
                    m_moduleLoader.CreateObject(name);
            }
            if(p)
            {
                _EraseImage(id);

                m_graph[id] = NodeData(p);

                m_actors[id] = new ProcessThread(id,
                                                 p,
                                                 m_queue);

                m_output.Push(new Message(Message::MSG_ACTOR_INSTALLED, id, name));

                std::cerr << "Added processor type " << name 
                          << " to actor " << id << std::endl;
            }
            else
            {
                m_output.Push(new Message(Message::MSG_ERROR,
                              id,
                              std::string("Bad name")));
            }
        }
    }
    _RebuildDependencies();
    m_scheduler.Invalidate(id);
}

void
Controller::OnMsgSetInput(int32_t id, std::pair<uint32_t, int32_t> link)
{
    uint32_t slot_id = link.first;
    uint32_t src_id = link.second;
 
    uint32_t input_type = CLS_INVALID;
    uint32_t slot_type = CLS_INVALID;

    // First establish what kind of data we have in our data
    // source.    
    if(src_id)
    {
        // We need to find an actor
        std::map<int32_t, NodeData>::iterator i =
            m_graph.find(src_id);

        if(i != m_graph.end())
        {
            std::list<std::pair<uint32_t, uint32_t> >t1, t2;
            input_type = i->second.GetIO(t1, t2);
        }
        else
        {
            std::cerr << "Unable to find actor " << id << std::endl;
        }
    }
    else
    {
        // link type is to source image.
        spectral::Image *img = _GetImage(src_id);

        if(img)
        {
            // Image is loaded.  Use whatever colour space.
            input_type = img->get_colour_space();
        }
        else
        {
            // Assume xyY.  We like xyY.  It's our friend.
            input_type = CLS_CIE_XYY;
        }
    }

    // Now establish what we have in our slot.
    if(id)
    {
        std::map<int32_t, NodeData>::iterator i =
            m_graph.find(id);
        if(i != m_graph.end())
        {
            slot_type = i->second.GetInputType(slot_id);
        }
        else
        {
            std::cerr << "Unable to find dest actor." << std::endl;
        }
    }

    // Now see if we can add the link.
    if((input_type != CLS_INVALID) &&
       (input_type == slot_type))
    {                
        // Currently just check for circular deps.
        if(m_scheduler.AddEdge(src_id, id))
        {
            std::map<int32_t, NodeData>::iterator i =
                m_graph.find(id);
            if(i != m_graph.end())
            {
                i->second.LinkInput(slot_id, src_id);
                std::cerr << "Linked " << id << ":" << slot_id << " <- " << src_id << std::endl;
                m_output.Push(new Message(Message::MSG_INPUT_ADDED,
                                          src_id,
                                          std::pair<int32_t, uint32_t>(slot_id, id)));
            }
        }
        //_RebuildDependencies(); // A belt & braces approach.
    }
    else
    {
        std::cerr << "slot type mismatch: src = " << input_type 
                  << ", dest = " << slot_type << std::endl;
    }
}

void
Controller::OnMsgLoadImage(int32_t id, spectral::Image *img)
{
    _RemoveActor(id);

    _SetImage(id, img);
    m_scheduler.SetComplete(id);

    if(img)
    {
        m_output.Push(new Message(Message::MSG_IMAGE_LOAD_SUCCEEDED, 
                                  id));
    }
    else
    {
        m_output.Push(new Message(Message::MSG_IMAGE_LOAD_FAILED, 
                                  id));
    }
}

void
Controller::OnMsgSaveImage(int32_t id, const std::string &fn)
{
    spectral::Image *img(NULL);

    std::map<int32_t, std::pair<uint32_t, spectral::Image*> >::iterator i =
        m_images.find(id);

    if(i != m_images.end())
    {
        img = i->second.second;
    }
    if(img)
    {
        spectral::SpSaveImage(m_moduleLoader, img, fn);
    }
}


void
Controller::OnMsgResult(int32_t id, spectral::Image *img, bool &terminate)
{
    if(id != 0)
    {
        _EraseImage(id);

        if(img)
        {
            {
                _SetImage(id, img);
                m_scheduler.SetComplete(id); 
            }
            terminate = !StartNextProcess();
        }
        else
        {
            // NULL result.  Report error and bail.
            m_queue.Push(new Message(Message::MSG_ERROR, 
                         id,
                         std::string("Process failed")));
        }
    } 
}

////////////////////////////////////////////////////////////////////////////////
// Internal utils.
////////////////////////////////////////////////////////////////////////////////

void
Controller::LoadModules(const std::list<std::string> &modules)
{
    std::list<std::string>::const_iterator i = modules.begin();

    while(i != modules.end())
    {
        m_moduleLoader.LoadModule(m_modulePath, *i);
        i++;
    }
}

void
Controller::_RebuildDependencies(void)
{
    m_scheduler.Clear();
    m_scheduler.AddNode(0);

    std::map<int32_t, NodeData>::iterator i;

    // First add all the nodes.
    i = m_graph.begin();
    while(i != m_graph.end())
    {
        m_scheduler.AddNode(i->first);
        i++;
    }

    // Now add edges.
    i = m_graph.begin();
    while(i != m_graph.end())
    {
        std::map<uint32_t, int32_t>::const_iterator j;

        j = i->second.GetInputs().begin();

        while(j != i->second.GetInputs().end())
        {
            m_scheduler.AddEdge(j->second, i->first);
            j++;
        }
        i++;
    }

    // Set everything but input to rebuild.
    m_scheduler.Invalidate(0);
    m_scheduler.SetComplete(0);
}

void
Controller::Stop(void)
{
    {
        std::map<int32_t, ProcessThread*>::iterator i = m_actors.begin();
 
        while(i != m_actors.end())
        {
            if(i->second)
            {
                std::cerr << "stopping actor " << i->first << std::endl;
                i->second->Shutdown();
                delete i->second;
            }
            i++;
        }
        m_actors.clear();
    }

    {
        std::map<int32_t, std::pair<uint32_t, spectral::Image*> >::iterator i =
            m_images.begin();

        while(i != m_images.end())
        {
            if(i->second.second)
            {
                delete i->second.second;
            }
            i++;
        }
        m_images.clear();
    }

    m_graph.clear();
}

uint32_t 
Controller::_SetImage(int32_t id, spectral::Image *img)
{
    uint32_t result;

    _EraseImage(id);
    
    result = _GetSeqNumber();

    m_images[id] = std::pair<uint32_t, spectral::Image*>(result, img);

    return result;
}

spectral::Image *
Controller::_GetImage(int32_t id)
{
    spectral::Image *result(NULL);

    std::map<int32_t, std::pair<uint32_t, spectral::Image*> >::const_iterator i =
        m_images.find(id);

    if(i != m_images.end())
    {
        result = i->second.second;
    }

    return result;
}

uint32_t
Controller::_GetImageSeq(int32_t id)
{
    uint32_t result(0);

    std::map<int32_t, std::pair<uint32_t, spectral::Image*> >::const_iterator i =
        m_images.find(id);

    if(i != m_images.end())
    {
        result = i->second.first;
    }

    return result;
}

void
Controller::_EraseImage(int id)
{
    std::map<int32_t, std::pair<uint32_t, spectral::Image*> >::iterator i =
        m_images.find(id);
    if(i != m_images.end())
    {
        if(i->second.second)
        {
            delete i->second.second;
            m_images[id] = std::pair<uint32_t, spectral::Image*>
                (0, (spectral::Image *)NULL);
        }
    }
    m_scheduler.Invalidate(id);
}

uint32_t
Controller::_GetSeqNumber(void)
{
    m_sequenceNumber++;
    return m_sequenceNumber;
}

void
Controller::_RemoveActor(int32_t id)
{
    bool removed(false);
 
    // First get rid of any images.
    _EraseImage(id);

    // Next get rid of any actor threads
    {
        std::map<int32_t, ProcessThread*>::iterator i =
            m_actors.find(id);

        if(i != m_actors.end())
        {
            if(i->second)
            {
                i->second->Shutdown();
                delete i->second;
            }
            
            removed = true;
            m_actors.erase(i);
        }
    }

    // Now get rid of any interface specs.
    {
        std::map<int32_t, NodeData>::iterator i =
            m_graph.find(id);

        if(i != m_graph.end())
        {
            // Report removal of links from this interface.
            std::map<uint32_t, int32_t>::const_iterator it = i->second.GetInputs().begin();

            while(it != i->second.GetInputs().end())
            {
                m_output.Push(new Message(Message::MSG_INPUT_REMOVED, 
                                          it->second,
                                          std::pair<uint32_t, int32_t>(it->first, id)));
                it++;
            }

            // Remove interface.
            m_graph.erase(i);
        }
    }

    // Now get rid of any inputs pointing to this actor.
    {
        std::map<int32_t, NodeData>::iterator i =
            m_graph.begin();

        while(i != m_graph.end())
        {
            if(i->second.HasLink(id))
            {
                // Report unlink.
                std::map<uint32_t, int32_t>::const_iterator it(i->second.GetInputs().begin());
                while(it != i->second.GetInputs().end())
                {
                    if(it->second == id)
                    {
                        m_output.Push(new Message(Message::MSG_INPUT_REMOVED, 
                                                  id, 
                                                  std::pair<uint32_t, int32_t>(it->first, i->first)));
                        
                    }
                    it++;
                }
                
                // do unlink.
                i->second.UnlinkInput(id);
            }
            i++;
        } 
    }

    if(removed)
    {
        m_output.Push(new Message(Message::MSG_ACTOR_REMOVED, id));
    }
}
}
