#include <math.h>
#include <string.h>
#include "image.h"
#include "utils.h"


int32_t constrain_reflect(int32_t x, uint32_t extent)
{
    extent--;

    if(extent)
    {
        x = abs(x);

        while(uint32_t(x) > extent)
        {
            x = abs((2 * (int32_t)extent) - x);
        }
    }
    return x;
}

int32_t constrain_wrap(int32_t x, uint32_t extent) 
{
    if(extent)
    {
        while(x < 0)
        {
            x+= extent;
        }
        while(x >= (int32_t)extent)
        {
            x-= extent;
        }
    }
    return x;
}


namespace spectral
{

Image::Image(uint32_t width,
             uint32_t height,
             uint32_t cs,
             float *buffer)
    : image<float>(width, height, get_pixel_size(cs), buffer)
    , m_cs(cs)
{
}

Image::Image(const Image &other)
    : image<float>(other.get_width(), 
                   other.get_height(), 
                   get_pixel_size(other.get_colour_space()),
                   other.get_buffer())
    , m_cs(other.get_colour_space())
{
}

Image::~Image()
{
}

// Create an srgb thumbnail.
Image *
Image::CreateThumbnail(uint32_t width, uint32_t height) const
{
    Image *result(NULL);

    if(!width)
        width = 1;
    if(!height)
        height = 1;

    result = new Image(width, height, CLS_SRGB_LINEAR);

    uint32_t x, y;
    float x_mul, y_mul;
    float max(0);

    x_mul = get_width();
    x_mul/= width;
    y_mul = get_height();
    y_mul/= height;

    for(y=0;y<height;y++)
    {
        uint32_t yy;
        float fy;
        fy = y;
        fy*= y_mul;
        yy = floor(fy);
        for(x=0;x<width;x++)
        {
            uint32_t xx;
            float fx;
            const float *from_pixel;
            float to_pixel[3];

            fx = x;
            fx*= x_mul;
            xx = floor(fx);

            from_pixel = get_pixel(xx, yy);
            if(from_pixel)
            {
                switch(get_colour_space())
                {
                case CLS_SRGB_LINEAR:
                case CLS_SRGB: // TODO: Fix.
                    to_pixel[0]=from_pixel[0];
                    to_pixel[1]=from_pixel[1];
                    to_pixel[2]=from_pixel[2];
                    break;
                case CLS_CIE_XYY:
                    to_pixel[0]=from_pixel[0];
                    to_pixel[1]=from_pixel[1];
                    to_pixel[2]=from_pixel[2];
                    xyy_to_linear_srgb(to_pixel);
                    break;
                case CLS_CIE_XYZ:
                    to_pixel[0]=from_pixel[0];
                    to_pixel[1]=from_pixel[1];
                    to_pixel[2]=from_pixel[2];
                    xyz_to_linear_srgb(to_pixel);
                    break;
                case CLS_LUMINANCE:
                    to_pixel[0]=from_pixel[0];
                    to_pixel[1]=from_pixel[0];
                    to_pixel[2]=from_pixel[0];
                    break;
                case CLS_DENSITY:
                    to_pixel[0]=exp(from_pixel[0]);
                    to_pixel[1]=to_pixel[0];
                    to_pixel[2]=to_pixel[0];
                    break;
                default:
                    to_pixel[0]=to_pixel[1]=to_pixel[2]=0.0;
                }
            }
            else
            {
                to_pixel[0]=to_pixel[1]=to_pixel[2]=0.0;
            }
            for(size_t i=0;i<3;i++)
            {
                if(to_pixel[i] > max)
                {
                    max = to_pixel[i];
                }
            }
            result->set_pixel(x, y, to_pixel);
        }
    }
    if(max > 1.0e-5)
    {
        size_t size = size_t(width) * size_t(height) * 3;
        float *buf = result->get_buffer();

        for(uint32_t i=0;i<size;i++)
        {
            buf[i]/= max;
        }
    }
    return result;
}

uint32_t
Image::get_colour_space(void) const
{
    return m_cs;
}

// Create a tile from an image, reflecting or wrapping if necessary.
Image *
Image::CreateTile(int32_t x1, int32_t y1,
                  uint32_t width, uint32_t height,
                  bool wrap_x, bool wrap_y) const
{
    Image *result(NULL);

    result = new Image(width, height, get_colour_space());

    if(result)
    {
        for (uint32_t y=0;uint32_t(y)<height;y++)
        {
            int32_t yy;
            yy = int32_t(y) + int32_t(y1);
            for(int32_t x=0;uint32_t(x)<width;x++)
            {
                int32_t xx;
                xx = int32_t(x) + int32_t(x1);

                result->set_pixel(x,y,get_constrained_pixel(xx,yy,
                                                            wrap_x,wrap_y));

            }
        }
    }
    return result;
}
 
// Expand an image, adding a reflected border.
Image *
Image::Expand(uint32_t x_border, uint32_t y_border,
              bool wrap_x, bool wrap_y) const
{
    Image *result(NULL);    

    uint32_t new_width  = get_width() + (x_border * 2);
    uint32_t new_height = get_height()+ (y_border * 2);
    result = new Image(new_width, new_height, get_colour_space());

    if(result)
    {
        for (int32_t y=0;uint32_t(y)<new_height;y++)
        {
            int32_t yy;
            yy = int32_t(y) - int32_t(y_border);
            for(int32_t x=0;uint32_t(x)<new_width;x++)
            {
                int32_t xx;
                xx = int32_t(x) - int32_t(x_border);
                result->set_pixel(x,y,get_constrained_pixel(xx,yy,
                                                            wrap_x,wrap_y));
            }
        }
    }

    return result;
}

Image *
Image::Contract(uint32_t x_border, uint32_t y_border) const
{
    Image *result(NULL);    

    if((get_height() > (y_border * 2) &&
       (get_width() > (x_border * 2))))
    {
        uint32_t new_width  = get_width() - (x_border * 2);
        uint32_t new_height = get_height()- (y_border * 2);
        result = new Image(new_width, new_height, get_colour_space());
   
        if(result)
        {
            for (int32_t y=0;uint32_t(y)<new_height;y++)
            {
                int32_t yy;
                yy = y + y_border;
                for(int32_t x=0;uint32_t(x)<new_width;x++)
                {
                    int32_t xx;
                    xx = x + x_border;
                    result->set_pixel(x,y,get_constrained_pixel(xx,yy));
                }
            }
        }
    }
    return result;
}

////////////////////////////////////////////////////////////////////////////////

IntegralHistogram::IntegralHistogram(uint32_t bins, const Image &img,
                                     uint32_t(*pixel_to_bin)(uint32_t,
                                                             uint32_t,
                                                             const Image *,
                                                             void *),
                                     void *data)

    : image<uint32_t>(img.get_width(), img.get_height(), bins)
{
    BuildHistogram(bins, img, 0, 0, img.get_width(), img.get_height(),
                   pixel_to_bin, data);
}

IntegralHistogram::IntegralHistogram(uint32_t bins, const Image &img,
                                     uint32_t x0, uint32_t y0,
                                     uint32_t width, uint32_t height,
                                     uint32_t(*pixel_to_bin)(uint32_t,
                                                             uint32_t,
                                                             const Image *,
                                                             void *),
                                     void *data)
    : image<uint32_t>(width, height, bins)
{
    BuildHistogram(bins, img, x0, y0, width, height, pixel_to_bin, data);
}

IntegralHistogram::~IntegralHistogram()
{
}

void
IntegralHistogram::BuildHistogram(uint32_t bins, const Image &img,
                                  uint32_t x0, uint32_t y0,
                                  uint32_t width, uint32_t height,
                                  uint32_t(*pixel_to_bin)(uint32_t,
                                                          uint32_t,
                                                          const Image *,
                                                          void *),
                                  void *data)
{
    uint32_t x,y;
    size_t index(0);

    memset(get_buffer(), 0, 
           get_width() * get_height() * bins * sizeof(uint32_t));

    for(y=0;y<get_height();y++)
    {
        for(x=0;x<get_width();x++)
        {
            uint32_t bin;

            // Add 1 to this bin if we are in range.
            if((x < width) && (y < height))
            {
                bin = pixel_to_bin(x + x0, y + y0, &img, data);

                if(bin < bins)
                {
                    get_buffer()[index + size_t(bin)] = 1;
                }
            }

            if(x)
            {
                // Add previous pixel in x
                const uint32_t *pixel;
                uint32_t i;
                pixel = get_pixel(x-1, y);
                for(i=0;i<bins;i++)
                {
                    get_buffer()[index + i]+= pixel[i];
                }
            }
            if(y)
            {
                // Add previous pixel in y
                const uint32_t *pixel;
                uint32_t i;

                pixel = get_pixel(x, y-1);
                for(i=0;i<bins;i++)
                {
                    get_buffer()[index + i]+= pixel[i];
                }
            }
            if(x && y)
            {
                // subtract previous pixel in x,y
                const uint32_t *pixel;
                uint32_t i;

                pixel = get_pixel(x-1, y-1);
                for(i=0;i<bins;i++)
                {
                    get_buffer()[index + i]-= pixel[i];
                }
            }

            index+= bins;
        }
    }
}

void 
IntegralHistogram::GetHistogram(uint32_t x1, uint32_t y1,
                                uint32_t x2, uint32_t y2,
                                uint32_t *result) const
{
    uint32_t i;

    if(x1 > x2)
    {
        uint32_t tmp = x1;
        x1 = x2;
        x2 = tmp;
    }
    if(y1 > y2)
    {
        uint32_t tmp = y1;
        y1 = y2;
        y2 = tmp;
    }

    for(i=0;i<get_channels();i++)
    {
        result[i] = 0;
    }

    if((x1 < get_width()) && (y1 < get_height()))
    {
        const uint32_t *a, *b, *c, *d;
        //
        //  a-----b
        //  |     |
        //  d-----c
        //
        // We want to get inclusive points, so we move the top left up, 
        // potentially making a, b, c invalid in the process.

        a = b = c = d = NULL;

        if(x2 >= get_width())
        {
            x2 = get_width() - 1;
        }
        if(y2 >= get_height())
        {
            y2 = get_height() - 1;
        }

        c = get_pixel(x2, y2);

        if(x1 && y1)
        {
            a = get_pixel(x1 - 1, y1 - 1);
        }
        if(x1)
        {
            d = get_pixel(x1 - 1, y2);
        }
        if(y1)
        {
            b = get_pixel(x2, y1 - 1);
        }

        // calculate histogram.
        // TODO: optimise
        for(i=0;i<get_channels();i++)
        {
            if(a) result[i]+=a[i];
            if(c) result[i]+=c[i];
            if(b) result[i]-=b[i];
            if(d) result[i]-=d[i];
        }
    }
}

}
