#ifndef _WIN32
#include <dlfcn.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <iostream>

#include "plugin.h"
#include "modules.h"

// function pointer def
#if !defined(_WIN32)
typedef void(*FUNCTION)(void);
#endif

namespace spectral {

// Strings for building library name.

#if defined(OS_MAC)
const std::string LIB_EXTENSION(".dylib");
const std::string LIB_PRELUDE("lib");
#elif defined(_WIN32)
const std::string LIB_EXTENSION(".dll");
const std::string LIB_PRELUDE("");
#else
const std::string LIB_EXTENSION(".so");
const std::string LIB_PRELUDE("lib");
#endif



Modules::Modules()
{
}

Modules::~Modules()
{
    // unload modules.
    std::map<std::string, ModuleData>::iterator i = m_modules.begin();

    while(i != m_modules.end())
    {
#ifdef _WIN32
        FreeLibrary((i->second).m_module);
#else
        dlclose((i->second).m_module);
#endif
        i++;
    }
    m_modules.clear();
    m_fileLoaders.clear();
    m_fileSavers.clear();
    m_processors.clear();
}

bool
Modules::LoadModule(std::string path, std::string name)
{
    bool result(false);

    if(m_modules.find(name) == m_modules.end())
    {
        MODULE_HANDLE library(0);
        std::string libname(path);

#ifdef _WIN32
        libname+= "\\";
#else
        libname+= "/";
#endif
        libname+= LIB_PRELUDE;
        libname+= name;
        libname+= LIB_EXTENSION; 

        std::cerr << "attempting to load library " << libname << std::endl;

#ifdef _WIN32
        library = LoadLibrary(libname.c_str());
#else
        library = dlopen(libname.c_str(), RTLD_LAZY);
#endif

        if(library)
        {
            // Get entry points.
#ifdef _WIN32
            FARPROC v_get_type  = GetProcAddress(library,
                                                 STM_PLUGIN_GET_TYPE_STRING);
            FARPROC v_get_string = GetProcAddress(library,
                                                  STM_PLUGIN_GET_STRING_STRING);
            FARPROC v_get_object = GetProcAddress(library,
                                                STM_PLUGIN_GET_OBJECT_STRING);
#else
            FUNCTION v_get_type   = ((FUNCTION(*)(...))dlsym)(library, STM_PLUGIN_GET_TYPE_STRING);
            FUNCTION v_get_string = ((FUNCTION(*)(...))dlsym)(library, STM_PLUGIN_GET_STRING_STRING);
            FUNCTION v_get_object = ((FUNCTION(*)(...))dlsym)(library, STM_PLUGIN_GET_OBJECT_STRING);
#endif

            if(v_get_type && v_get_string && v_get_object)
            {
                int(*get_type)(void) = (int(*)(void))v_get_type;
                char*(*get_string)(int) = (char*(*)(int))v_get_string;
                void*(*get_object)(void) = (void*(*)(void))v_get_object;

                result = true;

                {
                    // Populate module data field.
                    char *n = get_string(STM_STRING_NAME);

                    ModuleData data;
                    data.m_module = library;
                    data.m_createFunction = get_object;
                    if(n)
                    {
                        data.m_name = n;
                        free(n);
                    }

                    m_modules[name] = data;
                }

                // Classify module type.
                switch(get_type())
                {
                case STM_TYPE_LOADER:
                    {
                        char *n = get_string(STM_STRING_FILE_EXT);
                        if(n && strlen(n))
                        {
                            char *ext, *nxt;
                            ext = n;
                            nxt = ext;
                            while(ext)
                            {
                                NextExt(&nxt);
                                m_fileLoaders[std::string(ext)] = name;
                                std::cerr << "Loader "
                                          << libname
                                          << " handles \""
                                          << ext
                                          << "\""
                                          << std::endl;
                                ext = nxt;
                            }
                            free(n);
                        }
                    }
                    break;
                case STM_TYPE_SAVER:
                    {
                        char *n = get_string(STM_STRING_FILE_EXT);
                        if(n && strlen(n))
                        {
                            char *ext, *nxt;
                            ext = n;
                            nxt = ext;
                            while(ext)
                            {
                                NextExt(&nxt);
                                m_fileSavers[std::string(ext)] = name;
                                std::cerr << "Saver "
                                          << libname
                                          << " handles \""
                                          << ext
                                          << "\""
                                          << std::endl;
                                ext = nxt;
                            }
                            free(n);
                        }
                    }
                    break;
                case STM_TYPE_PROCESS:
                    {
                        m_processors.insert(name);
                        std::cerr << "Loaded Processor \""
                                  << name
                                  << "\""
                                  << std::endl;
                    }
                }
            }
            else
            {
                fputs("failed to get all function pointers:", stderr);
                if(!v_get_type) fputs("v_get_type, ", stderr);
                if(!v_get_string) fputs("v_get_string, ", stderr);
                if(!v_get_object) fputs("v_get_object, ", stderr);
                fputs("\n", stderr);
#ifdef _WIN32
                FreeLibrary(library);
#else
                dlclose(library);
#endif
            }
        }
        else
        {
#ifdef _WIN32
            fprintf(stderr, "Failed to open library%s\"\n", libname.c_str());
#else
            fprintf(stderr, "dlopen failed with \"%s\"\n", dlerror());
#endif
        }
    }
    else
    {
        result = false;
    }
    return result;
}

void *
Modules::CreateObject(const std::string name) const
{
    void *result(NULL);

    std::map<std::string, ModuleData>::const_iterator i =
        m_modules.find(name);

    if(i != m_modules.end())
    {
        if(i->second.m_createFunction)
        {
            result = (i->second.m_createFunction)();
        }
    }

    return result;
}

bool
Modules::GetModuleDesc(const std::string &name, std::string &desc) const
{
    bool result(false);

    std::map<std::string, ModuleData>::const_iterator i =
        m_modules.find(name);

    if(i != m_modules.end())
    {
        desc = i->second.m_name;
        result = true;
    }

    return result;
}

void
Modules::GetLoaders(std::map<std::string, std::string> &result) const
{
    result.clear();
    result.insert(m_fileLoaders.begin(), m_fileLoaders.end());
}

void
Modules::GetSavers(std::map<std::string, std::string> &result) const
{
    result.clear();
    result.insert(m_fileSavers.begin(), m_fileSavers.end());
}

void
Modules::GetProcessors(std::set<std::string> &result) const
{
    result.clear();
    result.insert(m_processors.begin(), m_processors.end());
}

void
Modules::NextExt(char **stringp)
{
#ifdef _WIN32
    if(stringp)
    {
        char *p(*stringp);
        if(p)
        {
            while(*p && (*p != ';'))
            {
                p++;
            }

            if(*p)
            {
                *p = '\0';
                p++;
                *stringp = p;
            }
            else
            {
                *stringp = NULL;
            }
        }
    }
    
#else
    strsep(stringp, ";");
#endif
}

}
