#include "parameters.h"

namespace spectral
{

////////////////////////////////////////////////////////////////////////////////
// FloatParameter class
//
void
FloatParameter ::GetValue(float &value) const
{
    value = m_value;
}

void
FloatParameter::GetRange(float &min, float &max) const
{
    min = m_min;
    max = m_max;
}

bool
FloatParameter::SetValue(float value)
{
    bool result(false);

    if((value >= m_min) && (value <= m_max))
    {
        m_value = value;
        result = true;
    }
    return result;
}

////////////////////////////////////////////////////////////////////////////////
// IntParameter class
//
void
IntParameter::GetValue(int32_t &value) const
{
    value = m_value;
}

void
IntParameter::GetRange(int32_t &min, int32_t &max) const
{
    min = m_min;
    max = m_max;
}

bool
IntParameter::SetValue(int32_t value)
{
    bool result(false);

    if((value >= m_min) && (value <= m_max))
    {
        m_value = value;
        result = true;
    }
    return result;
}


EnumParameter::EnumParameter()
    : m_value(0)
{
}

EnumParameter::EnumParameter(const EnumParameter &other)
{
    std::list<std::pair<uint32_t, std::string> > range;

    other.GetRange(range);
    SetRange(range);

    {
        std::string tmp;
        other.GetValue(m_value, tmp);
    }

}

EnumParameter::EnumParameter(const std::list<std::pair<uint32_t,
                             std::string> > &range, uint32_t value)
    : m_value(value)
{
    SetRange(range);
}

void
EnumParameter::SetRange(
    const std::list<std::pair<uint32_t, std::string> > &range)
{
    std::list<std::pair<uint32_t, std::string> >::const_iterator i =
        range.begin();

    m_range.clear();

    while(i != range.end())
    {
        m_range[i->first] = i->second;

        i++;
    }
} 

void
EnumParameter::GetRange(std::list<std::pair<uint32_t, std::string> > &out) const
{
    std::map<uint32_t, std::string>::const_iterator i =
        m_range.begin();

    out.clear();

    while(i != m_range.end())
    {
        out.push_back(std::pair<uint32_t, std::string>(i->first, i->second));
        i++;
    }
}

void
EnumParameter::GetValue(uint32_t &value, std::string &name) const
{
    value = m_value;
    name = "";
    std::map<uint32_t, std::string>::const_iterator i = m_range.find(m_value);

    if(i != m_range.end())
    {
        name = i->second;
    }
}

bool
EnumParameter::SetValue(uint32_t value)
{
    bool result(false);

    if(m_range.find(value) != m_range.end())
    {
        m_value = value;
        result = true;
    }

    return result;
}        


////////////////////////////////////////////////////////////////////////////////
// Parameters class.
//

Parameters::Parameters()
{
}

Parameters::~Parameters()
{
}

uint32_t
Parameters::GetParameterType(std::string name) const
{
    uint32_t result(PARAM_TYPE_NONE);

    std::map<std::string, uint32_t>::const_iterator i =
        m_allParams.find(name);

    if(i != m_allParams.end())
    {
        result = i->second;
    }
    return result;
}

// Generate a sorted list of parameter names.
void
Parameters::GetParameterNames(std::list<std::string> &names) const
{
    // Initial implementation just dumps all names in alphabetical order.

    std::list<std::string>::const_iterator i = m_names.begin();

    while(i != m_names.end())
    {
        names.push_back(*i);
        i++;
    }
}
////////////////////////////////////////////////////////////////////////////////
// Boolean parameters.
bool
Parameters::AddParameter(std::string name, bool value)
{
    bool result(false);

    if(m_allParams.find(name) == m_allParams.end())
    {
        // Parameter is not defined.
        result = true;

        m_allParams[name] = PARAM_TYPE_BOOL;
        m_boolParams[name] = value;
        m_names.push_back(name);
    }

    return result;
}

bool
Parameters::SetParameter(std::string name, bool value)
{
    bool result(false);
    
    std::map<std::string, bool>::iterator i =
        m_boolParams.find(name);

    if(i != m_boolParams.end())
    {
        SetDirty();
        m_boolParams[name] = value;
    }

    return result;
}

bool
Parameters::GetParameter(std::string name, bool &value) const
{
    bool result(false);
    
    std::map<std::string, bool>::const_iterator i =
        m_boolParams.find(name);

    if(i != m_boolParams.end())
    {
        value = i->second;
    }

    return result;
}

////////////////////////////////////////////////////////////////////////////////
// Floating point parameters.
bool 
Parameters::AddParameter(std::string name, float min, float max, float value)
{
    bool result(false);

    if(m_allParams.find(name) == m_allParams.end())
    {
        // Parameter is not defined.
        result = true;

        m_allParams[name] = PARAM_TYPE_FLOAT;
        m_floatParams[name] = FloatParameter(min, max, value);
        m_names.push_back(name);
    }

    return result;
}

bool
Parameters::GetParameter(std::string name, float &value) const
{
    bool result(false);

    std::map<std::string, FloatParameter>::const_iterator i =
        m_floatParams.find(name);

    if(i != m_floatParams.end())
    {
        result = true;
        i->second.GetValue(value);
    }

    return result;
}

bool
Parameters::GetRange(std::string name, float &min, float &max) const
{
    bool result(false);

    std::map<std::string, FloatParameter>::const_iterator i =
        m_floatParams.find(name);

    if(i != m_floatParams.end())
    {
        result = true;
        i->second.GetRange(min, max);
    }

    return result;
}

bool 
Parameters::SetParameter(std::string name, float value)
{
    bool result(false);
    
    std::map<std::string, FloatParameter>::iterator i =
        m_floatParams.find(name);

    if(i != m_floatParams.end())
    {
        SetDirty();
        result = i->second.SetValue(value);
    }

    return result;
}

////////////////////////////////////////////////////////////////////////////////
// Integer parameters.
bool 
Parameters::AddParameter(std::string name,
                         int32_t min,
                         int32_t max,
                         int32_t value)
{
    bool result(false);

    if(m_allParams.find(name) == m_allParams.end())
    {
        // Parameter is not defined.
        result = true;

        m_allParams[name] = PARAM_TYPE_INT;
        m_intParams[name] = IntParameter(min, max, value);
        m_names.push_back(name);
    }

    return result;
}

bool
Parameters::GetParameter(std::string name, int32_t &value) const
{
    bool result(false);

    std::map<std::string, IntParameter>::const_iterator i =
        m_intParams.find(name);

    if(i != m_intParams.end())
    {
        result = true;
        i->second.GetValue(value);
    }

    return result;
}

bool
Parameters::GetRange(std::string name, int32_t &min, int32_t &max) const
{
    bool result(false);

    std::map<std::string, IntParameter>::const_iterator i =
        m_intParams.find(name);

    if(i != m_intParams.end())
    {
        result = true;
        i->second.GetRange(min, max);
    }

    return result;
}

bool 
Parameters::SetParameter(std::string name, int32_t value)
{
    bool result(false);
    { 
        std::map<std::string, IntParameter>::iterator i =
            m_intParams.find(name);

        if(i != m_intParams.end())
        {
            SetDirty();
            result = i->second.SetValue(value);
        }
    }
    if(!result)
    {
        // Try to set as an enum.
        if(value >= 0)
        {
            std::map<std::string, EnumParameter>::iterator i =
                m_enumParams.find(name);
    
            if(i != m_enumParams.end())
            {
                SetDirty();
                result = i->second.SetValue(value);
            }
        }
    }
    if(!result)
    {
        // Try to set as a float.
        std::map<std::string, FloatParameter>::iterator i =
            m_floatParams.find(name);
   
        if(i != m_floatParams.end())
        {
            SetDirty();
            float fval = value;
            result = i->second.SetValue(fval);
        }
    }

    return result;
}

////////////////////////////////////////////////////////////////////////////////
// Enum parameters.
bool
Parameters::AddParameter(
    std::string name,
    const std::list<std::pair<uint32_t, std::string> > &range,
    uint32_t value)
{
    bool result(false);

    if(m_allParams.find(name) == m_allParams.end())
    {
        // Parameter is not defined.
        result = true;

        m_allParams[name] = PARAM_TYPE_ENUM;

        m_enumParams[name] = 
            EnumParameter(range, value);

        m_names.push_back(name);
    }

    return result;

}

bool
Parameters::SetParameter(std::string name, uint32_t value)
{
    bool result(false);

    std::map<std::string, EnumParameter>::iterator i =
        m_enumParams.find(name);

    if(i != m_enumParams.end())
    {
        SetDirty();
        result = i->second.SetValue(value);
    }

    return result;
}

bool
Parameters::GetParameter(std::string name, uint32_t &value, std::string &str) const
{
    bool result(false);

    std::map<std::string, EnumParameter>::const_iterator i =
        m_enumParams.find(name);

    if(i != m_enumParams.end())
    {
        result = true;
        i->second.GetValue(value, str);
    }

    return result;
}

bool
Parameters::GetRange(
    std::string name,
    std::list<std::pair<uint32_t, std::string> > &range) const
{
    bool result(false);

    std::map<std::string, EnumParameter>::const_iterator i =
        m_enumParams.find(name);

    if(i != m_enumParams.end())
    {
        result = true;
        i->second.GetRange(range);
    }

    return result;
}

////////////////////////////////////////////////////////////////////////////////
// White point parameters.
bool
Parameters::AddParameter(std::string name, Chromaticity value)
{
    bool result(false);

    if(m_allParams.find(name) == m_allParams.end())
    {
        // Parameter is not defined.
        result = true;

        m_allParams[name] = PARAM_TYPE_CHROMATICITY;
        m_chromaParams[name] = value;
        m_names.push_back(name);
    }

    return result;
}

bool
Parameters::SetParameter(std::string name, Chromaticity value)
{
    bool result(false);
    
    std::map<std::string, Chromaticity>::iterator i =
        m_chromaParams.find(name);

    if(i != m_chromaParams.end())
    {
        SetDirty();
        m_chromaParams.erase(i);
        m_chromaParams[name] = value;
    }

    return result;
}

bool
Parameters::GetParameter(std::string name, Chromaticity &value) const
{
    bool result(false);

    std::map<std::string, Chromaticity>::const_iterator i =
        m_chromaParams.find(name);

    if(i != m_chromaParams.end())
    {
        result = true;
        value = i->second;
    }

    return result;
}

}
