#include "image.h"
#include "processthread.h"
#include "parameters.h"
#include "processor.h"


namespace control {

////////////////////////////////////////////////////////////////////////////////
// ProcessThread class.
// Wraps a processor object, and allows it to perform image processing tasks in
//  its own thread.
//
ProcessThread::ProcessThread(int id,
                             spectral::Processor *processor,
                             MessageQueue &output)
    : m_outputQueue(output)
    , m_processor(processor)
    , m_active(false)
    , m_id(id)
{
    if(m_processor)
    {
        m_processor->SetThread(this);
    }
    Start();
} 

ProcessThread::~ProcessThread()
{
    if(m_processor)
    {
        delete m_processor;
    }
}

void
ProcessThread::Process(void)
{
    m_inputQueue.Push(new Message(Message::MSG_START)); 
}

void
ProcessThread::Shutdown(void)
{
    m_inputQueue.Push(new Message(Message::MSG_SHUTDOWN)); 
    Join();
}

void
ProcessThread::SetImage(int slot, spectral::Image *image)
{
    m_inputQueue.Push(new Message(Message::MSG_SET_IMAGE, slot, image)); 
}

void
ProcessThread::SetParameters(spectral::Parameters *params)
{
    m_inputQueue.Push(new Message(Message::MSG_PARAMETERS, 0, params));
}

void
ProcessThread::Progress(float progress)
{
    Message *msg = new Message(Message::MSG_PROGRESS, m_id, progress);

    if(msg)
    {
        m_outputQueue.Push(msg);
    }
}

void *
ProcessThread::MainLoop(void)
{
    bool terminate(false);

    while(!terminate)
    {
        Message *msg;
        msg = m_inputQueue.Pop();
        if(msg)
        {
            switch(msg->GetType())
            { 
            case Message::MSG_START:
                {
                    // Process an image.
                    spectral::Image *img(NULL);

                    if(m_processor)
                    {
                        {
                            Lock lock(this);
                            m_active = true;
                        }
                        img = m_processor->Process();
                        {
                            Lock lock(this);
                            m_active = false;
                        }
                    }

                    Message *out_msg = new Message(Message::MSG_RESULT,
                                                   m_id,
                                                   img);
                    if(out_msg)
                    {
                        m_outputQueue.Push(out_msg);
                    }
                }
                break;
            case Message::MSG_SET_IMAGE:
                {
                    // Receive an image.
                    if(m_processor)
                    {
                        m_processor->SetImage(msg->GetId(),
                                              msg->GetImage());
                    }
                }
                break;
            case Message::MSG_PARAMETERS:
                {
                    // Receive parameters.
                    if(m_processor && msg->GetParameters())
                    {
                        spectral::Parameters *params = msg->GetParameters();
                        m_processor->SetParameters(params);
                        delete params;
                    }
                }
                break;
            case Message::MSG_SHUTDOWN:
                // Terminate thread.
                terminate = true;
                break;
            default:
                // Nothing required
                break;
            }
            delete msg;
        }
    }

    return NULL;
}

}
