#include "tsort.h"
#include "scheduler.h"

namespace control {

void Scheduler::Clear(void)
{
    m_status.clear();
    m_dependencies.clear();
}
void
Scheduler::AddNode(int id)
{
    m_status[id] = std::pair<JobStatus, float>(PENDING, (float)0.0);
}

bool
Scheduler::AddEdge(int from, int to)
{
    bool result(false);

    if((m_status.find(from) != m_status.end()) &&
       (m_status.find(to) != m_status.end()))
    {
        std::list<int> order;
        m_dependencies.insert(std::pair<int,int>(from, to));

        if(topological_sort(m_dependencies, order))
	{
	    result = true;
	}
	else
	{
	    // circular dependency!
	    m_dependencies.erase(m_dependencies.find(
	        std::pair<int,int>(from, to)));
	}

    }
    return result;
}

// Invalidate the output of a node and all its descendants.
void
Scheduler::Invalidate(int id)
{
    std::set<int> to_invalidate;
    std::set<int> root;
    std::set<int>::iterator i;

    root.insert(id);

    find_descendants(m_dependencies,
                     root,
                     to_invalidate);

    to_invalidate.insert(id);

    i = to_invalidate.begin();

    while(i != to_invalidate.end())
    {
        std::map<int, std::pair<JobStatus,float> >::iterator status;

        status = m_status.find(*i);

        if(status != m_status.end())
        {
            m_status.erase(status);

            m_status[*i] = std::pair<JobStatus,float>(PENDING, 0.0);
        }
        i++;
    }
}

void
Scheduler::SetStarted(int id)
{
    std::map<int, std::pair<JobStatus,float> >::iterator status;

    status = m_status.find(id);

    if(status != m_status.end())
    {
        float progress;
	progress = (*status).second.second;
	m_status.erase(status);
	m_status[id] = std::pair<JobStatus, float>(STARTED, progress);
    }
}

void
Scheduler::SetComplete(int id)
{
    std::map<int, std::pair<JobStatus,float> >::iterator status;

    status = m_status.find(id);

    if(status != m_status.end())
    {
        float progress;
	progress = (*status).second.second;
	m_status.erase(status);
	m_status[id] = std::pair<JobStatus, float>(COMPLETE, progress);
    }
}

void
Scheduler::ClearProgress(void)
{
}

void
Scheduler::SetProgress(int id, float progress)
{
    std::map<int, std::pair<JobStatus,float> >::iterator status;

    status = m_status.find(id);

    if(status != m_status.end())
    {
        JobStatus js;
	js = (*status).second.first;
	m_status.erase(status);
	m_status[id] = std::pair<JobStatus, float>(js, progress);
    }
}

void
Scheduler::GetNextScheduledJobs(std::set<int> &out) const
{
    out.clear();

    std::set<int> complete;
    std::set<int> can_run;

    GetCompletedJobs(complete);

    calculate_schedule(m_dependencies, complete, can_run);

    if(can_run.size())
    {
        std::set<int>::iterator i = can_run.begin();
	while(i != can_run.end())
	{
	    if(GetStatus(*i) == PENDING)
	    {
	        out.insert(*i);
	    }
	    i++;
	}
    }
}

void
Scheduler::GetCompletedJobs(std::set<int> &out) const
{
    std::map<int, std::pair<JobStatus,float> >::const_iterator i;

    out.clear();

    i = m_status.begin();

    while(i != m_status.end())
    {
        if((*i).second.first == COMPLETE)
	{
	    out.insert(i->first);
	}
        i++;
    }
}

void
Scheduler::GetPendingJobs(std::set<int> &out) const
{
    std::map<int, std::pair<JobStatus,float> >::const_iterator i;

    out.clear();

    i = m_status.begin();
    while(i != m_status.end())
    {
        if((*i).second.first == PENDING)
	{
	    out.insert(i->first);
	}
        i++;
    }
}

void
Scheduler::GetStartedJobs(std::set<int> &out) const
{
    std::map<int, std::pair<JobStatus,float> >::const_iterator i;

    out.clear();

    i = m_status.begin();
    while(i != m_status.end())
    {
        if((*i).second.first == STARTED)
	{
	    out.insert(i->first);
	}
        i++;
    }
}

float
Scheduler::GetProgress(int id) const
{
    float progress(0.0);

    std::map<int, std::pair<JobStatus,float> >::const_iterator i;

    i = m_status.find(id);

    if(i != m_status.end())
    {
        progress = i->second.second;
    }

    return progress;
}

Scheduler::JobStatus
Scheduler::GetStatus(int id) const
{
    JobStatus status(UNKNOWN);

    std::map<int, std::pair<JobStatus,float> >::const_iterator i;

    i = m_status.find(id);

    if(i != m_status.end())
    {
        status = i->second.first;
    }

    return status;
}

}
