#include "thread.h"
#include <time.h>

#ifdef _WIN32
#include <Winbase.h>
#else
#include <sys/time.h>
#endif

namespace thread {

WorkerThread::WorkerThread(JobQueue &output)
    : Thread()
    , m_terminate(false)
    , m_output(output)
{
    Start();
}

WorkerThread::~WorkerThread()
{
}

void WorkerThread::Terminate(void)
{
    m_terminate = true;
    m_input.Ping();
}

void *
WorkerThread::MainLoop(void)
{
    while(!m_terminate)
    {
        Job *job;
        job = m_input.Pop();

        if(job)
        {
            job->DoProcess();
            m_output.Push(job);
        }
    }
    return NULL;
}

void
WorkerThread::Push(Job *job)
{
    m_input.Push(job);
}

////////////////////////////////////////////////////////////////////////////////
// ThreadPool


// Statics
static ThreadPool *s_threadPool(NULL);
static Mutex s_threadPoolMutex;

ThreadPool*
ThreadPool::GetInstance(void)
{
    ThreadPool *result(NULL);

    s_threadPoolMutex.Lock();
    if(!s_threadPool)
    {
        s_threadPool = new ThreadPool();
    }
    result = s_threadPool;
    s_threadPoolMutex.Unlock();

    return result;
}

void ThreadPool::Shutdown(void)
{
    s_threadPoolMutex.Lock();

    if(s_threadPool)
    {
        s_threadPool->Terminate();

        delete s_threadPool;

        s_threadPool = NULL;
    }

    s_threadPoolMutex.Unlock();
}

bool
ThreadPool::CanUsePool(void)
{
    bool result(false);

    s_threadPoolMutex.Lock();

    if(s_threadPool)
    {
        result = s_threadPool->_CanUsePool();
    }

    s_threadPoolMutex.Unlock();

    return result;
}

// memers
ThreadPool::ThreadPool()
    : Thread()
    , m_allowJobs(true)
    , m_targetThreadCount(0)
    , m_nextThreadId(0)
    , m_terminate(false)
{
    SetNumThreads(1);

    Start();
}

ThreadPool::~ThreadPool()
{
    {
        Lock lock(this);

        m_allowJobs = false;
        m_farmers.clear();
        m_destinations.clear();
    }
    Resize(0);
}

bool
ThreadPool::_CanUsePool(void)
{
    Lock lock(this);

    std::map<size_t, WorkerThread*>::const_iterator i = m_threads.begin();

#ifdef _WIN32
    STM_THREAD_TYPE this_thread = GetCurrentThread();
#else
    STM_THREAD_TYPE this_thread = pthread_self();
#endif

    while(i != m_threads.end())
    {
#if 0
        if(GetThreadId((i->second)->GetThread()) == GetThreadId(this_thread))
        {
            return false;
        }
#else
        if((i->second)->GetThread() == this_thread)
        {
            return false;
        }
#endif
        i++;
    }
    return true;
}

void
ThreadPool::Terminate(void)
{
    Lock lock(this);

    m_terminate = true;
    m_queue.Ping();
}

void *
ThreadPool::MainLoop(void)
{
    while(!m_terminate)
    {
        Job *job;
        job = m_queue.Pop();

        if(job)
        {
            Lock lock(this);


            if(job->Processed())
            {
                // We have processed this job.  See where it is going.
                bool delete_job(true);

                std::map<Job*, Farmer*>::iterator f_it =
                    m_destinations.find(job);

                // and which thread it is from.
                std::map<Job *, size_t>::iterator t_it =
                    m_jobs.find(job);

                // Do we have a destination?
                if(f_it != m_destinations.end())
                {
                    // Is this farmer registered?
                    if(m_farmers.find(f_it->second) != m_farmers.end())
                    {
                        delete_job = false;
                        f_it->second->GetComplete().Push(job);
                    }

                    m_destinations.erase(f_it);
                }

                // Set its thread back to inactive.
                if(t_it != m_jobs.end())
                {
                    size_t thread_id = t_it->second;

                    std::set<size_t>::iterator a_it =
                        m_activeThreads.find(thread_id);

                    if(a_it != m_activeThreads.end())
                    {
                        m_activeThreads.erase(a_it);
                    }
                    m_idleThreads.insert(thread_id);
                }

                if(delete_job)
                {
                    delete job;
                }
            }
            else
            {
                // An unprocessed job.  Pop it on the pending queue.
                m_pending.push(job);
            }


            //job->DoProcess();
            //m_output.Push(job);
        }

        // See if we have any unwanted threads
        TrimThreads();

        // See if we can assign any pending jobs tp threads.
        AssignJobs();
    }
    Resize(0);
    return NULL;
}

void
ThreadPool::AssignJobs(void)
{
    Lock lock(this);


    while(m_pending.size() && m_idleThreads.size())
    {
        Job *job;
        job = m_pending.front();
        m_pending.pop();

        if(job)
        {
            bool assign_job(false);

            std::map<Job *, Farmer *>::iterator dest_it =
                m_destinations.find(job);

            // Make sure its destination exists.
            if(dest_it != m_destinations.end())
            {
                std::set<Farmer *>::iterator farmer_it =
                    m_farmers.find(dest_it->second);
                if(farmer_it != m_farmers.end())
                {
                    assign_job = true;
                }
            }

            if(assign_job)
            {
                // Find the first idle thread.
                size_t thread_id = *m_idleThreads.begin();

                std::map<size_t, WorkerThread*>::iterator thread_it =
                    m_threads.find(thread_id);

                if(thread_it != m_threads.end())
                {
                    // Set thread associated with this job.
                    m_jobs[job] = thread_id;
                    // Mark thread active.
                    m_idleThreads.erase(m_idleThreads.begin());
                    m_activeThreads.insert(thread_id);
                    // Push job.
                    thread_it->second->Push(job);
                }
            }
            else
            {
                // farmer has gone away.
                delete job;
            }
        }
    }
}


void
ThreadPool::SetNumThreads(size_t count)
{
    if(count == 0)
    {
        count = 1;
    }
    Resize(count);
}

void
ThreadPool::Resize(size_t count)
{
    {
        Lock lock(this);


        if(count > m_threads.size())
        {
            // We are to add threads to the pool.
            size_t to_add = count - m_threads.size();
    
            for(size_t i=0;i<to_add;i++)
            {
                WorkerThread *new_thread;
                new_thread = NULL;
    
                new_thread = new WorkerThread(m_queue);
    
                if(new_thread)
                {
                    m_threads[m_nextThreadId] = new_thread;
                    m_idleThreads.insert(m_nextThreadId);
                    m_nextThreadId++;
                }
            }
        }
        m_targetThreadCount = count;
    }

    TrimThreads();
}

void 
ThreadPool::TrimThreads(void)
{
    Lock lock(this);
  

    if(m_targetThreadCount < m_threads.size())
    {
        // We are eliminating threads.
        size_t to_go = m_threads.size() - m_targetThreadCount;

        while(to_go && m_idleThreads.size())
        {
            std::map<size_t, WorkerThread*>::iterator i;
            size_t thread_id;

            thread_id = *(m_idleThreads.begin());
            i = m_threads.find(thread_id);

            // Clean up idle threads.
            if(i != m_threads.end())
            {
                i->second->Terminate();
                i->second->Join();
                delete (i->second);
                m_threads.erase(i);
                m_idleThreads.erase(m_idleThreads.begin());
                to_go--;
            }
        }
    }
}


void 
ThreadPool::Push(Job *job, Farmer *farmer)
{
    Lock lock(this);


    if(job)
    {
        if(farmer)
        {
            if(m_farmers.find(farmer) != m_farmers.end())
            {
                m_destinations[job] = farmer;
                m_queue.Push(job);
            }
        }
        else
        {
            delete job;
        }
    }
}

void 
ThreadPool::RegisterFarmer(Farmer *farmer)
{

    if(farmer)
    {
        Lock lock(this);
        m_farmers.insert(farmer);
    }
}

void
ThreadPool::UnRegisterFarmer(Farmer *farmer)
{
    if(farmer)
    {
        Lock lock(this);
        std::set<Farmer*>::iterator i = m_farmers.find(farmer);
        if(i != m_farmers.end())
        {
            m_farmers.erase(i);
        }
    }
}

////////////////////////////////////////////////////////////////////////////////

Farmer::Farmer()
    : m_numSubmitted(0)
    , m_numCompleted(0)
    , m_terminate(false)
{
    m_pool = ThreadPool::GetInstance();
    if(m_pool)
    {
        m_pool->RegisterFarmer(this);
    }
}

Farmer::~Farmer()
{
    if(m_pool)
    {
        m_pool->UnRegisterFarmer(this);
    }
}

void
Farmer::Push(Job *job)
{
    if(job)
    {
        m_numSubmitted++;
        m_pending.Push(job);
    }
}

void
Farmer::Process(void)
{
    m_terminate = false;

    if(m_pool)
    {
        while(m_numCompleted < m_numSubmitted)
        {
            Job *job;

            while(m_pending.Peek())
            {
                m_pool->Push(m_pending.Pop(), this);
            }

            job = m_complete.Pop();

            if(job)
            {
                ProcessCompletedJob(job);

                m_numCompleted++;
                delete job;
            }
        }
    }
}

void
Farmer::ProcessCompletedJob(Job *)
{
}

}

