#include "threadutils.h"
#include <time.h>
#include <iostream>

#ifndef _WIN32
#include <sys/time.h>
#endif

namespace thread {

#ifdef _WIN32

////////////////////////////////////////////////////////////////////////////////
// WINDOWS CODE
////////////////////////////////////////////////////////////////////////////////

uint32_t GetProcessorCount(void)
{
    uint32_t count = 0;

    SYSTEM_INFO info;
    GetSystemInfo(&info);
    count = info.dwNumberOfProcessors;

    return count;
}


Mutex::Mutex()
{
    m_mutex = CreateMutex(NULL, FALSE, NULL);
}

Mutex::~Mutex()
{
    CloseHandle(m_mutex);
}

void
Mutex::Lock(void)
{
    WaitForSingleObject(m_mutex, INFINITE);
}

bool
Mutex::TryLock(void)
{
    return(WaitForSingleObject(m_mutex, 0) == WAIT_OBJECT_0);
}

void
Mutex::Unlock(void)
{
    ReleaseMutex(m_mutex);
}

////////////////////////////////////////////////////////////////////////////////

Condition::Condition()
{
    m_cond.handles[0] = CreateEvent(NULL, TRUE, FALSE, NULL);  // manual
    m_cond.handles[1] = CreateEvent(NULL, FALSE, FALSE, NULL); // auto
}

Condition::~Condition()
{
    CloseHandle(m_cond.handles[0]);
    CloseHandle(m_cond.handles[1]);
}

void
Condition::Wait(Mutex &lock)
{
    lock.Unlock();
    WaitForMultipleObjects(2, m_cond.handles, FALSE, INFINITE);
    lock.Lock();
}

void
Condition::TimedWait(Mutex &lock, uint32_t ms)
{

/*
    struct timespec timeout;
    uint32_t s = ms / 1000;
    uint32_t ns = (ms % 1000) * 1000000;

    //clock_gettime(CLOCK_REALTIME, &timeout);
    {
        struct timeval tv;
        gettimeofday(&tv, NULL);
        timeout.tv_sec = tv.tv_sec;
        timeout.tv_nsec = tv.tv_usec * 1000;
    }

    timeout.tv_sec+= s;
    timeout.tv_nsec+= ns;

    pthread_cond_timedwait(&m_cond, &(lock.GetMutex()), &timeout);
*/
}

void
Condition::Signal(void)
{
    SetEvent(m_cond.handles[1]);
}

void
Condition::Broadcast(void)
{
    SetEvent(m_cond.handles[0]);
    ResetEvent(m_cond.handles[0]);
}

////////////////////////////////////////////////////////////////////////////////

void *
Thread::Entry(void *p)
{
    void *result = NULL;

    Thread *thread = (Thread *)p;

    if(thread)
    {
        result = thread->MainLoop();
    }

    return result;
}

DWORD
Thread::WinEntry(void *p)
{
    Entry(p);
    return 0;
}

void
Thread::Start(void)
{
    Lock lock(this);
    
    m_thread = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)WinEntry, (void *)this, 0, NULL);
}

void
Thread::Join(void)
{
    Lock lock(this);

    WaitForSingleObject(m_thread, INFINITE);
}

#else

////////////////////////////////////////////////////////////////////////////////
// UNIX/POSIX code
////////////////////////////////////////////////////////////////////////////////

uint32_t GetProcessorCount(void)
{
    unsigned int count = 1;
#ifdef _SC_NPROCESSORS_CONF
    count = sysconf(_SC_NPROCESSORS_CONF);
#endif

    return count;
}


Mutex::Mutex()
{
    pthread_mutex_init(&m_mutex, NULL);
}

Mutex::~Mutex()
{
    pthread_mutex_destroy(&m_mutex);
}

void
Mutex::Lock(void)
{
    pthread_mutex_lock(&m_mutex);
}

bool
Mutex::TryLock(void)
{
    return(pthread_mutex_trylock(&m_mutex) == 0);
}

void
Mutex::Unlock(void)
{
    pthread_mutex_unlock(&m_mutex);
}

////////////////////////////////////////////////////////////////////////////////

Condition::Condition()
{
    pthread_cond_init(&m_cond, NULL);
}

Condition::~Condition()
{
    pthread_cond_destroy(&m_cond);
}

void
Condition::Wait(Mutex &lock)
{
    pthread_cond_wait(&m_cond, &(lock.GetMutex()));
}

void
Condition::TimedWait(Mutex &lock, uint32_t ms)
{
    struct timespec timeout;
    uint32_t s = ms / 1000;
    uint32_t ns = (ms % 1000) * 1000000;

    //clock_gettime(CLOCK_REALTIME, &timeout);
    {
        struct timeval tv;
        gettimeofday(&tv, NULL);
        timeout.tv_sec = tv.tv_sec;
        timeout.tv_nsec = tv.tv_usec * 1000;
    }

    timeout.tv_sec+= s;
    timeout.tv_nsec+= ns;

    pthread_cond_timedwait(&m_cond, &(lock.GetMutex()), &timeout);
}

void
Condition::Signal(void)
{
    pthread_cond_signal(&m_cond);
}

void
Condition::Broadcast(void)
{
    pthread_cond_broadcast(&m_cond);
}

////////////////////////////////////////////////////////////////////////////////

void *
Thread::Entry(void *p)
{
    void *result = NULL;

    Thread *thread = (Thread *)p;

    if(thread)
    {
        result = thread->MainLoop();
    }

    return result;
}


void
Thread::Start(void)
{
    Lock lock(this);
    pthread_create(&m_thread, NULL, Entry, (void *)this);
}

void
Thread::Join(void)
{
    Lock lock(this);

    void *result;
    pthread_join(m_thread, &result);
}

#endif

////////////////////////////////////////////////////////////////////////////////

Lockable::Lockable()
{
}

Lockable::~Lockable()
{
}

Lockable::Lock::Lock(Lockable *obj)
    : m_obj(obj)
{
    if(m_obj)
    {
        m_obj->GetLock().Lock();
    }
}

Lockable::Lock::~Lock()
{
    if(m_obj)
    {
        m_obj->GetLock().Unlock();
    }
}

}
