#include "tsort.h"

namespace control {

// Given a list of pairs of ints, extract all the ids referenced.
void get_nodes(const std::set<std::pair<int, int> > &edges,
               std::set<int> &nodes)
{
    nodes.clear();

    std::set<std::pair<int, int> >::const_iterator i = edges.begin();

    while(i != edges.end())
    {
        nodes.insert(i->first);
        nodes.insert(i->second);

        i++;
    }
}

// find out how many dependencies a node has.  I.e. how many edges point to
// it.
int count_dependencies(const std::set<std::pair<int, int> > &edges, int id)
{
    int result(0);

    std::set<std::pair<int, int> >::const_iterator i = edges.begin();

    while(i != edges.end())
    {
        if(i->second == id)
        {
            result++;
        }
        i++;
    }

    return result;
}

// Remove edges coming from a particular node.
void remove_edges(std::set<std::pair<int, int> > &edges, int id)
{
    std::set<std::pair<int, int> >::iterator i = edges.begin();

    while(i != edges.end())
    {
        if(i->first == id)
        {
            std::set<std::pair<int, int> >::iterator next = i;
            next++;
            edges.erase(i);
            i = next;
        }
        else
        {
            i++;
        }
    }
}

// Internal topological sort implementation.
bool topological_sort(std::set<std::pair<int, int> > &edges,
                      std::set<int>  &nodes,
                      std::list<int> &order)
{
    bool result(true);

    while(result && nodes.size())
    {
        std::set<int>::iterator i;
        std::list<int> to_go;
        result = false;

        i = nodes.begin();

        while((i != nodes.end()))
        {
            if(count_dependencies(edges, *i) == 0)
            {
               // Nothing pointing at this node, so it can go on the
               // output list.

               // put this result on output.
               order.push_back(*i);
               to_go.push_back(*i);

               result = true;
            }
            i++;
        }

        if(to_go.size())
        {
            std::list<int>::iterator j = to_go.begin();
            while(j != to_go.end())
            {
                remove_edges(edges, *j);
                nodes.erase(nodes.find(*j));
                j++;
            }
        }
    }

    return result;
}

////////////////////////////////////////////////////////////////////////////////
//
// External functions
//
////////////////////////////////////////////////////////////////////////////////

bool topological_sort(const std::set<std::pair<int, int> > &edges,
                      std::list<int> &order)
{
    bool result(true);
    std::set<int> nodes;
    std::set<std::pair<int, int> > edges_copy(edges);

    order.clear();

    get_nodes(edges_copy, nodes);

    if(nodes.size())
    {
        result = topological_sort(edges_copy, nodes, order);
    }

    if(!result)
    {
        // Cyclic graph
        order.clear();
    }

    return result;
}

// Analyse a graph for concurrent jobs.
void calculate_schedule(const std::set<std::pair<int, int> > &edges,
                        const std::set<int> &completed,
                        std::set<int> &can_run)
{
    can_run.clear();

    if(edges.size())
    {
        std::set<std::pair<int,int> > edges_copy(edges);
        std::set<int> nodes;

        get_nodes(edges_copy, nodes);

        // Changes related to completed jobs.
        {
            std::set<int>::iterator i = completed.begin();

            while(i != completed.end())
            {
                // Remove completed jobs from node list.
                {
                    std::set<int>::iterator to_go;
                    to_go = nodes.find(*i);

                    if(to_go != nodes.end())
                    {
                        nodes.erase(to_go);
                    }
                }

                // Now remove edges running from completed nodes.
                remove_edges(edges_copy, *i);

                i++;
            }
        }

        // Now look at remaining edges.
        {
            std::set<std::pair<int,int> >::iterator i;

            i = edges_copy.begin();
            while(i != edges_copy.end())
            {
                std::set<int>::iterator to_go;

                to_go = nodes.find(i->second);

                if(to_go != nodes.end())
                {
                    nodes.erase(to_go);
                }

                i++;
            }
        }

        can_run.insert(nodes.begin(), nodes.end());
    }
}

// Find all descendants of a node.
// Ludicrously crap implementation.
void find_descendants(const std::set<std::pair<int, int> > &edges,
                      const std::set<int> &root,
                      std::set<int> &out)
{
    std::set<int> scratch[2];
    std::set<int> &cur(scratch[0]);
    std::set<int> &nxt(scratch[1]);

    std::set<std::pair<int, int> > edges_copy(edges);

    out.clear();
    cur.insert(root.begin(), root.end());

    do
    {
        std::set<std::pair<int, int> >::iterator j;
        nxt.clear();


        j = edges_copy.begin();

        while(j != edges_copy.end())
        {
            // does this edge come from the current set?
            if(cur.find(j->first) != cur.end())
            {
                // Find the items this edge points at.
                nxt.insert(j->second);
            }

            j++;
        }
           
        // Add nxt set to results.
        {
            out.insert(nxt.begin(), nxt.end());
        }
        {
            std::set<int>::iterator i = nxt.begin();
       
            while(i != nxt.end())
            {
                remove_edges(edges_copy, *i);

                i++;
            }
        }
        // swap cur & nxt;
        {
            std::set<int> &tmp(cur);
            cur = nxt;
            tmp = cur;
        }
    } while(cur.size());
}

} // namespace spectral
