#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "htond.h"
#include "utils.h"

uint32_t get_pixel_size(uint32_t colour_space)
{
    uint32_t result = 1;

    switch(colour_space)
    {
    case CLS_SRGB_LINEAR:
    case CLS_SRGB:
    case CLS_CIE_XYZ:
    case CLS_CIE_XYY:
    case CLS_LOG_LUV:
        result = 3;
        break;
    case CLS_LUMINANCE:
    case CLS_DENSITY:
    case CLS_NUMBER:
        result = 1;
    }
    return result;
}

float
scalar_product(const float *a, const float *b)
{
    float result = 0;

    if(a && b)
    {
        result = (a[0] * b[0]) +
                 (a[1] * b[1]) +
                 (a[2] * b[2]);
    }

    return result;
}

void
vector_product(const float *a, const float *b, float *out)
{
    if(a && b && out)
    {
        out[0] = (a[1] * b[2]) - (b[1] * a[2]);
        out[1] = (b[0] * a[2]) - (a[0] * b[2]); 
        out[2] = (a[0] * b[1]) - (b[0] * a[1]);
    }
}

void
vector_normalise(float *v)
{
    if(v)
    {
        int i;
        float r(0.0);

        for(i=0;i<3;i++)
        {
            r+= (v[i] * v[i]);
        }

        r = sqrt(r);

        if(r > 1.0e-10)
        {
            for(i=0;i<3;i++)
            {
                v[i]/= r;
            }
        }
    }
}

// matrix/vector multiplier.
void transform_colour(const Matrix33 matrix, float *colour)
{
    transform_vector(matrix, colour);
}

// matrix/vector multiplier.
void transform_vector(const Matrix33 matrix, float *vector)
{
    if(vector)
    {
        float x, y, z;
        x = scalar_product(matrix[0], vector);
        y = scalar_product(matrix[1], vector);
        z = scalar_product(matrix[2], vector);
        vector[0] = x;
        vector[1] = y;
        vector[2] = z;
    }
}

void matrix_identity(Matrix33 *out)
{
    if(out)
    {
        for(int i=0;i<3;i++)
        {
            for(int j=0;j<3;j++)
            {
                if(i == j)
                {
                    (*out)[i][j] = 1.0;
                }
                else
                {
                    (*out)[i][j] = 0.0;
                }
            }
        }
    }
}
// matrix/matrix multiplier.
void matrix_multiply(const Matrix33 m1,
                     const Matrix33 m2,
                     Matrix33 *out)
{
    if(out)
    {
        for(int i=0;i<3;i++)
        {
            for(int j=0;j<3;j++)
            {
                (*out)[i][j] = (m1[i][0] * m2[0][j]) +
                               (m1[i][1] * m2[1][j]) +
                               (m1[i][2] * m2[2][j]);
            }
        }
    }
}

float matrix_inverse(const Matrix33 m,
                     Matrix33 *out)
{
    float det(0.0);
    
    if(out)
    {
        det=(m[0][0] * ((m[1][1] * m[2][2]) - (m[2][1] * m[1][2]))) -
            (m[0][1] * ((m[1][0] * m[2][2]) - (m[1][2] * m[2][0]))) +
            (m[0][2] * ((m[1][0] * m[2][1]) - (m[1][1] * m[2][0])));
    
        if(fabs(det) > 1e-6)
        {
            (*out)[0][0]= ((m[1][1] * m[2][2]) - (m[2][1] * m[1][2]))/det;
            (*out)[1][0]=-((m[1][0] * m[2][2]) - (m[1][2] * m[2][0]))/det;
            (*out)[2][0]= ((m[1][0] * m[2][1]) - (m[2][0] * m[1][1]))/det;
            
            (*out)[0][1]=-((m[0][1] * m[2][2]) - (m[0][2] * m[2][1]))/det;
            (*out)[1][1]= ((m[0][0] * m[2][2]) - (m[0][2] * m[2][0]))/det;
            (*out)[2][1]=-((m[0][0] * m[2][1]) - (m[2][0] * m[0][1]))/det;
            
            (*out)[0][2]= ((m[0][1] * m[1][2]) - (m[0][2] * m[1][1]))/det;
            (*out)[1][2]=-((m[0][0] * m[1][2]) - (m[1][0] * m[0][2]))/det;
            (*out)[2][2]= ((m[0][0] * m[1][1]) - (m[1][0] * m[0][1]))/det;
        }
    }
    
    return det;
}


// Create a transformation matrix to convert a colour from one wp to another
void chromatic_adaptation(const float *ws, const float *wd, Matrix33 *out)
{

    if(ws && wd && out)
    {
        int i;
#define VK
        const float MA[3][3] =
        {
            /* linear*/
#if defined(LINEAR)
             {1,0,0},{0,1,0},{0,0,1}
#elif defined(BRADFORD)
            { 0.8951,-0.7502, 0.0389},
            { 0.2664, 1.7135,-0.0685},
            {-0.1614, 0.0367, 1.0296}
#elif defined(VK)
            { 0.40024,-0.22630, 0.00000},
            { 0.70760, 1.16532, 0.00000},
            {-0.08081, 0.04570, 0.91822} 
#endif
        };
    
        const float MA_inv[3][3] =
        {
#if defined(LINEAR)
             {1,0,0},{0,1,0},{0,0,1}
#elif defined(BRADFORD)
            { 0.986993, 0.432305,-0.008529},
            {-0.147054, 0.518360, 0.040043},
            { 0.159963, 0.049291, 0.968487}
#elif defined(VK)
            { 1.8599364, 0.3611914, 0.0000000},
            {-1.1293816, 0.6388125, 0.0000000},
            { 0.2198974,-0.0000064, 1.0890636}
#endif
        };

        float diag[3][3] =
        {
            {0, 0, 0},
            {0, 0, 0},
            {0, 0, 0}
        };
    
        float cs[3], cd[3];
    
        for(i=0;i<3;i++)
        {
            cs[i]=ws[i];
            cd[i]=wd[i];
        }
    
        transform_colour(MA, cs);
        transform_colour(MA, cd);

        for(i=0;i<3;i++)
        {
            diag[i][i] = cd[i]/cs[i];
        }

        Matrix33 tmp1;
        matrix_multiply(diag, MA, &tmp1);
        matrix_multiply(MA_inv, tmp1, out);
    }
}

void linear_srgb_to_xyy(float *colour)
{
    if(colour)
    {
        // first convert to xyz.
        {
            const float matrix[3][3] =
                {{ 0.412424,  0.357579,  0.180464},
                 { 0.212656,  0.715158,  0.0721856},
                 { 0.0193324, 0.119193,  0.950444}};
            transform_colour(&(matrix[0]), colour); 
        } 
        xyz_to_xyy(colour);
    }
}

void xyy_to_linear_srgb(float *colour)
{
    if(colour)
    {
        xyy_to_xyz(colour);
        // xyz to srgb.
        {
            const float matrix[3][3] =
                {{ 3.24071, -1.53726, -0.498571},
                 {-0.969258,  1.87599,  0.0415557},
                 { 0.0556352, -0.203996,  1.05707}};
            transform_colour(&(matrix[0]), colour); 
        } 
    }
}

void xyy_to_xyz(float *colour)
{
    if(colour)
    {
        float X(0.0), Y(0.0), Z(0.0);

	if(colour[2] > 1.0e-6)
	{
            X = (colour[2] * colour[0]) / colour[1];
            Y = colour[2];
            Z = ((1.0 - (colour[0] + colour[1])) * colour[2])/colour[1];
        }

        colour[0] = X;
        colour[1] = Y;
        colour[2] = Z;
    }
}

void xyz_to_xyy(float *colour)
{
    if(colour)
    {
        // Now xyy.
        float total;
        total = colour[0] + colour[1] + colour[2];
        colour[0]/= total;
        colour[2] = colour[1];
        colour[1]/= total;
    }
}

void xyz_to_linear_srgb(float *colour)
{
    if(colour)
    {
        // xyz to srgb.
        const float matrix[3][3] =
            {{ 3.24071, -1.53726, -0.498571},
             {-0.969258,  1.87599,  0.0415557},
             { 0.0556352, -0.203996,  1.05707}};
        transform_colour(&(matrix[0]), colour); 
    }
}

float remove_srgb_gamma(float x)
{
    float result;
    if(x < 0.004045)
    {
        result = x / 12.92;
    }
    else
    {
        result = pow(((x + 0.055) / 1.055), 2.4);
    }
    return result;
}

float apply_srgb_gamma(float x)
{
    float result;
    if (x < 0.0031308)
    {
           result = x * 12.92;
    } 
    else
    {
        result = (pow(x, (float)0.4166667) * 1.055) - 0.055;
    }
    return result;
}

double luminance_to_density(double luminance)
{
    if(luminance < 1e-20)
    {
        luminance = 1e-20;
    }
    return log(luminance);
}

double density_to_luminance(double density)
{
    return exp(density);
}

double halton(uint32_t base, uint32_t n)
{
    double result = 0.0;
    double rem;
    double frac = 1.0 / (double)base;
    //int N1 = 0;
 
    while(n)
    {
        //N1  = (n / base);
        rem = n % base;
        result+= (frac * rem);
        n/= base;
        frac/= (double)base;
    }
    return result;
}

double gaussian(double sigma, double r)
{
    const double pi = 3.1415926535897932;
    double two_sigma_squared = (2.0 * sigma * sigma);
    double denom = sqrt(two_sigma_squared * pi);

    double exponent = -((r * r) / two_sigma_squared);

    return (1.0 / denom) * exp(exponent);
}

float
round_float(float x)
{
#ifdef _WIN32
	float fx = floor(x);
	if((x - fx) > 0.5)
	{
		return fx + 1;
	}
	else
	{
		return fx;
	}
#else
	return round(x);
#endif
}

double
round_double(double x)
{
#ifdef _WIN32
	double fx = floor(x);
	if((x - fx) > 0.5)
	{
		return fx + 1;
	}
	else
	{
		return fx;
	}
#else
	return round(x);
#endif
}

float
trunc_float(float x)
{
#ifdef _WIN32
	float fx = floor(x);
	if(x > 0)
	{
		return floor(x);
	}
	else if(x < 0)
	{
		return ceil(x);
	}
	else
	{
		return 0;
	}
#else
	return trunc(x);
#endif
}

double
trunc_double(double x)
{
#ifdef _WIN32
	float fx = floor(x);
	if(x > 0)
	{
		return floor(x);
	}
	else if(x < 0)
	{
		return ceil(x);
	}
	else
	{
		return 0;
	}
#else
	return trunc(x);
#endif
}
