#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include <string>

#include "utils.h"
#include "tiff_loader.h"
#include "bufutils.h"

#define D65_X 0.31271
#define D65_Y 0.32902

#if 0
void matrix_display(const char *name, const Matrix33 m)
{
    size_t i, j;

    if(name)
    {
        fprintf(stderr, "\n\r%s =\n\r", name);
    }

    fprintf(stderr, "{{%f, %f, %f},\n\r", m[0][0],m[0][1],m[0][2]);
    fprintf(stderr, " {%f, %f, %f},\n\r", m[1][0],m[1][1],m[1][2]);
    fprintf(stderr, " {%f, %f, %f}};\n\r\n\r", m[2][0],m[2][1],m[2][2]);
}
#else
void matrix_display(const char *, const Matrix33 )
{
}
#endif

//
// Utilities to act on buffers.
//


void
TiffLoader::buf_transform(Matrix33 matrix, size_t size, float *buf)
{
    if(buf)
    {
        for(size_t i = 0, index=0; i<size; i++,index+=3)
        {
            transform_colour(matrix, &(buf[index]));
        }
    }
}


//
// Tiff reading helpers.
//

// Get the size of a tiff image in pixels.
bool
TiffLoader::tiff_get_size(TIFF *t, uint32_t *width, uint32_t *height)
{
    bool valid = true;

    if(t && width && height)
    {
        uint32_t tmp;

        if(TIFFGetField(t, TIFFTAG_IMAGEWIDTH, &tmp) == 1)
        {
            *width = tmp;
        }
        else
        {
            valid = false;
        }

        if(TIFFGetField(t, TIFFTAG_IMAGELENGTH,&tmp) == 1)
        {
            *height= tmp;
        }
        else
        {
            valid = false;
        }
    }
    else
    {
        valid = false;
    }

    return valid;
} 

// get the luminance scale factor of a tiff image.
bool
TiffLoader::tiff_get_stonits(TIFF *t, double &stonits)
{
    bool valid = true;

    if(t)
    {
        double tmp;

        if(TIFFGetField(t, TIFFTAG_STONITS, &tmp) == 1)
        {
            stonits = tmp;
        }
        else
        {
            stonits = 1.0;
        }
    }
    else
    {
        valid = false;
    }
    return valid;
}

bool
TiffLoader::tiff_get_whitepoint(TIFF *t, float *wp)
{
    bool valid = true;

    if(t && wp)
    {
        float *tmp;
        if(TIFFGetField(t, TIFFTAG_WHITEPOINT, &tmp) == 1)
        {
            wp[0] = tmp[0];
            wp[1] = tmp[1];

            fprintf(stderr, "read white = (%f, %f)\n\r", wp[0], wp[1]);
        }
        else
        {
            // Default to d65
            wp[0] = D65_X;
            wp[1] = D65_Y;

            fprintf(stderr, "default white = (%f, %f)\n\r", wp[0], wp[1]);
        }
    }
    else
    {
        valid = false;
    }
    return valid;
}

bool
TiffLoader::tiff_get_primaries(TIFF *t, float *r, float *g, float *b)
{
    bool result(true);

    if(t)
    {
        float *primaries;

        if(TIFFGetField(t, TIFFTAG_PRIMARYCHROMATICITIES, &primaries) == 1)
        {
            fprintf(stderr, "primaries = (%f,%f) (%f,%f) (%f,%f)\n\r",
                    primaries[0], primaries[1], 
                    primaries[2], primaries[3],
                    primaries[4], primaries[5]);
            if(r)
            {
                r[0] = primaries[0];
                r[1] = primaries[1];
            }
            if(g)
            {
                g[0] = primaries[2];
                g[1] = primaries[3];
            }
            if(b)
            {
                b[0] = primaries[4];
                b[1] = primaries[5];
            }
        }
        else
        {
            result = false;
        }
    }
    else
    {
        result = false;
    }

    return result;
}

// Handle reading of ieee rgb pixels.
bool
TiffLoader::tiff_read_ieee_rgb(TIFF *t, uint32_t width, uint32_t height, float *buf)
{
    bool valid(true);

    size_t size = size_t(width) * size_t(height);

    if(buf)
    {
        // Get scanlines.
        //
        // IEEE rgb is the same size as our target cie xyY buffer, so we can
        // read straight into the buffer then convert in situ.
        uint32_t y;
        float *sl = buf;

        for(y=0;y<height;y++)
        {
            TIFFReadScanline(t, sl, y, 0);
            sl+= (width * 3);
        }
        {
             bool calculate_transform(true);

             float white[3], red[3], green[3], blue[3];

             calculate_transform&= tiff_get_whitepoint(t, white);
             calculate_transform&= tiff_get_primaries(t, red, green, blue);

             if(calculate_transform)
             {
                  // RGB conversion based on chromaticities.
                  Matrix33 chrom_matrix, phos_matrix, inv_phos_matrix, tfm_matrix;
                  size_t i;
                  float s[3];

                  // First get coords for primaries & white.
                  red[2]   = 1.0;
                  green[2] = 1.0;
                  blue[2]  = 1.0;
                  white[2] = 1.0;

                  xyy_to_xyz(red);
                  xyy_to_xyz(green);
                  xyy_to_xyz(blue);
                  xyy_to_xyz(white);

                  for(i=0;i<3;i++)
                  {
                      phos_matrix[i][0] = red[i];
                      phos_matrix[i][1] = green[i];
                      phos_matrix[i][2] = blue[i];
                      
                      s[i] = white[i];
                  }

                  matrix_display("phos_matrix", phos_matrix);

                  matrix_inverse(phos_matrix, &inv_phos_matrix);

                  matrix_display("inv_phos_matrix", inv_phos_matrix);

                  transform_colour(inv_phos_matrix, s);

                  fprintf(stderr, "s = (%f, %f, %f)\n\r", s[0], s[1], s[2]);

                  for(i=0;i<3;i++)
                  {
                      tfm_matrix[i][0] = s[0] * phos_matrix[i][0];
                      tfm_matrix[i][1] = s[1] * phos_matrix[i][1];
                      tfm_matrix[i][2] = s[2] * phos_matrix[i][2];
                  }


                  matrix_display("tfm_matrix", tfm_matrix);

                 {
                      float wd[3] = {D65_X, D65_Y, 1}; // d65
                      xyy_to_xyz(wd);
                      chromatic_adaptation(white, wd, &chrom_matrix);
                  }

                  matrix_display("chrom_matrix", chrom_matrix);

                  // TODO: Combine these transforms.

                  // rgb to xyz.
                  buf_transform(tfm_matrix, (width * height), buf);
                  // fix white point.
                  buf_transform(chrom_matrix, (width * height), buf);
                  // Convert to xyY
                  buf_xyz_to_xyy(buf, size);
             }
             else
             {
                // Default to srgb.
                buf_linear_srgb_to_xyy(buf, size);
             }
        }
    }
    else
    {
        valid = false;
    }

    return valid;
}

bool
TiffLoader::tiff_read_logluv(TIFF *t, uint32_t width, uint32_t height, float *buf)
{
    bool valid(true);

    Matrix33 chrom_matrix;

    {
        float ws[3];
        float wd[3] = {D65_X, D65_Y, 1}; // d65
        tiff_get_whitepoint(t, ws); // source white point.
        ws[2] = 1;
        xyy_to_xyz(ws);
        xyy_to_xyz(wd);
        chromatic_adaptation(ws, wd, &chrom_matrix);
    }


    if(buf)
    {
        float *line = new float[width * 3];

        if(line)
        {
            uint32_t y, index(0);
        
            TIFFSetField(t, TIFFTAG_SGILOGDATAFMT, SGILOGDATAFMT_FLOAT);

            for(y=0;y<height;y++)
            {
                uint32_t x;
                float *xyz;
                xyz = line;

                TIFFReadScanline(t, line, y, 0);
 
                for(x=0;x<width;x++)
                {
                    // chromatic adaptation.
                    transform_colour(chrom_matrix, xyz);

                    // convert to xyy.
                    xyz_to_xyy(xyz);

                    buf[index++] = xyz[0];
                    buf[index++] = xyz[1];
                    buf[index++] = xyz[2];
                    xyz+= 3;
                }
            }

            delete line;
        }
        else
        {
            valid = false;
        }
    }
    else
    {
        valid = false;
    }
    return valid;
}

// Read the image.
bool
TiffLoader::tiff_read_image_pixels(TIFF *t, spectral::Image *image)
{
    float *buf = NULL;

    bool valid = true;

    if(image)
    {
        buf = image->get_buffer();
    }

    if(buf && t)
    {
        const uint32_t width(image->get_width());
        const uint32_t height(image->get_height());

        uint16 bps, spp, s_format, p_config, photometric, compression;

        TIFFGetField(t, TIFFTAG_BITSPERSAMPLE, &bps);
        TIFFGetField(t, TIFFTAG_SAMPLESPERPIXEL, &spp);
        TIFFGetField(t, TIFFTAG_SAMPLEFORMAT, &s_format);
        TIFFGetField(t, TIFFTAG_PLANARCONFIG, &p_config);
        TIFFGetField(t, TIFFTAG_PHOTOMETRIC, &photometric);
        TIFFGetField(t, TIFFTAG_COMPRESSION, &compression);

        if((s_format == SAMPLEFORMAT_IEEEFP) &&
           (bps == 32) &&
           (spp == 3) &&
           (p_config == PLANARCONFIG_CONTIG) &&
           (photometric == PHOTOMETRIC_RGB))
        {
            // IEEE floating point RGB image.
            valid = tiff_read_ieee_rgb(t, width, height, buf);
        }
        else if((photometric == PHOTOMETRIC_LOGLUV) &&
                (compression == COMPRESSION_SGILOG) &&
                (bps == 16) &&
                (spp == 3) &&
                (p_config == PLANARCONFIG_CONTIG) &&
                (s_format == SAMPLEFORMAT_INT))
        {
            tiff_read_logluv(t, width, height, buf);
        }
        else
        {
            valid = false;
        }
    }
    else
    {
        valid = false;
    }

    return valid;
}

//
// Main interface
//

// laod a tiff image.
spectral::Image *
TiffLoader::LoadImage(const std::string &fn)
{
    spectral::Image *result = NULL;
    TIFF *t = NULL;

    t = TIFFOpen(fn.c_str(), "r");

    if(t)
    {
        uint32_t width, height;
        double stonits;

        tiff_get_stonits(t, stonits);

        if(tiff_get_size(t, &width, &height))
        {
            const size_t size = size_t(width) * size_t(height);

            result = new spectral::Image(width, height, CLS_CIE_XYY);

            if(result)
            {
                if(result->get_buffer())
                {
                    if(!tiff_read_image_pixels(t, result))
                    {
                        delete result;
                        result = NULL;
                    }
                    else
                    {
                        //fprintf(stderr, "stonits = %lf\n\r", stonits);
                        buf_scale_xyy_y(result->get_buffer(), size, stonits);
                    }
                }
                else
                {
                    delete result;
                    result = NULL;
                }
            }
        }
        TIFFClose(t);
    }

    return result;
}

