

/*! \file
  \verbatim
  
    Copyright (c) 2006, Sylvain Paris and Frdo Durand

    Permission is hereby granted, free of charge, to any person
    obtaining a copy of this software and associated documentation
    files (the "Software"), to deal in the Software without
    restriction, including without limitation the rights to use, copy,
    modify, merge, publish, distribute, sublicense, and/or sell copies
    of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be
    included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
    HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
    WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.

  \endverbatim
*/



#include <stdio.h>
#include <stdlib.h>
#include <math.h>
//#include <assert.h>

#include <vector>
#include <set>
// p&d includes start.
#define NO_XML
#include <cmath>

#include <algorithm>

#define CHRONO
#include "geom.h"
#include "fast_lbf.h"
// p&d includes end.


#include "parameters.h"
#include "image.h"
#include "utils.h"
#include "bufutils.h"
#include "thread.h"

#include "bilateral.h"

#define IMG_LUMINANCE 0

namespace bilateral
{

////////////////////////////////////////////////////////////////////////////////
//
// Bilateral processor implementation.
//
////////////////////////////////////////////////////////////////////////////////

Bilateral::Bilateral()
    : spectral::Processor()
    , m_lumImage(NULL)
    , m_sigma(10)
    , m_threshold(0.3)
    , m_wrap_horizontal(false)
    , m_wrap_vertical(false)
{
}

Bilateral::~Bilateral()
{
}

// Get slots & output type.  Returns output colour space.
// pairs are <slot id, colour space>
uint32_t
Bilateral::GetInterface(
    std::list<std::pair<uint32_t, uint32_t> > &mandatory,
    std::list<std::pair<uint32_t, uint32_t> > &optional) const
{
    mandatory.clear();
    optional.clear();

    mandatory.push_back(
        std::pair<uint32_t, uint32_t>(IMG_LUMINANCE, CLS_LUMINANCE));

    return CLS_LUMINANCE;
}


bool
Bilateral::SetImage(uint32_t slot, const spectral::Image *img)
{
    if(slot == IMG_LUMINANCE)
    {
        m_lumImage = NULL;

        if(img)
        {
            if(img->get_colour_space() == CLS_LUMINANCE)
            {
                m_lumImage = img;

                return true;
            }
        }
        else
        {
            return true;
        }
    }
    return false;
}

void
Bilateral::GetParameters(spectral::Parameters *params) const
{
    if(params)
    {
        std::string tmp;

        tmp = "sigma";
        params->AddParameter(tmp, 1, 100, m_sigma);
        
        tmp = "range sigma";
        params->AddParameter(tmp, 0.1, 10.0, m_threshold);

        //tmp = "wrap horizontal";
        //params->AddParameter(tmp, m_wrap_horizontal);

        //tmp = "wrap vertical";
        //params->AddParameter(tmp, m_wrap_vertical);
    }
}

void
Bilateral::SetParameters(const spectral::Parameters *params)
{
    if(params)
    {
       std::string tmp;

       tmp = "sigma";
       params->GetParameter(tmp, m_sigma);
       
       tmp = "range sigma";
       params->GetParameter(tmp, m_threshold);
        
        //tmp = "wrap horizontal";
        //params->GetParameter(tmp, m_wrap_horizontal);

        //tmp = "wrap vertical";
        //params->GetParameter(tmp, m_wrap_vertical);
    }
}

spectral::Image *
Bilateral::Process(void)
{
    spectral::Image *result(NULL);

    Progress(0);
    if(m_lumImage)
    {
        Array_2D<double> image(m_lumImage->get_width(), 
                               m_lumImage->get_height());
        Array_2D<double> filtered_image(m_lumImage->get_width(), 
                                        m_lumImage->get_height());
        result = new spectral::Image(*m_lumImage);
        buf_luminance_to_density(result->get_buffer(), 
                                 result->get_width() * 
                                 result->get_height(),
                                 1.0e-10);

        {
            float *p = result->get_buffer();
            for(uint32_t y=0;y<m_lumImage->get_height();y++)
            {
                for(uint32_t x=0;x<m_lumImage->get_width();x++)
                {
                    image(x, y) = *p;
                    p++;
                }
            }
        }

        Image_filter::fast_LBF(image, 
                               image,
                               double(m_sigma), 
                               double(m_threshold/logf(2.0)),
                               false,
                               &filtered_image,
                               &filtered_image);

        {
            float *p = result->get_buffer();
            for(uint32_t y=0;y<m_lumImage->get_height();y++)
            {
                for(uint32_t x=0;x<m_lumImage->get_width();x++)
                {
                    *p = float(filtered_image(x, y));
                    p++;
                }
            }
        }

        buf_density_to_luminance(result->get_buffer(), 
                                 result->get_width() * result->get_height());
    }

    Progress(100.0);
    return result;
}

void
Bilateral::ClearImage(uint32_t slot)
{
    if(slot == IMG_LUMINANCE)
    {
        m_lumImage = NULL;
    }
}


}
