#include <stdexcept>
#include <stdio.h>
#include "avltree.h"

#define MIN(_a, _b) ((_a<_b)?_a:_b)
#define MAX(_a, _b) ((_a>_b)?_a:_b)
////////////////////////////////////////////////////////////////////////////////
// Construction/destruction
bilateral::Node::Node(int64_t value, int64_t squared, int64_t cubed)
    : m_left(NULL)
    , m_right(NULL)
    , m_parent(NULL)
    , m_balance(0)
    , m_value(value)
    , m_square_value(squared)
    , m_cube_value(cubed)
    , m_count(1)
{
    UpdateTotals(false);
}

bilateral::Node::~Node()
{
    if(m_left)
    {
        delete m_left;
    }
    if(m_right)
    {
        delete m_right;
    }
}

////////////////////////////////////////////////////////////////////////////////
// External interface..

// insert.
bool
bilateral::Node::Insert(bilateral::Node *node, bilateral::Node **root)
{
    bool change(false);

    if(node)
    {
        if(node->GetValue() < m_value)
        {
            if(m_left)
            {
                if(m_left->Insert(node, root))
                {
                    m_balance--;
                    if(m_balance < 0)
                    {
                        // increase in depth.
                        change = true;
                    }
                }
                UpdateTotals(false);
            }
            else
            {
                SetLeft(node);
                UpdateTotals(false);

                m_balance--;
                if(m_balance < 0)
                {
                    // increase in depth.
                    change = true;
                }
            }
        }
        else if(node->GetValue() > m_value)
        {
            if(m_right)
            {
                if(m_right->Insert(node, root))
                {
                    m_balance++;
                    if(m_balance > 0)
                    {
                        // increase in depth.
                        change = true;
                    }
                }
                UpdateTotals(false);
            }
            else
            {
                SetRight(node);
                UpdateTotals(false);

                m_balance++;
                if(m_balance > 0)
                {
                    // increase in depth.
                    change = true;
                }
            }
        }
        else
        {
            throw std::logic_error("double insert of value");
        }
    }

    if((m_balance == 2) || (m_balance == -2))
    {
        change = !Rebalance(root);
    }

    return change;
}


bool
bilateral::Node::Remove(bilateral::Node **root)
{
    Node *to_go = NULL;
    int64_t to_go_value = m_value;
    int64_t to_go_square_value;
    int64_t to_go_cube_value;
    uint32_t to_go_count;

    if(m_left)
    {
        to_go = m_left->FindMax();
    }
    else if(m_right)
    {
        to_go = m_right->FindMin();
    }

    if(to_go)
    {
        // cache data from node to be erased.
        to_go_value        = to_go->GetValue();
        to_go_square_value = to_go->GetSquareValue();
        to_go_cube_value   = to_go->GetCubeValue();
        to_go_count        = to_go->GetCount();

        (*root)->Delete(to_go, root);
        //delete to_go;

        m_value        = to_go_value;
        m_square_value = to_go_square_value;
        m_cube_value   = to_go_cube_value;
        m_count        = to_go_count;

        UpdateTotals(true);
    }
    else
    {
        (*root)->Delete(this, root);
    }

    return true;
}

bool
bilateral::Node::Delete(Node *node, Node **root)
{
    bool change(false);

    if(node->GetValue() < m_value)
    {
        // delete left
        if(m_left)
        {
            if(m_left->Delete(node, root))
            {
                m_balance++;
                if(!m_balance)
                {
                    change = true;
                }
            }
            UpdateTotals(false);
        }
        else
        {
            throw std::logic_error("delete of non member node");
        }
    }
    else if(node->GetValue() > m_value)
    {
        // delete right
        if(m_right)
        {
            if(m_right->Delete(node, root))
            {
                m_balance--;
                if(!m_balance)
                {
                    change = true;
                }
            }
            UpdateTotals(false);
        }
        else
        {
            throw std::logic_error("delete of non member node");
        }
    }
    else
    {
        // Delete this
        if(m_parent)
        {
            Node *replacement;

            if(m_left)
            {
                replacement = m_left;
            }
            else
            {
                replacement = m_right;
            }

            // replace self in tree with children, if any..
            if(m_parent->GetRight() == this)
            {
                m_parent->SetRight(replacement);
            }
            else
            {
                m_parent->SetLeft(replacement);
            }

            // flag height change.
            change = true;
        }
        else
        {
            // change root;
            if(m_left)
            {
                *root = m_left;
            }
            else
            {
                *root = m_right;
            }
            change = false;
        }
        m_left = m_right = m_parent = NULL;

        delete this;
        return change;
    }


    if((m_balance == 2) || (m_balance == -2))
    {
        change = Rebalance(root);
    }
/*
    else
    {
        if((m_left == NULL) && (m_right == NULL))
        {
            change = true;
        }
        else
        {
            change = false;
        }
    }*/

    return change;
}

bilateral::Node *
bilateral::Node::Find(int64_t value)
{
    if(value < m_value)
    {
        if(m_left)
        {
            return m_left->Find(value);
        }
        else
        {
            return NULL;
        }
    }
    else if(value > m_value)
    {
        if(m_right)
        {
            return m_right->Find(value);
        }
        else
        {
            return NULL;
        }
    }
    else
    {
        return this;
    }
}

void
bilateral::Node::GetTotals(int64_t min, int64_t max,
                           int64_t &total_values,
                           int64_t &total_squares,
                           int64_t &total_cubes,
                           uint32_t &total_count)
{
    total_values  = GetTotalValue();
    total_squares = GetTotalSquareValue();
    total_cubes   = GetTotalCubeValue();
    total_count   = GetTotalCount();

    ExcludeMin(min,
               total_values, total_squares,
               total_cubes, total_count);

    ExcludeMax(max,
               total_values, total_squares,
               total_cubes, total_count);
}

void
bilateral::Node::ExcludeMin(int64_t min,
                            int64_t &total_values,
                            int64_t &total_squares,
                            int64_t &total_cubes,
                            uint32_t &total_count)
{
    if(m_value < min)
    {
        // Subtract this.
        total_values -= (m_count * m_value);
        total_squares-= (m_count * m_square_value);
        total_cubes  -= (m_count * m_cube_value);
        total_count  -= m_count;

        // subtract left subtree.
        if(m_left)
        {
            total_values -= m_left->GetTotalValue();
            total_squares-= m_left->GetTotalSquareValue();
            total_cubes  -= m_left->GetTotalCubeValue();
            total_count  -= m_left->GetTotalCount();
        }

        // recurse right
        if(m_right)
        {
            m_right->ExcludeMin(min, total_values,
                                total_squares, total_cubes,
                                total_count);
        }
    }
    else if(m_left)
    {
        m_left->ExcludeMin(min, total_values,
                           total_squares, total_cubes,
                           total_count);
    }
}

void
bilateral::Node::ExcludeMax(int64_t max,
                            int64_t &total_values,
                            int64_t &total_squares,
                            int64_t &total_cubes,
                            uint32_t &total_count)
{
    if(m_value > max)
    {
        // Subtract this.
        total_values -= (m_count * m_value);
        total_squares-= (m_count * m_square_value);
        total_cubes  -= (m_count * m_cube_value);
        total_count  -= m_count;

        // subtract right subtree.
        if(m_right)
        {
            total_values -= m_right->GetTotalValue();
            total_squares-= m_right->GetTotalSquareValue();
            total_cubes  -= m_right->GetTotalCubeValue();
            total_count  -= m_right->GetTotalCount();
        }

        // recurse left
        if(m_left)
        {
            m_left->ExcludeMax(max, total_values,
                               total_squares, total_cubes,
                               total_count);
        }
    }
    else if(m_right)
    {
        m_right->ExcludeMax(max, total_values,
                            total_squares, total_cubes,
                            total_count);
    }
}



bilateral::Node *
bilateral::Node::GetLeft(void) const
{
    return m_left;
}

bilateral::Node *
bilateral::Node::GetRight(void) const
{
    return m_right;
}

int64_t
bilateral::Node::GetValue(void) const
{
    return m_value;
}

int64_t
bilateral::Node::GetSquareValue(void) const
{
    return m_square_value;
}

int64_t
bilateral::Node::GetCubeValue(void) const
{
    return m_cube_value;
}

uint32_t
bilateral::Node::GetCount(void) const
{
    return m_count;
}

void
bilateral::Node::SetCount(uint32_t count)
{
    m_count = count;
    UpdateTotals(true);
}

int64_t
bilateral::Node::GetTotalValue(void) const
{
    return m_total_value;
}

int64_t
bilateral::Node::GetTotalSquareValue(void) const
{
    return m_total_square_value;
}

int64_t
bilateral::Node::GetTotalCubeValue(void) const
{
    return m_total_cube_value;
}

uint32_t
bilateral::Node::GetTotalCount(void) const
{
    return m_total_count;
}

////////////////////////////////////////////////////////////////////////////////
// Internals
void
bilateral::Node::SetLeft(bilateral::Node *p)
{
    m_left = p;
    if(p)
    {
        p->SetParent(this);
    }
}

void
bilateral::Node::SetRight(bilateral::Node *p)
{
    m_right = p;
    if(p)
    {
        p->SetParent(this);
    }
}

void
bilateral::Node::SetParent(bilateral::Node *p)
{
    m_parent = p;
}

bilateral::Node *
bilateral::Node::FindRoot(void)
{
    if(m_parent)
    {
        return m_parent->FindRoot();
    }
    else
    {
        return this;
    }
}

bilateral::Node *
bilateral::Node::FindMax(void)
{
    if(m_right)
    {
        return m_right->FindMax();
    }
    else
    {
        return this;
    }
}

bilateral::Node *
bilateral::Node::FindMin(void)
{
    if(m_left)
    {
        return m_left->FindMin();
    }
    else
    {
        return this;
    }
}

void
bilateral::Node::Dump(void)
{
    putchar('\n');
    Dump(0);
    putchar('\n');
}

void
bilateral::Node::Dump(int depth)
{
    printf("(%li,%i)", m_value, m_balance);

    if(m_left)
    {
        printf("-->");
        m_left->Dump(depth + 1);
    }
    else
    {
        putchar('\n');
    }
    if(m_right)
    {
        for(int i=0;i<=depth;i++)
        {
            printf("|        ");
        }
        printf("\n");
        for(int i=0;i<depth;i++)
        {
            printf("|        ");
        }
        printf("\\----->");
        m_right->Dump(depth+1);
    }
}

bool
bilateral::Node::Check(void)
{
    bool ok(true);

    // check leaf balance
    if(!(m_left || m_right))
    {
        if(m_balance != 0)
        {
            ok = false;
        }
    }

    if(m_left && ok)
    {
        ok = m_left->Check();
    }
    if(m_right && ok)
    {
        ok = m_right->Check();
    }

    return ok;
}

// Update the total values of this current node, and propagate to parents
// if arg is true.
void
bilateral::Node::UpdateTotals(bool propagate)
{
    m_total_value        = m_value * m_count;
    m_total_square_value = m_square_value * m_count;
    m_total_cube_value   = m_cube_value * m_count;
    m_total_count        = m_count;

    if(m_left)
    {
        m_total_value        += m_left->GetTotalValue();
        m_total_square_value += m_left->GetTotalSquareValue();
        m_total_cube_value   += m_left->GetTotalCubeValue();
        m_total_count        += m_left->GetTotalCount();
    }
    if(m_right)
    {
        m_total_value        += m_right->GetTotalValue();
        m_total_square_value += m_right->GetTotalSquareValue();
        m_total_cube_value   += m_right->GetTotalCubeValue();
        m_total_count        += m_right->GetTotalCount();
    }

    if(m_parent && propagate)
    {
        m_parent->UpdateTotals(true);
    }
}

////////////////////////////////////////////////////////////////////////////////
// Tree query functions.

int32_t
bilateral::Node::GetBalance(void) const
{
    return m_balance;
}

void
bilateral::Node::SetBalance(int32_t balance)
{
    m_balance = balance;
}

bilateral::Node *
bilateral::Node::GetParent(void) const
{
    return m_parent;
}

////////////////////////////////////////////////////////////////////////////////

bool
bilateral::Node::RotateRight(bilateral::Node *root)
{
    bool change(false);
    Node *old_root = root;
    Node *parent = root->GetParent();

    change = (root->GetLeft()->GetBalance() != 0);

    // rotate
    root = root->GetLeft();
    old_root->SetLeft(root->GetRight());
    root->SetRight(old_root);

    if(parent)
    {
        if(parent->GetLeft() == old_root)
        {
            parent->SetLeft(root);
        }
        else
        {
            parent->SetRight(root);
        }
    }
    else
    {
        root->SetParent(NULL);
    }

    // Update balance.
    old_root->SetBalance(old_root->GetBalance() +
                        (1 - MIN(root->GetBalance(), 0)));
    root->SetBalance(root->GetBalance() +
                     (1 + MAX(old_root->GetBalance(), 0)));

    // Update totals.
    old_root->UpdateTotals(false);
    root->UpdateTotals(false);
    return change;
}

bool
bilateral::Node::RotateLeft(bilateral::Node *root)
{
    bool change(false);
    Node *old_root = root;
    Node *parent = root->GetParent();

    change = (root->GetRight()->GetBalance() != 0);
    // rotate
    root = root->GetRight();
    old_root->SetRight(root->GetLeft());
    root->SetLeft(old_root);

    if(parent)
    {
        if(parent->GetLeft() == old_root)
        {
            parent->SetLeft(root);
        }
        else
        {
            parent->SetRight(root);
        }
    }
    else
    {
        root->SetParent(NULL);
    }

    // Update balance.
    old_root->SetBalance(old_root->GetBalance() -
                        (1 + MAX(root->GetBalance(), 0)));
    root->SetBalance(root->GetBalance() -
                     (1 - MIN(old_root->GetBalance(), 0)));

    // Update totals.
    old_root->UpdateTotals(false);
    root->UpdateTotals(false);
    return change;
}

bool
bilateral::Node::DoubleRotateRight(bilateral::Node *root)
{
     RotateLeft(root->GetLeft());
     RotateRight(root);
     return true;
}

bool
bilateral::Node::DoubleRotateLeft(bilateral::Node *root)
{
    RotateRight(root->GetRight());
    RotateLeft(root);
    return true;
}

bool
bilateral::Node::Rebalance(bilateral::Node **root)
{
    bool change(false);

    if(m_balance == 2)
    {
        if(GetRight()->GetBalance() >= 0)
        {
            change = RotateLeft(this);
        }
        else
        {
            change = DoubleRotateLeft(this);
        }

        if(this == *root)
        {
            *root = m_parent;
        }
    }
    else if(m_balance == -2)
    {
        if(GetLeft()->GetBalance() <= 0)
        {
            change = RotateRight(this);
        }
        else
        {
            change = DoubleRotateRight(this);
        }
        if(this == *root)
        {
            *root = m_parent;
        }
    }
    return change;
}

