#include <stdexcept>
#include <stdio.h>
#include "avltree2.h"
#include <list>

#define MIN(_a, _b) ((_a<_b)?_a:_b)
#define MAX(_a, _b) ((_a>_b)?_a:_b)
namespace bilateral {

////////////////////////////////////////////////////////////////////////////////
// Internal prototypes.
bool rotate_left(avl_node *);
bool rotate_right(avl_node *);
bool double_rotate_left(avl_node *);
bool double_rotate_right(avl_node *);


avl_node *find_root(avl_node *);
avl_node *find_min(avl_node *);
avl_node *find_max(avl_node *);

void exclude_min(avl_node *node,
                 int64_t min,
                 int64_t &total_values,
                 int64_t &total_squares,
                 int64_t &total_cubes,
                 uint32_t &total_count);

void exclude_max(avl_node *node,
                 int64_t min,
                 int64_t &total_values,
                 int64_t &total_squares,
                 int64_t &total_cubes,
                 uint32_t &total_count);

bool rebalance(avl_node *node,
               avl_node **root);

void set_left(avl_node *, avl_node *);
void set_right(avl_node *, avl_node *);
void set_parent(avl_node *, avl_node *);

////////////////////////////////////////////////////////////////////////////////
// External interface.
avl_node *
create_node(allocator &alloc,
            int64_t value,
            int64_t squared,
            int64_t cubed)
{
    avl_node *new_node;

    new_node = alloc.alloc();

    if(new_node)
    {
        new_node->m_left = new_node->m_right = new_node->m_parent = NULL;
        new_node->m_balance = 0;
        new_node->m_value = value;
        new_node->m_square_value = squared;
        new_node->m_cube_value = cubed;
        new_node->m_count = 1;
        update_totals(new_node, false);
    }
    return new_node;
}

void
destroy_node(allocator &alloc,
             avl_node *to_go)
{
    if(to_go->m_left)
    {
        destroy_node(alloc, to_go->m_left);
    }
    if(to_go->m_right)
    {
        destroy_node(alloc, to_go->m_right);
    }
    alloc.free(to_go);
}

bool
insert_node(avl_node *node, avl_node *to_insert, avl_node **root)
{
    bool change(false);

    if(to_insert && node)
    {
        if(to_insert->m_value < node->m_value)
        {
            if(node->m_left)
            {
                if(insert_node(node->m_left, to_insert, root))
                {
                    node->m_balance--;
                    if(node->m_balance < 0)
                    {
                        // increase in depth.
                        change = true;
                    }
                }
                update_totals(node, false);
            }
            else
            {
                set_left(node, to_insert);
                update_totals(node, false);

                node->m_balance--;

                if(node->m_balance < 0)
                {
                    // increase in depth.
                    change = true;
                }
            }
        }
        else if(to_insert->m_value > node->m_value)
        {
            if(node->m_right)
            {
                if(insert_node(node->m_right, to_insert, root))
                {
                    node->m_balance++;
                    if(node->m_balance > 0)
                    {
                        // increase in depth.
                        change = true;
                    }
                }
                update_totals(node, false);
            }
            else
            {
                set_right(node, to_insert);
                update_totals(node, false);

                node->m_balance++;
                if(node->m_balance > 0)
                {
                    // increase in depth.
                    change = true;
                }
            }
        }
        else
        {
            throw std::logic_error("double insert of value");
        }
    }

    if((node->m_balance == 2) || (node->m_balance == -2))
    {
        change = !rebalance(node, root);
    }

    return change;
}

bool remove_node(allocator &alloc, avl_node *node, avl_node **root)
{
    avl_node *to_go = NULL;
    int64_t to_go_value = node->m_value;
    int64_t to_go_square_value;
    int64_t to_go_cube_value;
    uint32_t to_go_count;

    if(node->m_left)
    {
        to_go = find_max(node->m_left);
    }
    else if(node->m_right)
    {
        to_go = find_min(node->m_right);
    }

    if(to_go)
    {
        // cache data from node to be erased.
        to_go_value        = to_go->m_value;
        to_go_square_value = to_go->m_square_value;
        to_go_cube_value   = to_go->m_cube_value;
        to_go_count        = to_go->m_count;

        delete_node(alloc, *root, to_go, root);

        node->m_value        = to_go_value;
        node->m_square_value = to_go_square_value;
        node->m_cube_value   = to_go_cube_value;
        node->m_count        = to_go_count;

        update_totals(node, true);
    }
    else
    {
        delete_node(alloc, *root, node, root);
    }

    return true;
}

bool
delete_node(allocator &alloc, avl_node *node, avl_node *to_go, avl_node **root)
{
    bool change(false);

    if(to_go->m_value < node->m_value)
    {
        // delete left
        if(node->m_left)
        {
            if(delete_node(alloc, node->m_left, to_go, root))
            {
                node->m_balance++;
                if(!node->m_balance)
                {
                    change = true;
                }
            }
            update_totals(node, false);
        }
        else
        {
            throw std::logic_error("delete of non member to_go");
        }
    }
    else if(to_go->m_value > node->m_value)
    {
        // delete right
        if(node->m_right)
        {
            if(delete_node(alloc, node->m_right, to_go, root))
            {
                node->m_balance--;
                if(!node->m_balance)
                {
                    change = true;
                }
            }
            update_totals(node, false);
        }
        else
        {
            throw std::logic_error("delete of non member to_go");
        }
    }
    else
    {
        // Delete this
        if(node->m_parent)
        {
            avl_node *replacement;

            if(node->m_left)
            {
                replacement = node->m_left;
            }
            else
            {
                replacement = node->m_right;
            }

            // replace self in tree with children, if any..
            if(node->m_parent->m_right == node)
            {
                set_right(node->m_parent, replacement);
            }
            else
            {
                set_left(node->m_parent, replacement);
            }

            // flag height change.
            change = true;
        }
        else
        {
            // change root;
            if(node->m_left)
            {
                *root = node->m_left;
            }
            else
            {
                *root = node->m_right;
            }
            change = false;
        }
        node->m_left = node->m_right = node->m_parent = NULL;

        destroy_node(alloc, node);
        return change;
    }


    if((node->m_balance == 2) || (node->m_balance == -2))
    {
        change = rebalance(node, root);
    }

    return change;
}

avl_node *find_node(avl_node *node, int64_t key)
{
    avl_node *result(node);

    while(result && (key != result->m_value))
    {
        if(key < result->m_value)
        {
            result = result->m_left;
        }
        else
        {
            result = result->m_right;
        }
    }

    return result;
}

void get_totals(avl_node *node,
                int64_t min, int64_t max,
                int64_t &total_values,
                int64_t &total_squares,
                int64_t &total_cubes,
                uint32_t &total_count)
{
    total_values  = node->m_total_value;
    total_squares = node->m_total_square_value;
    total_cubes   = node->m_total_cube_value;
    total_count   = node->m_total_count;

    exclude_min(node,
                min,
                total_values, total_squares,
                total_cubes, total_count);

    exclude_max(node,
                max,
                total_values, total_squares,
                total_cubes, total_count);
}

// Update the total values of this current node, and propagate to parents
// if arg is true.
void update_totals(avl_node *node, bool propagate)
{
    do 
    {
        node->m_total_value        = node->m_value * node->m_count;
        node->m_total_square_value = node->m_square_value * node->m_count;
        node->m_total_cube_value   = node->m_cube_value * node->m_count;
        node->m_total_count        = node->m_count;

        if(node->m_left)
        {
            node->m_total_value        += node->m_left->m_total_value;
            node->m_total_square_value += node->m_left->m_total_square_value;
            node->m_total_cube_value   += node->m_left->m_total_cube_value;
            node->m_total_count        += node->m_left->m_total_count;
        }
        if(node->m_right)
        {
            node->m_total_value        += node->m_right->m_total_value;
            node->m_total_square_value += node->m_right->m_total_square_value;
            node->m_total_cube_value   += node->m_right->m_total_cube_value;
            node->m_total_count        += node->m_right->m_total_count;
        }
        node = node->m_parent;
    } while(propagate && node);
}

////////////////////////////////////////////////////////////////////////////////
// Internal bits.

void exclude_min(avl_node *node,
                 int64_t min,
                 int64_t &total_values,
                 int64_t &total_squares,
                 int64_t &total_cubes,
                 uint32_t &total_count)

{
#ifndef RECURSIVE_EXCLUDE
    while(node)
    {
        if(node->m_value < min)
        {
            // Subtract this.
            total_values -= (node->m_count * node->m_value);
            total_squares-= (node->m_count * node->m_square_value);
            total_cubes  -= (node->m_count * node->m_cube_value);
            total_count  -= node->m_count;

            // subtract left subtree.
            if(node->m_left)
            {
                total_values -= node->m_left->m_total_value;
                total_squares-= node->m_left->m_total_square_value;
                total_cubes  -= node->m_left->m_total_cube_value;
                total_count  -= node->m_left->m_total_count;
            }

            node = node->m_right;
        }
        else
        {
            node = node->m_left;
        }
    }
#else
    if(node->m_value < min)
    {
        // Subtract this.
        total_values -= (node->m_count * node->m_value);
        total_squares-= (node->m_count * node->m_square_value);
        total_cubes  -= (node->m_count * node->m_cube_value);
        total_count  -= node->m_count;

        // subtract left subtree.
        if(node->m_left)
        {
            total_values -= node->m_left->m_total_value;
            total_squares-= node->m_left->m_total_square_value;
            total_cubes  -= node->m_left->m_total_cube_value;
            total_count  -= node->m_left->m_total_count;
        }

        // recurse right
        if(node->m_right)
        {
            exclude_min(node->m_right, 
                        min, total_values,
                        total_squares, total_cubes,
                        total_count);
        }
    }
    else if(node->m_left)
    {
        exclude_min(node->m_left, 
                    min, total_values,
                    total_squares, total_cubes,
                    total_count);
    }
#endif
}

void exclude_max(avl_node *node,
                 int64_t max,
                 int64_t &total_values,
                 int64_t &total_squares,
                 int64_t &total_cubes,
                 uint32_t &total_count)
{
#ifndef RECURSIVE_EXCLUDE
    while(node)
    {
        if(node->m_value > max)
        {
            // Subtract this.
            total_values -= (node->m_count * node->m_value);
            total_squares-= (node->m_count * node->m_square_value);
            total_cubes  -= (node->m_count * node->m_cube_value);
            total_count  -= node->m_count;

            // subtract left subtree.
            if(node->m_right)
            {
                total_values -= node->m_right->m_total_value;
                total_squares-= node->m_right->m_total_square_value;
                total_cubes  -= node->m_right->m_total_cube_value;
                total_count  -= node->m_right->m_total_count;
            }

            node = node->m_left;
        }
        else
        {
            node = node->m_right;
        }
    }
#else
    if(node->m_value > max)
    {
        // Subtract this.
        total_values -= (node->m_count * node->m_value);
        total_squares-= (node->m_count * node->m_square_value);
        total_cubes  -= (node->m_count * node->m_cube_value);
        total_count  -= node->m_count;

        // subtract right subtree.
        if(node->m_right)
        {
            total_values -= node->m_right->m_total_value;
            total_squares-= node->m_right->m_total_square_value;
            total_cubes  -= node->m_right->m_total_cube_value;
            total_count  -= node->m_right->m_total_count;
        }

        // recurse left
        if(node->m_left)
        {
            exclude_max(node->m_left,
                        max, total_values,
                        total_squares, total_cubes,
                        total_count);
        }
    }
    else if(node->m_right)
    {
        exclude_max(node->m_right,
                    max, total_values,
                    total_squares, total_cubes,
                    total_count);
    }
#endif
}

avl_node *find_root(avl_node *node)
{
    while(node->m_parent)
    {
        node = node->m_parent;
    }

    return node;
}

avl_node *find_min(avl_node *node)
{
    while(node->m_left)
    {
        node = node->m_left;
    }
    return node;
}

avl_node *find_max(avl_node *node)
{
    while(node->m_right)
    {
        node = node->m_right;
    }
    return node;
}

/*
void
bilateral::Node::Dump(void)
{
    putchar('\n');
    Dump(0);
    putchar('\n');
}

void
bilateral::Node::Dump(int depth)
{
    printf("(%li,%i)", m_value, m_balance);

    if(m_left)
    {
        printf("-->");
        m_left->Dump(depth + 1);
    }
    else
    {
        putchar('\n');
    }
    if(m_right)
    {
        for(int i=0;i<=depth;i++)
        {
            printf("|        ");
        }
        printf("\n");
        for(int i=0;i<depth;i++)
        {
            printf("|        ");
        }
        printf("\\----->");
        m_right->Dump(depth+1);
    }
}

bool
bilateral::Node::Check(void)
{
    bool ok(true);

    // check leaf balance
    if(!(m_left || m_right))
    {
        if(m_balance != 0)
        {
            ok = false;
        }
    }

    if(m_left && ok)
    {
        ok = m_left->Check();
    }
    if(m_right && ok)
    {
        ok = m_right->Check();
    }

    return ok;
}
*/

bool rotate_left(avl_node *root)
{
    bool change(false);
    avl_node *old_root = root;
    avl_node *parent = root->m_parent;

    change = (root->m_right->m_balance != 0);
    // rotate
    root = root->m_right;
    set_right(old_root, root->m_left);
    set_left(root, old_root);

    if(parent)
    {
        if(parent->m_left == old_root)
        {
            set_left(parent, root);
        }
        else
        {
            set_right(parent, root);
        }
    }
    else
    {
        set_parent(root, NULL);
    }

    // Update balance.
    old_root->m_balance = (old_root->m_balance -
                           (1 + MAX(root->m_balance, 0)));
    root->m_balance = (root->m_balance -
                       (1 - MIN(old_root->m_balance, 0)));

    // Update totals.
    update_totals(old_root, false);
    update_totals(root, false);
    return change;
}

bool rotate_right(avl_node *root)
{
    bool change(false);
    avl_node *old_root = root;
    avl_node *parent = root->m_parent;

    change = (root->m_left->m_balance != 0);
    // rotate
    root = root->m_left;
    set_left(old_root, root->m_right);
    set_right(root, old_root);

    if(parent)
    {
        if(parent->m_left == old_root)
        {
            set_left(parent, root);
        }
        else
        {
            set_right(parent, root);
        }
    }
    else
    {
        set_parent(root, NULL);
    }

    // Update balance.
    old_root->m_balance = (old_root->m_balance +
                        (1 - MIN(root->m_balance, 0)));
    root->m_balance = (root->m_balance +
                     (1 + MAX(old_root->m_balance, 0)));

    // Update totals.
    update_totals(old_root, false);
    update_totals(root, false);
    return change;
}

bool double_rotate_left(avl_node *root)
{
    rotate_right(root->m_right);
    rotate_left(root);
    return true;
}

bool double_rotate_right(avl_node *root)
{
    rotate_left(root->m_left);
    rotate_right(root);
    return true;
}

bool
rebalance(avl_node *node, avl_node **root)
{
    bool change(false);

    if(node->m_balance == 2)
    {
        if(node->m_right->m_balance >= 0)
        {
            change = rotate_left(node);
        }
        else
        {
            change = double_rotate_left(node);
        }

        if(node == *root)
        {
            *root = node->m_parent;
        }
    }
    else if(node->m_balance == -2)
    {
        if(node->m_left->m_balance <= 0)
        {
            change = rotate_right(node);
        }
        else
        {
            change = double_rotate_right(node);
        }
        if(node == *root)
        {
            *root = node->m_parent;
        }
    }
    return change;
}

void set_left(avl_node *node, avl_node *left)
{
    node->m_left = left;

    if(left)
    {
        left->m_parent = node;
    }
}

void set_right(avl_node *node, avl_node *right)
{

    node->m_right = right;

    if(right)
    {
        right->m_parent = node;
    }
}

void set_parent(avl_node *node, avl_node *parent)
{
    node->m_parent = parent;
}

}

