#include <stdio.h>
#include <stdlib.h>
#include <math.h>
//#include <assert.h>

#include <vector>
#include <set>

#include "image.h"
#include "processor.h"
#include "histogram.h"
#include "filter.h"
#include "thread.h"

#define TILE_WIDTH  512
#define TILE_HEIGHT 1024

////////////////////////////////////////////////////////////////////////////////
// Farmer class.

namespace 
{

class BilateralFarmer : public thread::Farmer
{
public:
    BilateralFarmer(spectral::Processor *owner, float min, float max)
    : m_owner(owner)
    , m_min_progress(min)
    , m_max_progress(max)
    {}

    virtual void ProcessCompletedJob(thread::Job *);

private:
    spectral::Processor *m_owner;
    float m_min_progress, m_max_progress;
};


void
BilateralFarmer::ProcessCompletedJob(thread::Job *)
{
    float progress = GetNumCompleted() + 1;
    progress/= GetNumSubmitted();
    progress*= (m_max_progress - m_min_progress);
    progress+= m_min_progress;

    if(m_owner)
    {
        m_owner->Progress(progress);
    }
}

class BilateralTileJob : public thread::Job
{
public:
    BilateralTileJob(const spectral::Image *reference,
                     const spectral::Image *input,
                     spectral::Image *output,
                     int32_t radius,
                     bool wrap_horizontal,
                     bool wrap_vertical,
                     int32_t xmin,
                     int32_t ymin,
                     int32_t xmax,
                     int32_t ymax)
        : m_reference(reference)
        , m_input(input)
        , m_output(output)
        , m_radius(radius)
        , m_wrap_horizontal(wrap_horizontal)
        , m_wrap_vertical(wrap_vertical)
        , m_xmin(xmin)
        , m_ymin(ymin)
        , m_xmax(xmax)
        , m_ymax(ymax)
    {
    }

    virtual ~BilateralTileJob()
    {
    }

    virtual void Process(void)
    {
        bilateral::BoxHistogram hist(TILE_WIDTH, (m_radius * 2 + 1)); 
        const float *ref_buf;
        float *out_buf;
        size_t width;

        int32_t hist_xmin, hist_xmax;
        int32_t hist_ymin, hist_ymax, cur_y;
    
        hist_xmin = m_xmin - m_radius;
        hist_xmax = m_xmax + m_radius + 1;
    
        hist_ymin = m_ymin - m_radius;
        hist_ymax = m_ymin + m_radius + 1;

    
        // setup initial state of histogram.

        hist.Clear();
        {
            int32_t x, y, i;
            for(y=hist_ymin;y<hist_ymax;y++)
            {
                i = 0;
                for(x=hist_xmin;x<hist_xmax;x++)
                {
                    hist.AddValue(i, 
                                  *(m_input->get_constrained_pixel(
                                        x, y,
                                        m_wrap_horizontal,
                                        m_wrap_vertical)));
                    i++;
                }
            }
        }
        // Iterate.
    
        cur_y = m_ymin;

        width = m_output->get_width();
        ref_buf = m_reference->get_buffer();
        out_buf = m_output->get_buffer();

        size_t idx = (m_xmin + (m_ymin * width));

        while(cur_y < m_ymax)
        {
            int32_t x, i;
            i = 0;
            for(x = m_xmin;x < m_xmax;x++)
            {
                float cur_val;
                    
                cur_val = ref_buf[idx];
    
                float cur_sq_val = (cur_val * cur_val);
                float total, sq_total, cube_total, total_weight, total_value;
                uint32_t count;
    
                hist.GetSums(uint32_t(i), uint32_t(i + (2.0 * m_radius) + 1), 
                             (cur_val - 1.0), (cur_val + 1.0), 
                             total, sq_total, cube_total, count);
    
                total_weight = (float)count * (1.0 - cur_sq_val);
                total_weight+= (2.0 * cur_val * total);
                total_weight-= sq_total;
    
                total_value = (1.0 - cur_sq_val) * total;
                total_value+= (2.0 * cur_val * sq_total);
                total_value-= cube_total;
    
                total_value/= total_weight;
                out_buf[idx] = total_value; 
                
                idx++;
                i++;
            }
    
            // move down.            
    
            {
                int32_t xx;
                i = 0;
                for(xx=hist_xmin;xx<hist_xmax;xx++)
                {
                    hist.RemoveValue(i, 
                                     *(m_input->get_constrained_pixel(
                                           xx, hist_ymin,
                                           m_wrap_horizontal,
                                           m_wrap_vertical)));
                    hist.AddValue(i, 
                                  *(m_input->get_constrained_pixel(
                                        xx, hist_ymax,
                                        m_wrap_horizontal,
                                        m_wrap_vertical)));
                    i++;
                }
            }
            hist_ymin++;
            hist_ymax++;
            cur_y++;
            idx = (m_xmin + (cur_y * width));
        }
    }
private:
    const spectral::Image *m_reference, *m_input;
    spectral::Image *m_output;
    int32_t m_radius;
    bool m_wrap_horizontal, m_wrap_vertical;
    int32_t m_xmin, m_ymin, m_xmax, m_ymax;
};

}

void
bilateral::Filter(spectral::Processor *processor,
                  float min_progress,
                  float max_progress,
                  const spectral::Image *reference,
                  const spectral::Image *input,
                  int32_t radius,
                  bool wrap_horizontal,
                  bool wrap_vertical,
                  spectral::Image *result)
{
    uint32_t strip_size(TILE_WIDTH - ((uint32_t(radius) + 1) * 2));

    BilateralFarmer farmer(processor, min_progress, max_progress);

    uint32_t xmin(0), xmax(0);

    thread::Job *job;

    do
    {
        uint32_t ymin, ymax;
        // Set up strip to cover.
        xmin = xmax;
        xmax = xmin + strip_size;

        if(xmax > input->get_width())
        {
            xmax = input->get_width();
        }
        ymin = ymax = 0;

        do
        {
            ymin = ymax;
            ymax = ymin + TILE_HEIGHT;

            if(ymax > input->get_height())
            {
                ymax = input->get_height();
            }
           
            job = new BilateralTileJob(reference, input, result, radius, wrap_horizontal, wrap_vertical, xmin, ymin, xmax, ymax);
            farmer.Push(job);
        } while(ymax < input->get_height());
    } while(xmax < input->get_width());
    farmer.Process();
}

