#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "utils.h"
#include "histogram.h"

const float mul_factor       = 1000000;

#ifdef USE_ALLOCATOR
bilateral::SubHistogram::SubHistogram()
    : m_allocator(NULL)
    , m_tree(NULL)
{
}

#else
bilateral::SubHistogram::SubHistogram()
    : m_tree(NULL)
{
}
#endif

bilateral::SubHistogram::~SubHistogram()
{
    Clear();
}

#ifdef USE_ALLOCATOR
void
bilateral::SubHistogram::SetAllocator(allocator *p)
{
    m_allocator = p;
}
#endif

bool 
bilateral::SubHistogram::AddValue(int64_t val,
                                  int64_t square,
                                  int64_t cube)
{
#ifdef USE_ALLOCATOR
    bool result = false;

    if(m_tree)
    {
        avl_node *node = find_node(m_tree, val);

        if(node)
        {
            node->m_count++;
            update_totals(node, true);
        }
        else
        {
            avl_node *new_node = create_node(*m_allocator, val, square, cube);

            insert_node(m_tree, new_node, &m_tree);
        }
        result = true;
    }
    else
    {
        m_tree = create_node(*m_allocator, val, square, cube);
        result = true;
    }

    return result;
#else
    bool result = false;

    if(m_tree)
    {
        Node *node = m_tree->Find(val);

        if(node)
        {
            uint32_t count = node->GetCount();
            node->SetCount(count + 1);
        }
        else
        {
            Node *new_node = new Node(val, square, cube);

            m_tree->Insert(new_node, &m_tree);
        }
        result = true;
    }
    else
    {
        m_tree = new Node(val,
                          square,
                          cube);
        result = true;
    }

    return result;
#endif
}

bool
bilateral::SubHistogram::RemoveValue(int64_t val)
{
#ifdef USE_ALLOCATOR
    bool result = false;
    if(m_tree)
    {
        avl_node *to_go = find_node(m_tree, val);

        if(to_go)
        {
            if(to_go->m_count < 2)
            {
                remove_node(*m_allocator, to_go, &m_tree);
            }
            else
            {
                to_go->m_count--;
                update_totals(to_go, true);
            }
            result = true;
        }
    }

    return result;
#else
    bool result = false;
    if(m_tree)
    {
        Node *to_go = m_tree->Find(val);

        if(to_go)
        {
            uint32_t count = to_go->GetCount();

            if(count < 2)
            {
                to_go->Remove(&m_tree);
            }
            else
            {
                to_go->SetCount(count - 1);
            }
            result = true;
        }
    }

    return result;
#endif
}

void bilateral::SubHistogram::Clear()
{
#ifdef USE_ALLOCATOR
    if(m_tree)
    {
        destroy_node(*m_allocator, m_tree);
    }
    m_tree = NULL;
#else
    if(m_tree)
    {
        delete m_tree;
    }
    m_tree = NULL;
#endif
}

void
bilateral::SubHistogram::GetSums(int64_t min,
                                 int64_t max, 
                                 int64_t &sum, 
                                 int64_t &sum_squares,
                                 int64_t &sum_cubes,
                                 uint32_t &count) const
{
#ifdef USE_ALLOCATOR
    if(m_tree)
    {
        get_totals(m_tree, min, max, sum, sum_squares, sum_cubes, count);
    }
    else
    {
        sum = 0;
        sum_squares = 0;
        sum_cubes = 0;
        count = 0;
    }
#else
    if(m_tree)
    {
        m_tree->GetTotals(min, max, sum, sum_squares, sum_cubes, count);
    }
    else
    {
        sum = 0;
        sum_squares = 0;
        sum_cubes = 0;
        count = 0;
    }
#endif
}

bilateral::BoxHistogram::BoxHistogram(uint32_t width,
                                      uint32_t diameter)
    : m_width(width)
    , m_use_layers(0)
#ifdef USE_ALLOCATOR
    , m_allocator(4096)
#endif
{
    for(uint32_t i=0;i<NUM_LAYERS;i++)
    {
        int32_t layer_size;
        layer_size = div((int32_t)width, (int32_t)GetLayerSize(i)).quot + 1;

        // The i+1 term is there to reduce the number of layers
        // as insert/remove is so much more expensive than lookup.
        if((GetLayerSize(i+1) <= diameter) || (i == 0))
        {
            m_use_layers = i + 1;
            m_count[i] = layer_size;
            m_masks[i] = GetLayerSize(i) - 1;
            m_layers[i] = new SubHistogram[layer_size];
#ifdef USE_ALLOCATOR
            for(size_t j=0;j<size_t(layer_size);j++)
            {
                m_layers[i][j].SetAllocator(&m_allocator);
            }
#endif
        }
        else
        {
            m_count[i] = 0;
            m_masks[i] = 0;
            m_layers[i] = NULL;
        }
    }
}

bilateral::BoxHistogram::~BoxHistogram()
{
    for(uint32_t i=0;i<NUM_LAYERS;i++)
    {
        if(m_layers[i])
        {
            delete[] m_layers[i];
        }
    }
}

bool
bilateral::BoxHistogram::AddValue(uint32_t x, float val)
{
    if(x < m_width)
    {
        for(uint32_t i=0;i<m_use_layers;i++)
        {
            uint32_t xx;
            int64_t i_val, i_square, i_cube;
            xx = x >> (i * 2);

            i_val = round_float(val * mul_factor);
            i_square= round_float(val * val * mul_factor);
            i_cube= round_float(val * val * val * mul_factor);

            m_layers[i][xx].AddValue(i_val, i_square, i_cube);
        }
        return true;
    }
    else
    {
        return false;
    }
}

bool 
bilateral::BoxHistogram::RemoveValue(uint32_t x, float val)
{
    if(x < m_width)
    {
        for(uint32_t i=0;i<m_use_layers;i++)
        {
            uint32_t xx;
            int64_t i_val;
            i_val = round_float(val * mul_factor);
            xx = x >> (i * 2);
            m_layers[i][xx].RemoveValue(i_val);
        }
        return true;
    }
    else
    {
        return false;
    }
}

void
bilateral::BoxHistogram::Clear(void)
{
    
    for(uint32_t i=0;i<m_use_layers;i++)
    {
        for(uint32_t j=0;j<m_count[i];j++)
        {
            m_layers[i][j].Clear();
        }
    }
}

void 
bilateral::BoxHistogram::GetSums(uint32_t xmin, uint32_t xmax,
                                 float min, float max,
                                 float &sum,
                                 float &sum_squares,
                                 float &sum_cubes,
                                 uint32_t &count) const
{
    sum = sum_squares = sum_cubes = 0;
    count = 0;
    uint32_t x = xmin;
    uint32_t step_size(1);
    uint32_t layer(0);

    // Find starting level.
    //printf("xmin = %i, xmax = %i\n", xmin, xmax);
    while(x < xmax) 
    {

        if(layer && ((x + step_size) > xmax))
        {
            // go down a layer
            layer--;
            step_size>>= 2;
            //printf("moving down to layer %i\n", layer);
        }
        else if((layer < (m_use_layers - 1)) &&
                ((x & m_masks[layer + 1]) == 0) &&
                ((x + (step_size * 4)) <= xmax))
        {
            // go up a layer.
           layer++;
           step_size<<= 2;
           //printf("moving up to layer %i\n", layer);
        }
        else
        {
            // Process.
            uint32_t index = x >> (layer * 2);
            int64_t tmp_sum, tmp_sum_squares, tmp_sum_cubes, i_min, i_max;
            uint32_t tmp_count;

            i_min = round_float(min * mul_factor);
            i_max = round_float(max * mul_factor);

            m_layers[layer][index].GetSums(i_min, i_max, 
                                           tmp_sum, tmp_sum_squares, 
                                           tmp_sum_cubes, tmp_count);

            sum+=         (float(tmp_sum)/mul_factor);
            sum_squares+= (float(tmp_sum_squares)/mul_factor);
            sum_cubes+=   (float(tmp_sum_cubes)/mul_factor);
            count+=       tmp_count;

            x+= step_size;
        }
    } 
}


uint32_t
bilateral::BoxHistogram::GetLayerSize(uint32_t layer) const
{
    return 1 << (layer * 2);
}

