#include <string>
#include <stdlib.h>

#include "utils.h"
#include "parameters.h"
#include "compressor.h"

const uint32_t COMPRESS_ASHIKHMIN = 0;
const uint32_t COMPRESS_LOG       = 1;
const uint32_t COMPRESS_GAMMA     = 2;
const uint32_t COMPRESS_LINEAR    = 3;
const uint32_t COMPRESS_EXPONENT  = 4;

// spectral::Image slots.
const uint32_t IMG_PICTURE          = 0;
const uint32_t IMG_LOCAL_ADAPTATION = 1;

static int compare_floats(const void *p1, const void *p2)
{
    const float *f1 = (const float *)p1;
    const float *f2 = (const float *)p2;

    if(*f1 > *f2)
    {
        return 1;
    }
    else if(*f1 < *f2)
    {
        return -1;
    }
    else
    {
        return 0;
    }
}

namespace compressor {

Compressor::Compressor()
    : spectral::Processor()
    , m_picture(NULL)
    , m_lal(NULL)
    , m_min_luminance(1e-5)
    , m_max_luminance(1e10)
    , m_gamma(0.6)
    , m_contrast(1.0)
    , m_compression_function(compress_gamma)
    , m_discard_dark(5)
    , m_discard_bright(5)
    , m_normalise(true)
{
}

Compressor::~Compressor()
{
}

// Get slots & output type.  Returns output colour space.
// pairs are <slot id, colour space>
uint32_t
Compressor::GetInterface(
    std::list<std::pair<uint32_t, uint32_t> > &mandatory,
    std::list<std::pair<uint32_t, uint32_t> > &optional) const
{
    mandatory.clear();
    optional.clear();

    mandatory.push_back(
        std::pair<uint32_t, uint32_t>(IMG_PICTURE, CLS_CIE_XYY));

    optional.push_back(
        std::pair<uint32_t, uint32_t>(IMG_LOCAL_ADAPTATION, CLS_LUMINANCE));

    return CLS_CIE_XYY;
}

bool
Compressor::SetImage(uint32_t type, const spectral::Image *img)
{
    bool result = false;

    if(type == IMG_LOCAL_ADAPTATION)
    {
        m_lal = NULL;
        if(img)
        {
            if(img->get_colour_space() == CLS_LUMINANCE)
            {
                m_lal = img;
                result = true;
            }
        }
        else
        {
            result = true;
        }
    }
    if(type == IMG_PICTURE)
    {
        m_picture = NULL;
        if(img)
        {
            if(img->get_colour_space() == CLS_CIE_XYY)
            {
                m_picture = img;
                result = true;
            }
        }
        else
        {
            result = true;
        }
    }

    return result;
}


void
Compressor::ClearImage(uint32_t type)
{
    if(type == IMG_PICTURE)
    {
        if(m_picture)
        {
            m_picture = NULL;
        }
    }
    else if(type == IMG_LOCAL_ADAPTATION)
    {
        if(m_lal)
        {
            m_lal = NULL;
        }
    }
}

void
Compressor::GetParameters(spectral::Parameters *parameters) const
{
    if(parameters)
    {
        std::string tmp;


        tmp = "min luminance";
        parameters->AddParameter(tmp, 
                                 1e-5, 1e20, m_min_luminance);

        tmp = "max luminance";
        parameters->AddParameter(tmp, 
                                 1e-5, 1e10, m_max_luminance);


        tmp = "discard dark";
        parameters->AddParameter(tmp, 
                                 0, 50, m_discard_dark);

        tmp = "discard bright";
        parameters->AddParameter(tmp, 
                                 0, 50, m_discard_bright);

        tmp = "contrast";
        parameters->AddParameter(tmp,
                                 0.0, 20.0, m_contrast);

        tmp = "gamma";
        parameters->AddParameter(tmp,
                                 0.0, 20.0, m_gamma);

        {
            uint32_t function = COMPRESS_ASHIKHMIN;
            tmp = "compression function";
            std::list<std::pair<uint32_t, std::string> > options;

            options.push_back(
                std::pair<uint32_t, std::string>
                    (COMPRESS_ASHIKHMIN, "ashikhmin"));
            options.push_back(
                std::pair<uint32_t, std::string>
                    (COMPRESS_LOG, "logarithmic"));
            options.push_back(
                std::pair<uint32_t, std::string>
                    (COMPRESS_GAMMA, "gamma"));
            options.push_back(
                std::pair<uint32_t, std::string>
                    (COMPRESS_LINEAR, "linear"));

            if(m_compression_function == compress_log)
            {
                function = COMPRESS_LOG;
            }
            else if(m_compression_function == compress_gamma)
            {
                function = COMPRESS_GAMMA;
            }
            else if(m_compression_function == compress_linear)
            {
                function = COMPRESS_LINEAR;
            }

            parameters->AddParameter(tmp, options, function);
        }

        tmp = "normalise";
        parameters->AddParameter(tmp, m_normalise);
    }
}

void 
Compressor::SetParameters(const spectral::Parameters *parameters)
{
    if(parameters)
    {
        std::string tmp;

        tmp = "min_luminance";
        parameters->GetParameter(tmp, m_min_luminance);

        tmp = "max_luminance";
        parameters->GetParameter(tmp, m_max_luminance);

        tmp = "discard dark";
        parameters->GetParameter(tmp, m_discard_dark);

        tmp = "discard bright";
        parameters->GetParameter(tmp, m_discard_bright);

        tmp = "contrast";
        parameters->GetParameter(tmp, m_contrast);
 
        tmp = "gamma";
        parameters->GetParameter(tmp, m_gamma);

        {
            tmp = "compression function";
            std::string tmp2;
            uint32_t compressor;

            parameters->GetParameter(tmp, compressor, tmp2);

            switch(compressor)
            {
            case COMPRESS_ASHIKHMIN:
                m_compression_function = compress_ashikhmin;
                break;
            case COMPRESS_EXPONENT:
                m_compression_function = compress_exponent;
                break;
            case COMPRESS_LOG:
                m_compression_function = compress_log;
                break;
            case COMPRESS_GAMMA:
                m_compression_function = compress_gamma;
                break;
            case COMPRESS_LINEAR:
                m_compression_function = compress_linear;
                break;
            }
        }

        tmp = "normalise";
        parameters->GetParameter(tmp, m_normalise);

    }
}

spectral::Image *
Compressor::Process(void)
{
    spectral::Image *result = NULL;
    Progress(0.0);

    /*
    fprintf(stderr, "m_min_luminance = %f\n", m_min_luminance);
    fprintf(stderr, "m_max_luminance = %f\n", m_max_luminance);
    fprintf(stderr, "m_gamma = %f\n", m_gamma);
    fprintf(stderr, "m_contrast = %f\n", m_contrast);
    fprintf(stderr, "m_normalise = %i\n", m_normalise);
    */

    if(m_picture)
    {
        result = new spectral::Image(*m_picture);
    }

    if(result)
    {
        bool use_lal(false);
        size_t size = size_t(result->get_width()) *
                      size_t(result->get_height());
        float *buf = result->get_buffer();
        float min(m_max_luminance), 
              max(m_min_luminance);
        float cmin, cmax;
        size_t i;
        min = max = buf[2];

        // Find scene minimum and maximum luminances.
        for(i=0;i<size;i++)
        {
            size_t index;
            index = (i * 3) + 2; // cie xyY luminance channel.

            if((buf[index] < min)  && (buf[index] > m_min_luminance))
            {
                min = buf[index];
            }
            if((buf[index] > max) && (buf[index] < m_max_luminance))
            {
                max = buf[index];
            }
        }

        // Reset min/max to discard the darkest & brightest X%
        if(m_discard_dark || m_discard_bright)
        {
            float tmp[10000]; // 100 * 100 image.

            uint32_t width = m_picture->get_width();
            uint32_t height = m_picture->get_height();

            // Sample the luminance channel.
            for(i=0;i<10000;i++)
            {
                float fx, fy;
                uint32_t x, y;

                fx = halton(5, i);
                fy = halton(7, i);

                fx*= width;
                fy*= height;

                x = round_float(fx);
                y = round_float(fy);

                if(x >= width)
                {
                    x = width - 1;
                } 
                if(y >= height)
                {
                    y = height - 1;
                } 
                tmp[i] = m_picture->get_pixel(x, y)[2];
            }
            // Sort.
            qsort(tmp, 10000, sizeof(float), compare_floats); 

            if(m_discard_dark)
            {
                uint32_t index = m_discard_dark * 100;
                if(index < 10000)
                {
                    min = tmp[index];
                }
            }
            if(m_discard_bright)
            {
                uint32_t index = m_discard_bright * 100;
                if(index < 10000)
                {
                    max = tmp[10000 - index];
                }
            }
        }

        if(min < m_min_luminance) min = m_min_luminance;
        if(max > m_max_luminance) max = m_max_luminance;
        //fprintf(stderr, "Min = %f, Max = %f\n\r", min, max);

        cmin = m_compression_function(min, &m_gamma);
        cmax = m_compression_function(max, &m_gamma);

        Progress(25);

        if(m_lal)
        {
            // does local adaptation image match real image?
            if((m_lal->get_width() == m_picture->get_width()) &&
               (m_lal->get_height() == m_picture->get_height()))
            {
                use_lal = true;
            }
        }
        if(use_lal)
        {
            float *lal_buf = m_lal->get_buffer();
            for(i=0;i<size;i++)
            {
                size_t index;
		float Y, Lal;

                index = (i * 3) + 2; // cie xyY luminance channel.

                Y = buf[index];
                Lal = lal_buf[i];

                if(Y > max) Y = max;
		if(Y < min) Y = min;

		if(Lal > max) Lal = max;
		if(Lal < min) Lal = min;

                buf[index] = local_tone_map(m_compression_function,
                                            &m_gamma,
                                            cmin, cmax, 1.0, 
                                            Y, Lal, m_contrast);
            }
        }
        else
        {
            for(i=0;i<size;i++)
            {
	        float Y;
                size_t index;
                float value;
                index = (i * 3) + 2; // cie xyY luminance channel.

                Y = buf[index];
/*
		if(Y > max) Y = max;
		if(Y < min) Y = min;
*/
                value = tone_map(m_compression_function,
                                 &m_gamma,
                                 cmin, cmax, 1.0, 
                                 Y);
                 if(value < 0.0)
		 {
		     value = 0.0;
		 }
		 buf[index] = value;
            }
        }

        Progress(50);

        if(m_normalise)
        {
            // linear scale down to ymax = 1.0.  Prevents burnout due to added
            // contrast.
            float post_min, post_max;

            post_min = post_max = buf[2];

            for(i=0;i<size;i++)
            {
                size_t index;
                index = (i * 3) + 2; // cie xyY luminance channel.
                if(buf[index] < post_min)
                {
                    post_min = buf[index];
                }
                if(buf[index] > post_max)
                {
                    post_max = buf[index];
                }
            }
            for(i=0;i<size;i++)
            {
                size_t index;
                index = (i * 3) + 2; // cie xyY luminance channel.
                buf[index] = (buf[index] - post_min) / (post_max - post_min);
            }
        }
        Progress(100);
    }

    return result;
}

float 
Compressor::tone_map(float (*cmp)(float, float *),
                     float *opts,
                     float clmin,
                     float clmax,
                     float ldmax,
                     float l)
{
    float cl = cmp(l, opts);
    return ldmax * ((cl - clmin) / (clmax - clmin));
}

float
Compressor::local_tone_map(float (*cmp)(float, float *),
                           float *opts,
                           float clmin,
                           float clmax,
                           float ldmax,
                           float l,
                           float la,
                           float contrast)
{
    float result = 0;
    float tl = tone_map(cmp, opts, clmin, clmax, ldmax, l);


    {
        float tmp = ((la / l) - 1.0f);

        tmp*= contrast;

        //result = tla * (tmp + 1.0f);

        result = tl * (1.0 - tmp);

        if(result < 0.0)
        {
            result = 0.0;
        }
    }
    return result;
}

float
Compressor::compress_gamma(float l, float *options)
{
    float gamma = 0.6;

    if(options)
    {
        gamma = *options;
    }
    return pow(l, gamma);
}

float
Compressor::compress_exponent(float l, float *)
{
    return exp(l);
}

float
Compressor::compress_log(float l, float *)
{
    if(l <= 1e-10)
    {
        return log(1e-10);
    }
    return log(l);
}

float
Compressor::compress_ashikhmin(float l, float *)
{
    float result;

    if(l < 0.0)
    {
        result = 0.0;
    }
    else if(l < 0.0034)
    {
        result = l / 0.0014;
    }
    else if(l < 1.0)
    {
        result = 2.4483 + (log(l / 0.0034)/0.4027);
    }
    else if(l < 7.2444)
    {
        result = 16.563 + ((l - 1)/0.4027);
    }
    else
    {
        result = 32.0639 + (log(l/7.2444)/0.0556);
    }
    return result;
}

float
Compressor::compress_linear(float l, float *)
{
    return l;
}

}
