#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "utils.h"
#include "thread.h"
#include "image.h"
#include "parameters.h"

#include "contrast.h"

const uint32_t IMG_PICTURE = 0;
const uint32_t IMG_LAL = 1;
const uint32_t JOB_SIZE = 160000;// 400 * 400
const float MIN_INTENSITY = 1.0e-10;

#define MAX2(_a,_b) ((_a>_b)?_a:_b)

namespace contrast {

class ContrastJob : public thread::Job
{
public:
    ContrastJob(spectral::Image *img, 
                const spectral::Image *lal,
                float contrast,
                bool flatten,
                size_t start,
                size_t end)
        : m_img(img)
        , m_lal(lal)
        , m_start(start)
        , m_end(end)
        , m_contrast(contrast)
        , m_flatten(flatten)
    {
    }

    virtual void Process(void);
private:
    void ProcessLinear(void);
    void ProcessLog(void);

    spectral::Image *m_img;
    const spectral::Image *m_lal;
    size_t m_start, m_end;
    float m_contrast;
    bool m_flatten;
};


void 
ContrastJob::Process(void)
{
    ProcessLog();
}

void 
ContrastJob::ProcessLinear(void)
{
    if(m_img && m_lal)
    {
        float *buf = m_img->get_buffer();
        float *lal = m_lal->get_buffer();
        size_t index = (m_start * 3) + 2;

        for(size_t i=m_start;i<m_end;i++)
        {
            float L, La, difference;
            L = buf[index];
            La = lal[i];
            difference = L - La;
            L+= (difference * m_contrast);
            if(L < MIN_INTENSITY)
            {
                L = MIN_INTENSITY;
            }
            buf[index] = L;

            index+=3;
        }
    }
}


void 
ContrastJob::ProcessLog(void)
{
    if(m_img && m_lal)
    {
        float *buf = m_img->get_buffer();
        float *lal = m_lal->get_buffer();
        size_t index = (m_start * 3) + 2;

        if(m_flatten)
        {
            for(size_t i=m_start;i<m_end;i++)
            {
                float L, La, difference;
                L = log(MAX2(buf[index], MIN_INTENSITY));
                La = log(MAX2(lal[i], MIN_INTENSITY));
                difference = L - La;

                buf[index] = exp(difference);
    
                index+=3;
            }
        }
        else
        {
            for(size_t i=m_start;i<m_end;i++)
            {
                float L, La, difference;
                L = log(MAX2(buf[index], MIN_INTENSITY));
                La = log(MAX2(lal[i], MIN_INTENSITY));
                difference = L - La;
                L+= (difference * m_contrast);
                buf[index] = exp(L);
    
                index+=3;
            }
        }
    }
}

class ContrastFarmer : public thread::Farmer
{
public:
    ContrastFarmer(Contrast *owner)
    : m_owner(owner)
    {}

    virtual void ProcessCompletedJob(thread::Job *);
private:
    Contrast *m_owner;
};


void
ContrastFarmer::ProcessCompletedJob(thread::Job *)
{
    float progress = GetNumCompleted() + 1;
    progress/= GetNumSubmitted();
    progress*= 100.0;
    if(m_owner)
    {
        m_owner->Progress(progress);
    }
}

Contrast::Contrast()
    : spectral::Processor()
    , m_picture(NULL)
    , m_lal(NULL)
    , m_contrast(1.0)
    , m_flatten(false)
{
}

Contrast::~Contrast()
{
}

// Get slots & output type.  Returns output colour space.
// pairs are <slot id, colour space>
uint32_t
Contrast::GetInterface(
    std::list<std::pair<uint32_t, uint32_t> > &mandatory,
    std::list<std::pair<uint32_t, uint32_t> > &optional) const
{   
    mandatory.clear();
    optional.clear();

    mandatory.push_back(     
        std::pair<uint32_t, uint32_t>(IMG_PICTURE, CLS_CIE_XYY));
            
    mandatory.push_back(     
        std::pair<uint32_t, uint32_t>(IMG_LAL, CLS_LUMINANCE));
    return CLS_CIE_XYY;
}  

bool
Contrast::SetImage(uint32_t type, const spectral::Image *img)
{
    bool result = false;

    if(type == IMG_PICTURE)
    {
        m_picture = NULL;
        if(img)
        {
            if(img->get_colour_space() == CLS_CIE_XYY)
            {
                m_picture = img;
                result = true;
            }
        }
        else
        {
            result = true;
        }
    }

    if(type == IMG_LAL)
    {
        m_lal = NULL;
        if(img)
        {
            if(img->get_colour_space() == CLS_LUMINANCE)
            {
                m_lal = img;
                result = true;
            }
        }
        else
        {
            result = true;
        }
    }

    return result;
}

void
Contrast::ClearImage(uint32_t type)
{
    if(type == IMG_PICTURE)
    {
        m_picture = NULL;
    }
}

void 
Contrast::GetParameters(spectral::Parameters *params) const 
{
    if(params)
    {
        params->AddParameter(std::string("contrast"),
                             0.0, 10.0, m_contrast);
        params->AddParameter(std::string("flatten"), m_flatten);
    }
}

// Apply parameters to a processor.
void
Contrast::SetParameters(const spectral::Parameters *params)
{
    if(params)
    {
        params->GetParameter(std::string("contrast"), m_contrast);
        params->GetParameter(std::string("flatten"), m_flatten);
    }
}

spectral::Image *
Contrast::Process(void)
{
    spectral::Image *result = NULL;

    Progress(0);
    if(m_picture)
    {
        result = new spectral::Image(*m_picture);
    }

    if(result)
    {
        size_t size = size_t(result->get_width()) *
                      size_t(result->get_height());
        size_t start, end;
        const spectral::Image *lal= NULL;

        ContrastFarmer farmer(this);

        start = end = size;
        if(m_lal &&
           (m_lal->get_width() == result->get_width()) &&
           (m_lal->get_height() == result->get_height()))
        {
            // lal image must match source image.
            lal = m_lal;
        } 

        while(end)
        {
            if(end >= JOB_SIZE)
            {
                start-= JOB_SIZE;
            }
            else
            {
                start = 0;
            }
            farmer.Push(new ContrastJob(result,
                                        lal,
                                        m_contrast,
                                        m_flatten,
                                        start,
                                        end));

            end = start;
        }
        farmer.Process();    
    }
    return result;
}

}
