#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "utils.h"
#include "bufutils.h"
#include "thread.h"
#include "image.h"
#include "parameters.h"

#include "convolve_mono.h"

#include "kiss_fft.h"
#include "kiss_fftnd.h"

const uint32_t IMG_PICTURE = 0;
const uint32_t IMG_FILTER = 1;

namespace convolution {

static void complex_multiply(float r1, float i1,
                             float r2, float i2,
                             float &r_out, float &i_out)
{
    r_out = ((r1 * r2) - (i1 * i2));
    i_out = ((r1 * i2) + (r2 * i1));
}

ConvolveMono::ConvolveMono()
    : spectral::Processor()
    , m_picture(NULL)
    , m_filter(NULL)
    , m_wrap_horizontal(false)
    , m_wrap_vertical(false)
{
}

ConvolveMono::~ConvolveMono()
{
}

// Get slots & output type.  Returns output colour space.
// pairs are <slot id, colour space>
uint32_t
ConvolveMono::GetInterface(
    std::list<std::pair<uint32_t, uint32_t> > &mandatory,
    std::list<std::pair<uint32_t, uint32_t> > &optional) const
{   
    mandatory.clear();
    optional.clear();
        
    mandatory.push_back(     
        std::pair<uint32_t, uint32_t>(IMG_PICTURE, CLS_CIE_XYY));
            
    mandatory.push_back( 
        std::pair<uint32_t, uint32_t>(IMG_FILTER, CLS_LUMINANCE));

    return CLS_CIE_XYY;
}  

bool
ConvolveMono::SetImage(uint32_t type, const spectral::Image *img)
{
    bool result = false;

    if(type == IMG_PICTURE)
    {
        m_picture = NULL;
        if(img)
        {
            if(img->get_colour_space() == CLS_CIE_XYY)
            {
                m_picture = img;
                result = true;
            }
        }
        else
        {
            result = true;
        }
    }

    if(type == IMG_FILTER)
    {
        m_filter = NULL;
        if(img)
        {
            if(img->get_colour_space() == CLS_LUMINANCE)
            {
                m_filter = img;
                result = true;
            }
        }
        else
        {
            result = true;
        }
    }

    return result;
}

void
ConvolveMono::ClearImage(uint32_t type)
{
    if(type == IMG_PICTURE)
    {
        m_picture = NULL;
    }
}

void 
ConvolveMono::GetParameters(spectral::Parameters *params) const 
{
    if(params)
    {
        std::string tmp;
        tmp = "wrap horizontal";
        params->AddParameter(tmp, m_wrap_horizontal);

        tmp = "wrap vertical";
        params->AddParameter(tmp, m_wrap_vertical);
    }
}

// Apply parameters to a processor.
void
ConvolveMono::SetParameters(const spectral::Parameters *params)
{
    if(params)
    {
        std::string tmp;
        tmp = "wrap horizontal";
        params->GetParameter(tmp, m_wrap_horizontal);

        tmp = "wrap vertical";
        params->GetParameter(tmp, m_wrap_vertical);
    }
}

spectral::Image *
ConvolveMono::Process(void)
{
    spectral::Image *result = NULL;

    Progress(0);

    if(m_picture && m_filter)
    {
        result = new spectral::Image(*m_picture);
    }

    if(result)
    {
        kiss_fft_cpx *image_output_buffer(NULL);
        kiss_fft_cpx *filter_output_buffer(NULL);

        float scale(1.0);
        int dims[2];
        int ndims(2);

        uint32_t width, height, offset_x, offset_y;

        // Set up image size.
        GetFullSize(width, height, offset_x, offset_y);
        dims[0] = height;
        dims[1] = width;

        // Convert to xyz for processing.
        buf_xyy_to_xyz(result->get_buffer(), (result->get_width() * result->get_height()));

        // Set up forward & reverse transforms.
        kiss_fftnd_cfg  stf = kiss_fftnd_alloc(dims, ndims, 0, 0, 0);
        kiss_fftnd_cfg  sti = kiss_fftnd_alloc(dims, ndims, 1, 0, 0);

        // Set up temporary output buffer.
        image_output_buffer = 
            (kiss_fft_cpx *)calloc(sizeof(kiss_fft_cpx), width*height);

        // setup filter
        {
            kiss_fft_cpx *filter_input_buffer(NULL);
            filter_input_buffer = (kiss_fft_cpx*)CreateFilterBuffer(0);

            filter_output_buffer = 
                (kiss_fft_cpx *)calloc(sizeof(kiss_fft_cpx), width*height);

            kiss_fftnd(stf, filter_input_buffer, filter_output_buffer);

            free(filter_input_buffer);
        }
        // set up scale.
        scale = 1.0/(width * height);
        //scale*= scale;
        Progress(25);

        for(size_t channel=0;channel<3;channel++)
        {
            kiss_fft_cpx *image_input_buffer(NULL);

            image_input_buffer = (kiss_fft_cpx*)CreateImageBuffer(result, channel);
            
            {
                // Forward transformation
                kiss_fftnd(stf, image_input_buffer, image_output_buffer);
                // convolve
                for(uint32_t n=0;n<(width * height);n++)
                {
                    complex_multiply(image_output_buffer[n].r,
                                     image_output_buffer[n].i,
                                     filter_output_buffer[n].r,
                                     filter_output_buffer[n].i,
                                     image_output_buffer[n].r,
                                     image_output_buffer[n].i);
                }
                // backward transformation
                kiss_fftnd(sti, image_output_buffer, image_input_buffer);
            }

            UnpackImageBuffer((void *)image_input_buffer, result, channel);
            
            free((void*)image_input_buffer);
            {
                float progress;
                progress = channel+2;
                progress/=4;
                progress*= 100;
                Progress(progress);
            }
        }
        // clean up
        free(stf);
        free(sti);
        free(image_output_buffer); 
        free(filter_output_buffer); 
               
        // convert back to xyy
        buf_xyz_to_xyy(result->get_buffer(), (result->get_width() * result->get_height()));
        // compensate for fft scaling.
        buf_scale_xyy_y(result->get_buffer(), 
                        (result->get_width() * result->get_height()),
                         scale);

    }
    return result;
}


void
ConvolveMono::GetFullSize(uint32_t &width, uint32_t &height, 
                          uint32_t &offset_x, uint32_t &offset_y)
{
    width = m_filter->get_width() + m_picture->get_width();
    height = m_filter->get_height() + m_picture->get_height();
    offset_x = m_filter->get_width() / 2;
    offset_y = m_filter->get_height() / 2;
}

void *
ConvolveMono::CreateImageBuffer(spectral::Image *image, uint32_t n)
{
    uint32_t width, height, offset_x, offset_y;
    kiss_fft_cpx *result(NULL);

    GetFullSize(width, height, offset_x, offset_y);
    if(width && height)
    {
        result = (kiss_fft_cpx *)calloc(sizeof(kiss_fft_cpx), width*height);

        if(result)
        {
            int32_t x, y;
            uint32_t index(0);
            for(y=0;uint32_t(y)<height;y++)
            {
                for(x=0;uint32_t(x)<width;x++)
                {
                    result[index].r = 
                        (image->get_constrained_pixel(x - int32_t(offset_x),
                                                      y - int32_t(offset_y),
                                                      m_wrap_horizontal,
                                                      m_wrap_vertical))[n];
                    index++;
                }
            }
        }
    } 
    return(void *)result;
}
void
ConvolveMono::UnpackImageBuffer(void *p, spectral::Image *image, uint32_t n)
{
    uint32_t width, height, offset_x, offset_y;
    kiss_fft_cpx *buf = (kiss_fft_cpx*)(p);

    GetFullSize(width, height, offset_x, offset_y);

    if(width && height)
    {
        uint32_t x,y;
        for(y=0;y<image->get_height();y++)
        {
            uint32_t yy;
            yy = y + offset_y;
            for(x=0;x<image->get_width();x++)
            {
                uint32_t xx;
                uint32_t index, img_index;
 
                xx = x + offset_x;
                index = xx+(yy*width);
                
                img_index = ((x + (image->get_width() * y)) * 3) + n;

                if(buf[index].r < 0.0)
                {
                   image->get_buffer()[img_index] = 0.0;
                }
                else
                {
                   image->get_buffer()[img_index] = buf[index].r;
                }
            }
        }
    }

}

void *
ConvolveMono::CreateFilterBuffer(uint32_t n)
{
    uint32_t width, height, offset_x, offset_y;
    kiss_fft_cpx *result(NULL);

    GetFullSize(width, height, offset_x, offset_y);
    if(width && height)
    {
        result = (kiss_fft_cpx *)calloc(sizeof(kiss_fft_cpx), width*height);

        if(result)
        {
            int32_t x, y;
            uint32_t index(0);

            for(size_t i=0;i<width*height;i++)
            {
                result[i].r = result[i].i = 0.0;
            }

            for(y=0;uint32_t(y)<m_filter->get_height();y++)
            {
                int32_t yy;
                yy = y - offset_y;
                //yy+=midy;
                if(yy < 0)
                {
                    yy+= height;
                }
                for(x=0;uint32_t(x)<m_filter->get_width();x++)
                {
                    int32_t xx;
                    xx = x - offset_x;
                    //xx+= midx;
                    if(xx < 0)
                    {
                        xx+= width;
                    }

                    //xx = x;yy=y;
                    index = xx + (yy * width);

                    result[index].r = m_filter->get_pixel(x,y)[n];
                }
            }
        }
    }    
    return result;
}

}
