#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "utils.h"
#include "thread.h"
#include "image.h"
#include "parameters.h"

#include "cropimage.h"

const uint32_t IMG_PICTURE = 0;
const uint32_t MAX_SIZE = 65535;

namespace cropimage {

#define min2(_A, _B) ((_A > _B)?(_B):(_A))
#define max2(_A, _B) ((_A > _B)?(_A):(_B))


CropImage::CropImage()
    : spectral::Processor()
    , m_picture(NULL)
    , m_x1(0)
    , m_y1(0)
    , m_x2(0)
    , m_y2(0)
{
}

CropImage::~CropImage()
{
}

// Get slots & output type.  Returns output colour space.
// pairs are <slot id, colour space>
uint32_t
CropImage::GetInterface(
    std::list<std::pair<uint32_t, uint32_t> > &mandatory,
    std::list<std::pair<uint32_t, uint32_t> > &optional) const
{   
    mandatory.clear();
    optional.clear();
        
    mandatory.push_back(
        std::pair<uint32_t, uint32_t>(IMG_PICTURE, CLS_CIE_XYY));
            
    return CLS_CIE_XYY;
}  

bool
CropImage::SetImage(uint32_t type, const spectral::Image *img)
{
    bool result = false;

    if(type == IMG_PICTURE)
    {
        m_picture = NULL;
        if(img)
        {
            if(img->get_colour_space() == CLS_CIE_XYY)
            {
                m_picture = img;
                result = true;
            }
        }
        else
        {
            result = true;
        }
    }

    return result;
}

void
CropImage::ClearImage(uint32_t type)
{
    if(type == IMG_PICTURE)
    {
        m_picture = NULL;
    }
}

void 
CropImage::GetParameters(spectral::Parameters *params) const 
{
    if(params)
    {
        params->AddParameter(std::string("x1"), 0, MAX_SIZE, m_x1);
        params->AddParameter(std::string("y1"), 0, MAX_SIZE, m_y1);
        params->AddParameter(std::string("x2"), 0, MAX_SIZE, m_x2);
        params->AddParameter(std::string("y2"), 0, MAX_SIZE, m_y2);
    }
}

// Apply parameters to a processor.
void
CropImage::SetParameters(const spectral::Parameters *params)
{
    if(params)
    {
        int32_t x1,x2,y1,y2;
        params->GetParameter(std::string("x1"), x1);
        params->GetParameter(std::string("y1"), y1);
        params->GetParameter(std::string("x2"), x2);
        params->GetParameter(std::string("y2"), y2);

        m_x1 = min2(x1, x2);
        m_y1 = min2(y1, y2);
        m_x2 = max2(x1, x2);
        m_y2 = max2(y1, y2);
    }
}

spectral::Image *
CropImage::Process(void)
{
    spectral::Image *result = NULL;

    Progress(0);

    if(m_picture)
    {
        int32_t x1, y1, x2, y2, width, height;
        x1 = max2(0, min2((int32_t)m_picture->get_width() - 1, m_x1));
        x2 = max2(0, min2((int32_t)m_picture->get_width() - 1, m_x2));
        y1 = max2(0, min2((int32_t)m_picture->get_height() - 1, m_y1));
        y2 = max2(0, min2((int32_t)m_picture->get_height() - 1, m_y2));
        width = (x2 - x1) + 1;
        height = (y2 - y1) + 1;

        result = new spectral::Image(width, height, CLS_CIE_XYY);

        if(result)
        {
            for(int32_t y=0;y<height;y++) // target y
            {
                int32_t yy; // source y
                yy = y + y1;
                for(int32_t x=0;x<width;x++) // target x
                {
                    int32_t xx; // source x
                    xx = x + x1;

                    result->set_pixel(x,y,m_picture->get_pixel(xx,yy));
                }
            }
            
        }
    }
    return result;
}

}
