#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "utils.h"
#include "thread.h"
#include "image.h"
#include "parameters.h"

#include "filmic.h"

const uint32_t IMG_PICTURE = 0;
const uint32_t MAX_SIZE = 65535;
const uint32_t JOB_SIZE = 160000;// 400 * 400

namespace filmic {


inline float filmic_internal(float x)
{
    const float A(0.15);
    const float B(0.50);
    const float C(0.10);
    const float D(0.20);
    const float E(0.02);
    const float F(0.30);

    return ((x*(A*x + C*B)+D*E)/(x*(A*x)+B)+D*F)-E/F;
}


inline float tm_filmic(float prescale, float input)
{
    const float W(11.2);
    const float whiteScale (1.0/filmic_internal(W));

    float tmp = filmic_internal(prescale * input) * whiteScale;

    if(tmp > 1.0)
    {
        tmp = 1.0;
    }
    return tmp;
}


class FilmicJob : public thread::Job
{
public:
    FilmicJob(spectral::Image *img,
             float prescale,
             bool rgb,
             size_t start,
             size_t end)
        : m_img(img)
        , m_prescale(prescale)
        , m_start(start)
        , m_end(end)
        , m_rgb(rgb)
    {
    }

    virtual void Process(void);
private:

    spectral::Image *m_img;
    float m_prescale;
    size_t m_start, m_end;
    bool m_rgb;
};

void
FilmicJob::Process(void)
{
    if(m_img)
    {
        float *buf = m_img->get_buffer();

        if(m_rgb)
        {
            float *p = buf + (m_start * 3);
            float *px = p;

            for(size_t i=m_start;i<m_end;i++)
            {
                px = p;
                xyy_to_linear_srgb(px);
                *p= tm_filmic(m_prescale, *p); p++;
                *p = tm_filmic(m_prescale, *p); p++;
                *p = tm_filmic(m_prescale, *p); p++;
                linear_srgb_to_xyy(px);
            }
        }
        else
        {
            float *p = buf + ((m_start * 3) + 2);
            for(uint32_t i=m_start;i<m_end;i++)
            {
                *p = tm_filmic(m_prescale, *p);
    
                p+=3;
            }
        }
    }
}

class FilmicFarmer : public thread::Farmer
{
public:
    FilmicFarmer(Filmic *owner)
    : m_owner(owner)
    {}

    virtual void ProcessCompletedJob(thread::Job *);
private:
    Filmic *m_owner;
};


void
FilmicFarmer::ProcessCompletedJob(thread::Job *)
{
    float progress = GetNumCompleted() + 1;
    progress/= GetNumSubmitted();
    progress*= 100.0;
    if(m_owner)
    {
        m_owner->Progress(progress);
    }
}

// estimate log10 average luminance of the image
float
Filmic::GetAdaptation(spectral::Image *img, uint32_t num_samples)
{
    const float sigma = 1.0e-6;
    float result = sigma;

    if(!num_samples)
    {
        num_samples = 1;
    }
    if(img)
    {
        uint32_t width, height;
        float total(0.0);

        width = img->get_width();
        height = img->get_height();

        for(uint32_t i=0;i<num_samples;i++)
        {
            float fx, fy;
            int32_t x, y;

            fx = halton(3, i);
            fy = halton(5, i);
            fx*= width;
            fy*= height;

            x = (int32_t)round_float(fx);
            y = (int32_t)round_float(fy);

            if(x<0)x=0;
            if(y<0)y=0;
            if(x>=(int32_t)width)x=(int32_t)(width-1);
            if(y>=(int32_t)height)y=(int32_t)(height-1);


            {
                const float *pixel = img->get_pixel(x, y);
                total+= log(sigma + pixel[2]);

/*
                if(pixel[2] > max)
                {
                    max = pixel[2];
                }
*/
            }
        }
        total/= num_samples;
        result = exp(total);
    }
    return result;
}

Filmic::Filmic()
    : spectral::Processor()
    , m_samples(2000)
    , m_exposure(0.0)
    , m_rgb(false)
{
}

Filmic::~Filmic()
{
}

// Get slots & output type.  Returns output colour space.
// pairs are <slot id, colour space>
uint32_t
Filmic::GetInterface(
    std::list<std::pair<uint32_t, uint32_t> > &mandatory,
    std::list<std::pair<uint32_t, uint32_t> > &optional) const
{   
    mandatory.clear();
    optional.clear();
        
    mandatory.push_back(
        std::pair<uint32_t, uint32_t>(IMG_PICTURE, CLS_CIE_XYY));

    return CLS_CIE_XYY;
}  

bool
Filmic::SetImage(uint32_t type, const spectral::Image *img)
{
    bool result = false;

    if(type == IMG_PICTURE)
    {
        m_picture = NULL;
        if(img)
        {
            if(img->get_colour_space() == CLS_CIE_XYY)
            {
                m_picture = img;
                result = true;
            }
        }
        else
        {
            result = true;
        }
    }

    return result;
}

void
Filmic::ClearImage(uint32_t type)
{
    if(type == IMG_PICTURE)
    {
        m_picture = NULL;
    }
}

void 
Filmic::GetParameters(spectral::Parameters *params) const 
{
    if(params)
    {
        params->AddParameter(std::string("exposure"), -10.0, 10.0, m_exposure);
        params->AddParameter(std::string("rgb"), m_rgb);
        params->AddParameter(std::string("samples"),
                                         100, 10000000, m_samples);
    }
}

// Apply parameters to a processor.
void
Filmic::SetParameters(const spectral::Parameters *params)
{
    if(params)
    {
        params->GetParameter(std::string("exposure"), m_exposure);
        params->GetParameter(std::string("rgb"), m_rgb);
        params->GetParameter(std::string("samples"), m_samples);
    }
}

spectral::Image *
Filmic::Process(void)
{
    spectral::Image *result = NULL;
    Progress(0);

    if(m_picture)
    {
        result = new spectral::Image(*m_picture);

        if(result)
        {
            size_t size = size_t(m_picture->get_width()) * size_t(m_picture->get_height());
            float adaptation;
            float exposure;

            // get adaptation & max luminance.
            adaptation = GetAdaptation(result, m_samples);

            exposure = pow((float)2.0, m_exposure);

            {
                float scale;
                size_t start, end;
                FilmicFarmer farmer(this);

                scale = exposure / adaptation;

                start = end = size;

                while(end)
                {
                    if(end >= JOB_SIZE)
                    {
                        start-= JOB_SIZE;
                    }
                    else
                    {
                        start = 0;
                    }
                    farmer.Push(new FilmicJob(result,
                                             scale,
                                             m_rgb,
                                             start,
                                             end));
                    end = start;
                }
                farmer.Process();
            }
        }
    }

    return result;
}

}
