#include <stdlib.h>
#include <math.h>

#include <stdio.h>
#include <iostream>
#include <list>
#include <utility>

#include "utils.h"
#include "image.h"
#include "parameters.h"

#include "field.h"

const uint32_t IMG_PICTURE = 0;
const uint32_t IMG_FIELD   = 1;

#define STEP 0.5

namespace spectral {

Field::Field()
    : Processor()
    , m_picture(NULL)
    , m_field(NULL)
    , m_weight(1.0)
    , m_size(60)
    , m_fine(true)
    , m_contour(true)
    , m_streamline_length(500)
{
}

Field::~Field()
{
}

// Get slots & output type.  Returns output colour space.
// pairs are <slot id, colour space>
uint32_t
Field::GetInterface(
    std::list<std::pair<uint32_t, uint32_t> > &mandatory,
    std::list<std::pair<uint32_t, uint32_t> > &) const
{   
    mandatory.clear();
        
    mandatory.push_back(     
        std::pair<uint32_t, uint32_t>(IMG_PICTURE, CLS_CIE_XYY));

    mandatory.push_back(     
        std::pair<uint32_t, uint32_t>(IMG_FIELD, CLS_LUMINANCE));
    return CLS_CIE_XYY;
}  

bool
Field::SetImage(uint32_t type, const spectral::Image *img)
{
    bool result = false;

    if(type == IMG_PICTURE)
    {
        m_picture = NULL;
        if(img)
        {
            if(img->get_colour_space() == CLS_CIE_XYY)
            {
                m_picture = img;
                result = true;
            }
        }
        else
        {
            result = true;
        }
    }
    if(type == IMG_FIELD)
    {
        m_field = NULL;
        if(img)
        {
            if(img->get_colour_space() == CLS_LUMINANCE)
            {
                m_field = img;
                result = true;
            }
        }
        else
        {
            result = true;
        }
    }

    return result;
}

void
Field::ClearImage(uint32_t type)
{
    if(type == IMG_PICTURE)
    {
        m_picture = NULL;
    }
    if(type == IMG_FIELD)
    {
        m_field = NULL;
    }
}

void 
Field::GetParameters(Parameters *params) const 
{
    if(params)
    {
        params->AddParameter("fine", m_fine);
        params->AddParameter("weight", 1e-5, 20.0, m_weight);
        params->AddParameter("follow contours",  m_contour);
        params->AddParameter("stream length", 5, 5000, m_streamline_length);
    }
}

// Apply parameters to a processor.
void
Field::SetParameters(const Parameters *params)
{
    if(params)
    {
        params->GetParameter("fine", m_fine);
        params->GetParameter("weight", m_weight);
        params->GetParameter("follow contours",  m_contour);
        params->GetParameter("stream length", m_streamline_length);
    }
}

// Make a vector field from the input field.
spectral::image<float> *
Field::MakeField(void)
{
    spectral::image<float> *result = NULL;

    if(m_field)
    {
        result = new image<float>(m_field->get_width(),
                                  m_field->get_height(),
                                  2);
    }

    if(result)
    {
        const float oo_sqrt2 = 1.0/sqrt(2.0);
        float *out_buf = result->get_buffer();
        int width = m_field->get_width();
        int height = m_field->get_height();

        for(int y=0;y<height;y++)
        {
            for(int x=0;x<width;x++)
            {
                float cur_value;
                float *pixel, diff, total;
                pixel = out_buf + result->get_index(x, y);

                pixel[0] = 0.0;
                pixel[1] = 0.0;
                total = 0;

                cur_value = *(m_field->get_pixel(x,y));
                // nested loops unrolled to avoid apparent compiler bug.
                // Let the optimiser soer tout the logic.

                if((x > 0) && (y > 0))
                {
                    // -1,-1
                    diff = cur_value - *m_field->get_pixel(x-1,y-1);

                    diff*= oo_sqrt2;
                    pixel[0] -= diff;
                    pixel[1] -= diff;
                    total+= oo_sqrt2;
                }
                if(x > 0)
                {
                    // -1,0
                    diff = cur_value - *m_field->get_pixel(x-1,y);
                    pixel[0] -= diff;
                    total+= 1;
                }
                if((x > 0) && (y < (height - 1)))
                {
                    // -1,1
                    diff = cur_value - *m_field->get_pixel(x-1,y+1);

                    diff*= oo_sqrt2;
                    pixel[0] -= diff;
                    pixel[1] += diff;
                    total+= oo_sqrt2;
                }
                if(y > 0)
                {
                    // 0,-1
                    diff = cur_value - *m_field->get_pixel(x,y-1);
                    pixel[1]-=diff;
                    total+=1;
                }
                if(y < (height - 1))
                {
                    // 0,-1
                    diff = cur_value - *m_field->get_pixel(x,y+1);
                    pixel[1]+=diff;
                    total+=1;
                }
                if((x < (width-1)) && (y > 0))
                {
                    // 1,-1
                    diff = cur_value - *m_field->get_pixel(x+1,y-1);

                    diff*= oo_sqrt2;
                    pixel[0] += diff;
                    pixel[1] -= diff;
                    total+= oo_sqrt2;
                }
                if(x < (width-1))
                {
                    // 1,0
                    diff = cur_value - *m_field->get_pixel(x+1,y);
                    pixel[0] += diff;
                    total+= 1;
                }
                if((x < (width+1)) && (y < (height - 1)))
                {
                    // -1,1
                    diff = cur_value - *m_field->get_pixel(x+1,y+1);

                    diff*= oo_sqrt2;
                    pixel[0] += diff;
                    pixel[1] += diff;
                    total+= oo_sqrt2;
                }
                pixel[0]/=total;
                pixel[1]/=total;

            }
        }
    }

    return result;
}

void 
Field::GetWeight(int32_t width, int32_t height, float x, float y, struct weights &out)
{
    int32_t x1,x2,y1,y2;

    // Get positions of 4 pixels.

    x1 = (int32_t)round_float(floor(x));
    y1 = (int32_t)round_float(floor(y));

    x2 = x1+1;
    y2 = y1+1;

    // coverage in x and y.
    float xc1 = fmod(x, (float)1.0);
    float yc1 = fmod(x, (float)1.0);
    float xc2 = 1.0 - xc1;
    float yc2 = 1.0 - yc1;

    // Default indices oor.
    out.index[0] = out.index[1] = out.index[2] = out.index[3] = -1;

    if((x1 >= 0) && (x1 < width))
    {
        if((y1 >= 0) && (y1 < height))
        {
            out.index[0] = x1 + (width * y1);
        }
        
        if((y2 >= 0) && (y2 < height))
        {
            out.index[1] = x1 + (width * y2);
        }
    }

    if((x2 >= 0) && (x2 < width))
    {
        if((y1 >= 0) && (y1 < height))
        {
            out.index[2] = x2 + (width * y1);
        }
        
        if((y2 >= 0) && (y2 < height))
        {
            out.index[3] = x2 + (width * y2);
        }
    }

    out.weight[0] = xc1 * yc1;
    out.weight[1] = xc1 * yc2;
    out.weight[2] = xc2 * yc1;
    out.weight[3] = xc2 * yc2;
}
void
Field::BuildStreamline(float x, float y, 
                      spectral::image<float> *field, 
                      std::list<std::pair<float, float> > &result)
{
    result.clear();

    if(field)
    {
        float direction[2];
        float pos[2];
        int width = field->get_width();
        int height = field->get_height();
        pos[0] = x;
        pos[1] = y;

        if(GetFieldDirection(field, pos, direction))
        {
            ////std::cerr << "got first point" << std::endl;
            result.push_back(std::pair<float, float>(x, y));

            // Follow field in dx, dy;
            {
                bool follow;
                float old_direction[2];
                float new_pos[2];
                int len;

                pos[0] = x;
                pos[1] = y;
                old_direction[0] = direction[0];
                old_direction[1] = direction[1];
                follow = true;

                len = m_streamline_length;

                // Follow the streamline in the +ve direction.
                while(follow && len)
                {
                    len--;
                
                    old_direction[0] = direction[0];
                    old_direction[1] = direction[1];
                    follow = GetFieldDirection(field, pos, direction);

                    // lazy integrator;
                    if(follow)
                    {
                        // Bail out if direction changes by more than 90 
                        // degrees.
                        float dp = (direction[0] * old_direction[0]) +
                                   (direction[1] * old_direction[1]);
                        if(dp < -0.5)
                        {
                            follow = false;
                            //std::cerr << "change in dir\n";
                        }
                    }
                    else
                    {
                        //std::cerr << "GetFieldDirection broke\n";
                    }

                    // Step along the streamline.
                    if(follow)
                    {
                        new_pos[0] = pos[0] + (direction[0] * STEP);
                        new_pos[1] = pos[1] + (direction[1] * STEP);

                        if((new_pos[0] >= 0.0) && (new_pos[0] < width) &&
                           (new_pos[1] >= 0.0) && (new_pos[1] < height))
                        {
                            old_direction[0] = direction[0];
                            old_direction[1] = direction[1];
                            pos[0] = new_pos[0];
                            pos[1] = new_pos[1];
                            result.push_back(
                                std::pair<float, float>(pos[0], pos[1]));
                        }
                        else
                        {
                            //std::cerr << "Out of image.\n";
                            follow = false;
                        }
                    }
                }

                pos[0] = x;
                pos[1] = y;

                follow = GetFieldDirection(field, pos, direction);
                direction[0] = -direction[0];
                direction[1] = -direction[1];
                old_direction[0] = direction[0];
                old_direction[1] = direction[1];

                follow = true;
                len = m_streamline_length;

                // Follow the streamline in the -ve direction.
                while(follow && len)
                {
                    len--;
                
                    old_direction[0] = direction[0];
                    old_direction[1] = direction[1];

                    follow = GetFieldDirection(field, pos, direction);

                    direction[0] = -direction[0];
                    direction[1] = -direction[1];
                    // lazy integrator;
                    if(follow)
                    {
                        // Bail out if direction changes by more than 90 
                        // degrees.
                        float dp = (direction[0] * old_direction[0]) +
                                   (direction[1] * old_direction[1]);
                        if(dp < -0.5)
                        {
                            follow = false;
                            //std::cerr << "change in dir\n";
                        }
                    }
                    else
                    {
                        //std::cerr << "GetFieldDirection broke.\n";
                    }

                    // Step along the streamline.
                    if(follow)
                    {
                        new_pos[0] = pos[0] + (direction[0] * STEP);
                        new_pos[1] = pos[1] + (direction[1] * STEP);

                        if((new_pos[0] >= 0.0) && (new_pos[0] < width) &&
                           (new_pos[1] >= 0.0) && (new_pos[1] < height))
                        {
                            old_direction[0] = direction[0];
                            old_direction[1] = direction[1];
                            pos[0] = new_pos[0];
                            pos[1] = new_pos[1];
                            result.push_front(
                                std::pair<float, float>(pos[0], pos[1]));
                        }
                        else
                        {
                            //std::cerr << "oor\n";
                            follow = false;
                        }
                    }
                }
            }
        }
    }
}

bool
Field::GetFieldDirection(spectral::image<float> *field, 
                        float *pos, float *dir)
{
    bool result(false);

    if(field && pos && dir)
    {
        struct weights wt;
        GetWeight(field->get_width(), field->get_height(), pos[0], pos[1], wt);
        float *buf = (field->get_buffer());

        dir[0] = 0.0;
        dir[1] = 0.0;

        for(int i=0;i<4;i++)
        {
            if(wt.index[i] != -1)
            {
                float fx,fy;
                int idx;
                idx = wt.index[i] * 2;

                fx = buf[idx];
                idx++;
                fy = buf[idx];

                fx*= wt.weight[i];
                fy*= wt.weight[i];

                dir[0]+= fx;
                dir[1]+= fy;

                result = true;
            }
        }
        if(result)
        {
            float len = (dir[0] * dir[0]) + (dir[1] * dir[1]);
            len = sqrt(len);

            if(len > 1e-10)
            {
                dir[0]/=len;
                dir[1]/=len;
                if(m_contour)
                {
                    float tmp;
                    tmp = dir[1];
                    dir[1] =-dir[0];
                    dir[0] = tmp;
                }
            }
            else
            {
                result = false;
            }
        }
    }

    return result;
}

bool
Field::GetColour(const spectral::Image *img, float x, float y, float *colour)
{
    bool result(false);

    if(img && colour)
    {
        int j;
        struct weights wt;
        GetWeight(img->get_width(), img->get_height(), x, y, wt);
        float *buf = (img->get_buffer());

        for(j=0;j<3;j++)
        {
            colour[j] = 0.0;
        }

        for(int i=0;i<4;i++)
        {
            if(wt.index[i] != -1)
            {
                int idx;
                idx = wt.index[i] * 3;

                for(j=0;j<3;j++)
                {
                    colour[j]+= buf[idx + j] * wt.weight[i];
                }
                result = true;
            }
        }
    }

    return result;
}

bool
Field::SplatColour(float x, float y, float *colour,
                  spectral::Image *img,
                  spectral::image<float> *hits,
                  float weight)
{
    bool result(false);

    if(img && colour && hits)
    {
        int32_t j;
        struct weights wt;
        float *col_buf = (img->get_buffer());
        float *wt_buf = (hits->get_buffer());

        if(m_fine)
        {
            int32_t width = img->get_width();
            int32_t height = img->get_height();

            int32_t ix = (int32_t)round_float(x);
            int32_t iy = (int32_t)round_float(y);

            if((ix >= 0) && (ix < width) &&
               (iy >= 0) && (iy < height))
            {
                int32_t wt_idx = ix + (iy * width);
                int32_t col_idx = wt_idx * 3;

                wt_buf[wt_idx]+= weight;
                col_buf[col_idx]  += colour[0] * weight;
                col_buf[col_idx+1]+= colour[1] * weight;
                col_buf[col_idx+2]+= colour[2] * weight;

                result = true;
            }
        }
        else
        {
            GetWeight(img->get_width(), img->get_height(), x, y, wt);

            for(int32_t i=0;i<4;i++)
            {
                if(wt.index[i] != -1)
                {
                    int col_idx, wt_idx;
    
                    col_idx = wt.index[i] * 3;
                    wt_idx = wt.index[i];
    
                    for(j=0;j<3;j++)
                    {
                        col_buf[col_idx+j]+= (colour[j] * wt.weight[i])*weight;
                    }

                    wt_buf[wt_idx]+= wt.weight[i] * weight;

                    result = true;
                }
            }
        }
    }

    return result;
}

void 
Field::FollowStreamline(std::list<std::pair<float, float> >&line,
                       spectral::Image *src,
                       spectral::Image *result,
                       spectral::image<float> *hits)
{

    if(line.size())
    {
        float total[3];
        int32_t head_dist, tail_dist, length;
        int32_t filter_len = m_size * 4;
        std::list<std::pair<float, float> >::iterator head =
            line.begin();
        std::list<std::pair<float, float> >::iterator mid =
            head;
        std::list<std::pair<float, float> >::iterator tail =
            head;

        total[0]=total[1]=total[2] = 0.0;
        head_dist = 0;
        tail_dist = 0;

        // Add in first point.
        {
            float colour[3];
            GetColour(src, head->first, head->second, colour);
            total[0]+=colour[0];
            total[1]+=colour[1];
            total[2]+=colour[2];
            length = 1;
        }
        while(mid != line.end())
        {
            bool splat;
            splat = false;

            // Move "front" of filter along stream line.
            if(head != line.end())
            {
                head++;
                head_dist++;
                length++;
                {
                    // Add head point.
                    float colour[3];
                    GetColour(src, 
                              head->first,
                              head->second,
                              colour);
                    total[0]+=colour[0];
                    total[1]+=colour[1];
                    total[2]+=colour[2];
                }
            }
            // If front is far enough ahead, then move mid 
            // point.
            if((head_dist > filter_len) ||
               (head == line.end()))
            {
                if((mid == line.begin()) && length)
                {
                    // splat here
                    float colour[3];
                    colour[0] = total[0];
                    colour[1] = total[1];
                    colour[2] = total[2];
                    colour[0]/= length;
                    colour[1]/= length;
                    colour[2]/= length;
                    SplatColour(mid->first, mid->second,
                                colour, result, hits, m_weight);
                }
                mid++;
                head_dist--;
                tail_dist++;
                splat = true;
            }
            // Drag the back of the filter along.
            if(tail_dist > filter_len)
            {
                {
                    // subtract tail pointer.
                    float colour[3];
                    GetColour(src, 
                              tail->first,
                              tail->second,
                              colour);
                    total[0]-=colour[0];
                    total[1]-=colour[1];
                    total[2]-=colour[2];
                }
                length--;
                tail_dist--;
                tail++;
            }
            if(splat && length)
            {
                // splat here
                float colour[3];
                colour[0] = total[0];
                colour[1] = total[1];
                colour[2] = total[2];
                colour[0]/= length;
                colour[1]/= length;
                colour[2]/= length;
                SplatColour(mid->first, mid->second,
                            colour, result, hits, m_weight);
            }
        }
    }
}

void 
Field::DarkStreamline(std::list<std::pair<float, float> >&line,
                     spectral::Image *result,
                     spectral::image<float> *hits)
{

    if(line.size())
    {
        float total[3];
        //int32_t head_dist(0), tail_dist(0);
        std::list<std::pair<float, float> >::iterator it =
            line.begin();

        total[0]=total[1]=total[2] = 0.0;

        float colour[3] = {0,0,0};
        while(it != line.end())
        {
            //bool splat;
            //splat = false;

            // Move "front" of filter along stream line.
            SplatColour(it->first, it->second,
                        colour, result, hits, m_weight);
            it++;
        }
    }
}

}
