#include <stdio.h>
#include <stdlib.h>
#include <math.h>
//#include <assert.h>

#include <vector>
#include <set>

#include "parameters.h"
#include "image.h"
#include "thread.h"

#include "global.h"

namespace global {

const uint32_t IMG_LUMINANCE = 0;
const uint32_t MAGIC = 0xffffffff;

////////////////////////////////////////////////////////////////////////////////
// comparison function for quicksort.
static int compare_filter_pixels(const void *p1, const void *p2)
{
    const filter_pixel *f1 = (const filter_pixel *)p1;
    const filter_pixel *f2 = (const filter_pixel *)p2;
    if(f1->value < f2->value)
    {
        return -1;
    }
    else if(f1->value > f2->value)
    {
        return 1;
    }
    else
    {
        return 0;
    }
}

////////////////////////////////////////////////////////////////////////////////
//
// Global processor implementation.
//
////////////////////////////////////////////////////////////////////////////////
Global::Global()
    : spectral::Processor()
    , m_lumImage(NULL)
    , m_threshold(0.3)
{
}

Global::~Global()
{
}

// Get slots & output type.  Returns output colour space.
// pairs are <slot id, colour space>
uint32_t
Global::GetInterface(
    std::list<std::pair<uint32_t, uint32_t> > &mandatory,
    std::list<std::pair<uint32_t, uint32_t> > &optional) const
{
    mandatory.clear();
    optional.clear();

    mandatory.push_back(
        std::pair<uint32_t, uint32_t>(IMG_LUMINANCE, CLS_LUMINANCE));

    return CLS_LUMINANCE;
}


bool
Global::SetImage(uint32_t slot, const spectral::Image *img)
{
    if(slot == IMG_LUMINANCE)
    {
        m_lumImage = NULL;

        if(img)
        {
            if(img->get_colour_space() == CLS_LUMINANCE)
            {
                m_lumImage = img;

                return true;
            }
        }
        else
        {
            return true;
        }
    }
    return false;
}

void
Global::GetParameters(spectral::Parameters *params) const
{
    if(params)
    {
        std::string tmp;

        tmp = "threshold";
        params->AddParameter(tmp, 0.1, 10.0, m_threshold);
    }
}

void
Global::SetParameters(const spectral::Parameters *params)
{
    if(params)
    {
       std::string tmp;

       tmp = "threshold";
       params->GetParameter(tmp, m_threshold);
    }
}

spectral::Image *
Global::Process(void)
{
    spectral::Image *result(NULL);

    Progress(0);
    if(m_lumImage)
    {
        result = new spectral::Image(*m_lumImage);

        CrunchImage(m_lumImage, result, m_threshold/logf(2.0)); 
    }

    Progress(100.0);
    return result;
}

void
Global::ClearImage(uint32_t slot)
{
    if(slot == IMG_LUMINANCE)
    {
        m_lumImage = NULL;
    }
}

void
Global::CrunchImage(const spectral::Image *in, spectral::Image *out, float threshold)
{
    if(in && out)
    {
        //uint32_t width = in->get_width();
        //uint32_t height = in->get_height();
        size_t size = size_t(in->get_width()) * size_t(in->get_height());

        // Move input into log space
        {
            float *buf = in->get_buffer();
            for(size_t i=0;i<size;i++)
            {
                if(buf[i] < 1e-20)
                { 
                    buf[i]=1e-20;
                }
                buf[i] = logf(buf[i]);
                buf[i]/= threshold;
            }
        }

        filter_pixel *pixel_map;
        pixel_map = (filter_pixel *)calloc(size, sizeof(filter_pixel));


        if(pixel_map)
        {
            size_t *index_buf;
            float *buf = in->get_buffer();
            size_t i;
    
            index_buf = (size_t *)calloc(size, sizeof(size_t));

            if(index_buf)
            {
                // Step 1
        
                // Sort input tile in order of value..
        
                for(i=0;i<size;i++)
                {
                    pixel_map[i].value = buf[i];
                    pixel_map[i].index = i;
                }
        
                qsort((void*)pixel_map, size, sizeof(filter_pixel), 
                      compare_filter_pixels);
        
                for(i=0;i<size;i++)
                {
                    index_buf[pixel_map[i].index] = i;
                }

                // Set up filter.
                {
                    // Use double precision for this.
                    double low_total(0),
                           low_sq_total(0),
                           low_cube_total(0),
                           high_total(0),
                           high_sq_total(0),
                           high_cube_total(0),
                           cur_val(0);
                    size_t low_index(0), high_index(0), cur_index(0);

                    cur_val = pixel_map[0].value;

                    while(cur_index < size)
                    {
                        // Move upper end of window up.
                        while(((high_index+1) < size) && 
                              (pixel_map[high_index+1].value < (cur_val + 1.0)))
                        {
                            high_index++;
                            high_total+= pixel_map[high_index].value;
                            high_sq_total+= (pixel_map[high_index].value *
                                             pixel_map[high_index].value);
                            high_cube_total+= (pixel_map[high_index].value *
                                               pixel_map[high_index].value *
                                               pixel_map[high_index].value);
                        }
                        // move lower end of window up.
                        while((low_index < cur_index) && 
                              (pixel_map[low_index].value < (cur_val - 1.0)))
                        {
                            low_total-= pixel_map[low_index].value;
                            low_sq_total-= (pixel_map[low_index].value *
                                            pixel_map[low_index].value);
                            low_cube_total-= (pixel_map[low_index].value *
                                              pixel_map[low_index].value *
                                              pixel_map[low_index].value);
                            low_index++;
                        }
                        // Calculate values.
                        {
                            double total_weight, total_value;

                            {
                                size_t total_count = high_index - low_index;
                                double cur_sq_val = (cur_val * cur_val);
                                double total = low_total + high_total;
                                double sq_total = low_sq_total + high_sq_total;
                                double cube_total = low_cube_total + 
                                                    high_cube_total;

                                total_weight = total_count * (1.0 - cur_sq_val);
                                total_weight+= (2.0 * cur_val * total);
                                total_weight-= sq_total;

                                total_value = (1.0 - cur_sq_val) * total;
                                total_value+= (2.0 * cur_val * sq_total);
                                total_value-= cube_total;
                            }

                            out->get_buffer()[pixel_map[cur_index].index] = total_value / total_weight;
                        }
                        
                        // move centre of window up.
                        low_total+=     cur_val;
                        low_sq_total+=  (cur_val * cur_val);
                        low_cube_total+=  (cur_val * cur_val * cur_val);
                        cur_index++;
                        cur_val =   pixel_map[cur_index].value;
                        high_total-=     cur_val;
                        high_sq_total-= (cur_val * cur_val);
                        high_cube_total-= (cur_val * cur_val * cur_val);
                        if((cur_index % 100000) == 0)
                        {
                            float progress = cur_index * 100;
                            progress /= size;
                            Progress(progress);
                        }
                    }
                }

                free(index_buf);
            }
            free(pixel_map);
        }
        
        // convert output from log space.
        {
            float *buf = out->get_buffer();
            for(uint32_t i=0;i<size;i++)
            {
                buf[i]*= threshold;
                buf[i] = expf(buf[i]);
            }
        }
    }
}

}
