#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "utils.h"
#include "thread.h"
#include "image.h"
#include "parameters.h"

#include "drago.h"

const uint32_t IMG_PICTURE = 0;
const uint32_t MAX_SIZE = 65535;
const uint32_t JOB_SIZE = 160000;// 400 * 400

namespace logmap {


class DragoJob : public thread::Job
{
public:
    DragoJob(spectral::Image *img,
             float scale,
             float exponent, 
             float adaptation,
             float LwMax,
             size_t start,
             size_t end)
        : m_img(img)
        , m_scale(scale)
        , m_exponent(exponent)
        , m_adaptation(adaptation)
        , m_LwMax(LwMax)
        , m_start(start)
        , m_end(end)
    {
    }

    virtual void Process(void);
private:

    spectral::Image *m_img;
    const spectral::Image *m_lal;
    float m_scale, m_exponent, m_adaptation, m_LwMax;
    size_t m_start, m_end;
};

void
DragoJob::Process(void)
{
    if(m_img)
    {
        float *buf = m_img->get_buffer();
        size_t index = (m_start * 3) + 2;

        for(size_t i=m_start;i<m_end;i++)
        {
            float numerator, denominator, Lw;

            Lw = buf[index] / m_adaptation;

            numerator = log(Lw + 1.0);
            denominator = log(2.0 + (8.0 * (pow((Lw/m_LwMax), m_exponent))));
            buf[index] = (m_scale * numerator) / denominator;

            index+=3;
        }
    }
}

class DragoFarmer : public thread::Farmer
{
public:
    DragoFarmer(Drago *owner)
    : m_owner(owner)
    {}

    virtual void ProcessCompletedJob(thread::Job *);
private:
    Drago *m_owner;
};


void
DragoFarmer::ProcessCompletedJob(thread::Job *)
{
    float progress = GetNumCompleted() + 1;
    progress/= GetNumSubmitted();
    progress*= 100.0;
    if(m_owner)
    {
        m_owner->Progress(progress);
    }
}

// estimate log10 average luminance of the image
float
Drago::GetAdaptation(spectral::Image *img, uint32_t num_samples, float &max)
{
    const float sigma = 1.0e-6;
    float result = sigma;
    max = 1.0e-10;

    if(!num_samples)
    {
        num_samples = 1;
    }
    if(img)
    {
        uint32_t width, height;
        float total(0.0);

        width = img->get_width();
        height = img->get_height();

        for(uint32_t i=0;i<num_samples;i++)
        {
            float fx, fy;
            int32_t x, y;

            fx = halton(3, i);
            fy = halton(5, i);
            fx*= width;
            fy*= height;

            x = (int32_t)round_float(fx);
            y = (int32_t)round_float(fy);

            if(x<0)x=0;
            if(y<0)y=0;
            if(x>=(int32_t)width)x=(int32_t)(width-1);
            if(y>=(int32_t)height)y=(int32_t)(height-1);


            {
                const float *pixel = img->get_pixel(x, y);
                total+= log(sigma + pixel[2]);

                if(pixel[2] > max)
                {
                    max = pixel[2];
                }
            }
        }
        total/= num_samples;
        result = exp(total);
    }
    return result;
}

Drago::Drago()
    : spectral::Processor()
    , m_picture(NULL)
    , m_b(0.73)
    , m_exposure(0.0)
    , m_samples(2000)
{
}

Drago::~Drago()
{
}

// Get slots & output type.  Returns output colour space.
// pairs are <slot id, colour space>
uint32_t
Drago::GetInterface(
    std::list<std::pair<uint32_t, uint32_t> > &mandatory,
    std::list<std::pair<uint32_t, uint32_t> > &optional) const
{   
    mandatory.clear();
    optional.clear();
        
    mandatory.push_back(
        std::pair<uint32_t, uint32_t>(IMG_PICTURE, CLS_CIE_XYY));

    return CLS_CIE_XYY;
}  

bool
Drago::SetImage(uint32_t type, const spectral::Image *img)
{
    bool result = false;

    if(type == IMG_PICTURE)
    {
        m_picture = NULL;
        if(img)
        {
            if(img->get_colour_space() == CLS_CIE_XYY)
            {
                m_picture = img;
                result = true;
            }
        }
        else
        {
            result = true;
        }
    }

    return result;
}

void
Drago::ClearImage(uint32_t type)
{
    if(type == IMG_PICTURE)
    {
        m_picture = NULL;
    }
}

void 
Drago::GetParameters(spectral::Parameters *params) const 
{
    if(params)
    {
        params->AddParameter(std::string("bias"), 0.5, 2.0, m_b);
        params->AddParameter(std::string("exposure"), -10.0, 10.0, m_exposure);
        params->AddParameter(std::string("samples"),
                                         100, 10000000, m_samples);
    }
}

// Apply parameters to a processor.
void
Drago::SetParameters(const spectral::Parameters *params)
{
    if(params)
    {
        params->GetParameter(std::string("bias"), m_b);
        params->GetParameter(std::string("exposure"), m_exposure);
        params->GetParameter(std::string("samples"), m_samples);
    }
}

spectral::Image *
Drago::Process(void)
{
    spectral::Image *result = NULL;
    Progress(0);

    if(m_picture)
    {
        result = new spectral::Image(*m_picture);

        if(result)
        {
            float max_Y(1.0e-10);
            float LwMax;
            size_t size = size_t(m_picture->get_width()) * size_t(m_picture->get_height());
            float adaptation;
            float exposure;

            // get adaptation & max luminance.
            adaptation = GetAdaptation(result, m_samples, max_Y);

            exposure = pow((float)2.0, m_exposure);

            adaptation/= exposure;

            LwMax = max_Y / adaptation;

            {
                float exponent, scale;
                size_t start, end;
                DragoFarmer farmer(this);

                // precalculated exponent for biasb term.
                exponent = log(m_b) / log(0.5);
          

                // full form of scale 
                // LdMax * 0.01 / log10(LwMax + 1)
                // We postmultiply by an exposure valuse, and assume for a TFT that
                // LdMax = 100  so simplify to...
                scale = 1.0 / log10(LwMax + 1.0); 

                start = end = size;

                while(end)
                {
                    if(end >= JOB_SIZE)
                    {
                        start-= JOB_SIZE;
                    }
                    else
                    {
                        start = 0;
                    }
                    farmer.Push(new DragoJob(result,
                                             scale,
                                             exponent,
                                             adaptation,
                                             LwMax,
                                             start,
                                             end));
                    end = start;
                }
                farmer.Process();
            }
        }
    }

    return result;
}

}
