#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "utils.h"
#include "image.h"
#include "parameters.h"

#include "lummixer.h"

const uint32_t IMG_1 = 0;
const uint32_t IMG_2 = 1;

namespace lummixer {

LuminanceMixer::LuminanceMixer()
    : spectral::Processor()
    , m_image1(NULL)
    , m_image2(NULL)
    , m_weight1(1.0)
    , m_weight2(1.0)
{
}

LuminanceMixer::~LuminanceMixer()
{
}

// Get slots & output type.  Returns output colour space.
// pairs are <slot id, colour space>
uint32_t
LuminanceMixer::GetInterface(
    std::list<std::pair<uint32_t, uint32_t> > &mandatory,
    std::list<std::pair<uint32_t, uint32_t> > &optional) const
{   
    mandatory.clear();
    optional.clear();
        
    mandatory.push_back(
        std::pair<uint32_t, uint32_t>(IMG_1, CLS_LUMINANCE));
    mandatory.push_back(
        std::pair<uint32_t, uint32_t>(IMG_2, CLS_LUMINANCE));
            
    return CLS_LUMINANCE;
}  

bool
LuminanceMixer::SetImage(uint32_t type, const spectral::Image *img)
{
    bool result = false;

    if(type == IMG_1)
    {
        m_image1 = NULL;
        if(img)
        {
            if(img->get_colour_space() == CLS_LUMINANCE)
            {
                m_image1 = img;
                result = true;
            }
        }
        else
        {
            result = true;
        }
    }
    if(type == IMG_2)
    {
        m_image2 = NULL;
        if(img)
        {
            if(img->get_colour_space() == CLS_LUMINANCE)
            {
                m_image2 = img;
                result = true;
            }
        }
        else
        {
            result = true;
        }
    }

    return result;
}

void
LuminanceMixer::ClearImage(uint32_t type)
{
    if(type == IMG_1)
    {
        m_image1 = NULL;
    }
    if(type == IMG_2)
    {
        m_image2 = NULL;
    }
}

void 
LuminanceMixer::GetParameters(spectral::Parameters *params) const 
{
    if(params)
    {
        params->AddParameter(std::string("weight 1"),
                             1e-5, 1.0, m_weight1);
        params->AddParameter(std::string("weight 2"),
                             1e-5, 1.0, m_weight2);
    }
}

// Apply parameters to a processor.
void
LuminanceMixer::SetParameters(const spectral::Parameters *params)
{
    if(params)
    {
        params->GetParameter("weight 1", m_weight1);
        params->GetParameter("weight 2", m_weight2);
    }
}

spectral::Image *
LuminanceMixer::Process(void)
{
    spectral::Image *result = NULL;

    Progress(0.0);

    if(m_image1 && m_image2)
    {
        if((m_image1->get_width() == m_image2->get_width()) &&
           (m_image1->get_height() == m_image2->get_height()))
        {
            result = new spectral::Image(m_image1->get_width(),
                                         m_image1->get_height(),
                                         CLS_LUMINANCE);
        }
    }

    if(result)
    {
        float total = m_weight1 + m_weight2;
        size_t size = size_t(m_image1->get_width()) *
                      size_t(m_image1->get_height());
        float *buf1, *buf2, *out;

	buf1 = m_image1->get_buffer();
	buf2 = m_image2->get_buffer();
	out = result->get_buffer();

        for(size_t i=0;i<size;i++)
	{
            if((i % 262144) == 0) // Once per 512*512 equiv.
            {
                float progress = i;
                progress/= size;
                progress*= 100;
                Progress(progress);
            }
	    out[i] = ((buf1[i] * m_weight1) + (buf2[i] * m_weight2)) / total;
	}
        Progress(100);
    }

    return result;
}

}
