#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include <utility>
#include <queue>
#include <vector>

#include "parameters.h"
#include "bufutils.h"

#include "mlfilter.h"

namespace multilevel {

const uint32_t IMG_LUMINANCE = 0;

////////////////////////////////////////////////////////////////////////////////
//
// MlHPassJob implementation.
//
////////////////////////////////////////////////////////////////////////////////
HPassJob::HPassJob(const spectral::Image *sourceImage,
                   const std::vector<float> &kernel,
                   uint32_t first, uint32_t last,
                   bool wrap,
                   spectral::Image *destImage)
    : m_sourceImage(sourceImage)
    , m_destImage(destImage)
    , m_kernel(kernel)
    , m_first(first)
    , m_last(last)
    , m_wrap(wrap)
    {
    }


void
HPassJob::Process(void)
{
    // TODO: Add range checks?
    int32_t min, max;
    max = (m_kernel.size() - 1);
    min = -max;
    int32_t x, y;

    for(y=m_first;y<(int32_t)m_last;y++) 
    { 
        for(x=0;(uint32_t)x<m_sourceImage->get_width();x++) 
        { 
            float value; 
            value = 0.0; 
            for(int32_t ki=min;ki<=max;ki++)
            { 
                int32_t xx; 
                xx = x + ki; 
                value+= ((*m_sourceImage->get_constrained_pixel(xx, y, 
                                                                m_wrap)) *
                         (m_kernel[abs(ki)]));
            }
            m_destImage->set_pixel(x, y, &value);
        }
    } 
}

////////////////////////////////////////////////////////////////////////////////
//
// VPassJob implementation.
//
////////////////////////////////////////////////////////////////////////////////
VPassJob::VPassJob(const spectral::Image *sourceImage,
                   const std::vector<float> &kernel,
                   uint32_t first, uint32_t last,
                   bool wrap,
                   spectral::Image *destImage)
    : m_sourceImage(sourceImage)
    , m_destImage(destImage)
    , m_kernel(kernel)
    , m_first(first)
    , m_last(last)
    , m_wrap(wrap)
    {
    }


void
VPassJob::Process(void)
{
    // TODO: Add range checks?
    int32_t min, max;
    max = (m_kernel.size() - 1);
    min = -max;
    int32_t x, y;

    for(y=m_first;y<(int32_t)m_last;y++) 
    { 
        for(x=0;(uint32_t)x<m_sourceImage->get_width();x++) 
        { 
            float value; 
            value = 0.0; 
            for(int32_t ki=min;ki<=max;ki++)
            { 
                int32_t yy; 
                yy = y + ki; 
                value+= ((*m_sourceImage->get_constrained_pixel(x, yy, false,
                                                                m_wrap)) *
                         (m_kernel[abs(ki)]));
            }
            m_destImage->set_pixel(x, y, &value);
        }
    } 
}

spectral::Image *
MlFilter::CreateLayer(const spectral::Image *input, uint32_t layer, float start_progress, float weight)
{
    spectral::Image *result(NULL);
    spectral::Image *tmp(NULL);
    float filter_size(layer);
    std::vector<float> kernel;


    {
        float total(0.0);
        float r(0.0);

        float value;
        // build filter kernel.
        do
        {
            value = (float)gaussian(filter_size, r);

            kernel.push_back(value);

            total+= value;

            r+= 1.0;
         } while(value > 1e-6);

        {
            // Normalise
            total-= kernel[0];
            total*= 2.0;
            total+= kernel[0];

            for(uint32_t i=0;i<kernel.size();i++)
            {
                kernel[i]/= total;
            }
        }
    }
    
    if(input)
    {
        tmp = new spectral::Image(*input);

        result = new spectral::Image(*input);
    }

    if(tmp && result)
    {
        const uint32_t JOB_SIZE = 32;

        uint32_t start(0);
    
        // Calculate weights of jobs for progress reporting.
        float job_value(weight);
        job_value/=((input->get_height() / JOB_SIZE) + 1);
        job_value/= 2.0;

        // Tell farmers about progress.
        MlFilterFarmer hfarmer(this, start_progress, job_value);
        MlFilterFarmer vfarmer(this, start_progress + (weight / 2.0), job_value);

        while(start < input->get_height())
        {
            uint32_t end;
            thread::Job *hjob, *vjob;
            
            end = start + JOB_SIZE;

            if(end > input->get_height())
            {
                end = input->get_height();
            }
            // Vreate jobs
            hjob = new HPassJob(input, kernel, start, end, m_wrap_horizontal, tmp);
            vjob = new VPassJob(tmp, kernel, start, end, m_wrap_vertical, result);
            
            hfarmer.Push(hjob);
            vfarmer.Push(vjob);
            
            start = end;
        }

        // Process image.
        hfarmer.Process();
        vfarmer.Process();
    }

    if(tmp)
    {
        delete tmp;
    }

    return result;
}

////////////////////////////////////////////////////////////////////////////////
//
// MlFilterFarmer implementation.
//
////////////////////////////////////////////////////////////////////////////////

MlFilterFarmer::MlFilterFarmer(MlFilter *owner,
                               float initialProgress,
                               float jobWeight)
    : thread::Farmer()
    , m_owner(owner)
    , m_initialProgress(initialProgress)
    , m_jobWeight(jobWeight)
{
}

void
MlFilterFarmer::ProcessCompletedJob(thread::Job *)
{
    if(m_owner)
    {
         float progress = m_initialProgress + (m_jobWeight * GetNumCompleted());
         m_owner->Progress(progress);
    }
}

////////////////////////////////////////////////////////////////////////////////
//
// MlFilter processor implementation.
//
////////////////////////////////////////////////////////////////////////////////
MlFilter::MlFilter()
    : spectral::Processor()
    , m_lumImage(NULL)
    , m_tolerance(0.3)
    , m_num_layers(4)
    , m_wrap_horizontal(false)
    , m_wrap_vertical(false)
    , m_density(false)
{
}

MlFilter::~MlFilter()
{
    ClearLayers();
}

// Get slots & output type.  Returns output colour space.
// pairs are <slot id, colour space>
uint32_t
MlFilter::GetInterface(
    std::list<std::pair<uint32_t, uint32_t> > &mandatory,
    std::list<std::pair<uint32_t, uint32_t> > &optional) const
{
    mandatory.clear();
    optional.clear();

    mandatory.push_back(
        std::pair<uint32_t, uint32_t>(IMG_LUMINANCE, CLS_LUMINANCE));

    return CLS_LUMINANCE;
}


bool
MlFilter::SetImage(uint32_t slot, const spectral::Image *img)
{
    if(slot == IMG_LUMINANCE)
    {
        m_lumImage = NULL;
        ClearLayers();

        if(img)
        {
            if(img->get_colour_space() == CLS_LUMINANCE)
            {
                m_lumImage = img;

                return true;
            }
        }
        else
        {
            return true;
        }
    }
    return false;
}

void
MlFilter::GetParameters(spectral::Parameters *params) const
{
    if(params)
    {
        std::string tmp;

        tmp = "number of layers";
        params->AddParameter(tmp, 1, 10, m_num_layers);

        tmp = "threshold";
        params->AddParameter(tmp, 0.1, 10.0, m_tolerance);
        
        tmp = "wrap horizontal";
        params->AddParameter(tmp, m_wrap_horizontal);

        tmp = "wrap vertical";
        params->AddParameter(tmp, m_wrap_vertical);

        tmp = "logarithmic";
        params->AddParameter(tmp, m_density);
    }
}

void
MlFilter::SetParameters(const spectral::Parameters *params)
{
    if(params)
    {
        std::string tmp;

        tmp = "number of layers";
        params->GetParameter(tmp, m_num_layers);

        tmp = "threshold";
        params->GetParameter(tmp, m_tolerance);
        
        tmp = "wrap horizontal";
        params->GetParameter(tmp, m_wrap_horizontal);

        tmp = "wrap vertical";
        params->GetParameter(tmp, m_wrap_vertical);

        tmp = "logarithmic";
        params->GetParameter(tmp, m_density);
    }
}

spectral::Image *
MlFilter::Process(void)
{
    spectral::Image *result(NULL);

    Progress(0.0);
    if(m_lumImage)
    {
        const spectral::Image *input(NULL);
        spectral::Image *nc_input(NULL);
        result = new spectral::Image(*m_lumImage);

        if(m_density)
        {
            nc_input = new spectral::Image(*m_lumImage);
            buf_luminance_to_density(nc_input->get_buffer(),
                                     nc_input->get_width() * 
                                     nc_input->get_height(),
                                     1.0e-10);
            input = nc_input;
        }
        else
        {
            input = m_lumImage;
        }
        
        if(result && input)
        {
            int32_t total_work(0), progress(0);

            if((uint32_t)m_num_layers > m_layers.size())
            {
                // Calculate total work to do.
                for(uint32_t i=m_layers.size();i<(uint32_t)m_num_layers;i++)
                {
                    int32_t layer_size;
                    layer_size = 1 << i;
                    total_work+= layer_size;
                }

                // Create any new layers required.
                {
                    uint32_t first_layer = m_layers.size();

                    // First create jobs.
                    for(uint32_t i=first_layer;i<(uint32_t)m_num_layers;i++)
                    {
                        int32_t layer_size;
                        spectral::Image *img;
                        layer_size = 1 << i;

                        float start_progress, weight;
                        
                        // Figure out how much this contributes to our
                        // progress reporting.
                        start_progress = progress;
                        start_progress/= total_work;
                        start_progress*= 90;
                        weight = layer_size;
                        weight/= total_work;
                        weight*= 90;

                        img = CreateLayer(input,
                                          layer_size,
                                          start_progress,
                                          weight);

                        progress+= layer_size;

                        m_layers.push_back(img);
                    }
                }
            }
            // Merge layers.
            {
                // TODO: Make this concurrent?
                size_t size = size_t(input->get_width()) *
                              size_t(input->get_height());

                float tolerance = m_tolerance * logf(2.0);
                if(m_density)
                {
                    for(size_t i=0;i<size;i++)
                    {
                        float value, total_weight, pixel;
                        pixel = input->get_buffer()[i];
                        value = pixel;
                        total_weight = 1.0;
    
                        for(size_t j=0;j<m_layers.size();j++)
                        {
                            if(m_layers[j])
                            {
                                float lvalue;
                                lvalue = m_layers[j]->get_buffer()[i];
    
                                if(fabs(lvalue - pixel) < tolerance)
                                {
                                    float weight, tmp;
                                    tmp = (lvalue - pixel) / tolerance;

                                    tmp = 1.0 - fabs(tmp);
                                    
                                    weight = ((float)(1 << j)) * tmp;
    
                                    value+= (lvalue * weight);
                                    total_weight+= weight;
                                }
                            }
                        }
                        result->get_buffer()[i] = value/total_weight;
                    }
                }
                else
                {
                    for(size_t i=0;i<size;i++)
                    {
                        float value, total_weight, pixel, log_pixel;
                        pixel = input->get_buffer()[i];
                        value = pixel;
                        total_weight = 1.0;
    
                        if(value < 1e-20)
                        {
                            value = 1e-20;
                        }
    
                        log_pixel = log(value);
    
                        for(size_t j=0;j<m_layers.size();j++)
                        {
                            if(m_layers[j])
                            {
                                float lvalue, log_lvalue;
                                lvalue = m_layers[j]->get_buffer()[i];
    
                                if(lvalue < 1e-20)
                                {
                                    lvalue = 1e-20;
                                }
    
                                log_lvalue = log(lvalue);
    
                                if(fabs(log_lvalue - log_pixel) < tolerance)
                                {
                                    float weight;
                                    float tmp = 1.0 - (fabs(log_lvalue - log_pixel)/
                                                      tolerance);
                                    weight = (float)(1 << j) * (tmp * tmp);
    
                                    value+= (lvalue * weight);
                                    total_weight+= weight;
                                }
                            }
                        }
                        result->get_buffer()[i] = value/total_weight;
                    }
                }
            }
        }
        if(nc_input)
        {
            delete nc_input;
            buf_density_to_luminance(result->get_buffer(),
                                     result->get_width()*result->get_height());
        }
    }
    Progress(100.0);
    return result;
}

void
MlFilter::ClearImage(uint32_t slot)
{
    if(slot == IMG_LUMINANCE)
    {
        m_lumImage = NULL;
        ClearLayers();
    }
}


void
MlFilter::ClearLayers(void)
{
    for(size_t i=0;i<m_layers.size();i++)
    {
        if(m_layers[i])
        {
            delete m_layers[i];
        }
    }
    m_layers.clear();
}

}

