#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include <utility>
#include <queue>
#include <vector>

#include "parameters.h"
#include "bufutils.h"

#include "mlfilter2.h"

namespace multilevel {

const uint32_t IMG_LUMINANCE(0);

const int32_t JOB_SIZE(16000);
const int32_t STRIP_SIZE(32);
////////////////////////////////////////////////////////////////////////////////
//
// Blur jobs.
//
////////////////////////////////////////////////////////////////////////////////

class HorizontalJob : public thread::Job
{
public:
    HorizontalJob(spectral::Image *input,
                  spectral::Image *output,
                  int32_t blur_size,
                  bool wrap_horizontal,
                  int32_t y1,
                  int32_t y2)
        : m_input(input)
        , m_output(output)
        , m_y1(y1)
        , m_y2(y2)
        , m_wrap_horizontal(wrap_horizontal)
        , m_blur_size(blur_size)
    {
    }
    virtual ~HorizontalJob()
    {
    }

    virtual void Process(void)
    {
        const uint32_t width(m_input->get_width());
        const uint32_t total_size((m_blur_size * 2) + 1);
        const uint32_t right(width - (m_blur_size + 1));
        const double weight(1.0/double(total_size));

        for(int32_t y=m_y1;y<m_y2;y++)
        {
            double total;
            float *pixel; 
            const float *front, *back;
            int32_t x;

            // first set up the line.
            pixel = m_output->get_buffer() + (y * width);
            back = m_input->get_buffer() + (y * width);
            front = back + m_blur_size;

            total = 0.0;

            for(x=(0 - (int32_t)m_blur_size);x<=(int32_t)m_blur_size;x++)
            {
                total+= *(m_input->get_constrained_pixel(x, y, m_wrap_horizontal, false));
            }

            if(total_size < width)
            {
                for(x=0;x<m_blur_size;x++)
                {
                    *pixel++ = float(total * weight);
                    total-= *(m_input->get_constrained_pixel(x-m_blur_size, y, m_wrap_horizontal, false));
                    total+= *++front;
                }
    
                for(x = m_blur_size;x<int32_t(right);x++)
                {
                    *pixel++ = float(total * weight);
                    total-= *back++;
                    total+= *++front;
                }
    
                for(x = right;x<int32_t(width);x++)
                {
                    *pixel++ = float(total * weight);
                    total-= *back++;
                    total+= *(m_input->get_constrained_pixel(x+m_blur_size+1, y, m_wrap_horizontal, false));
                }
            }
            else
            { 
                for(x=0;x<int32_t(m_input->get_width());x++)
                {
                    *pixel++ = float(total*weight);
                    total-= *(m_input->get_constrained_pixel(x-m_blur_size, y, m_wrap_horizontal, false));
                    total+= *(m_input->get_constrained_pixel(x+m_blur_size+1, y, m_wrap_horizontal, false));
                }
            }
        }
    }

    spectral::Image *m_input;
    spectral::Image *m_output;
    int32_t m_y1, m_y2;
    bool m_wrap_horizontal;
    int32_t m_blur_size;
};

class VerticalJob : public thread::Job
{
public:
    VerticalJob(spectral::Image *input,
                  spectral::Image *output,
                  int32_t blur_size,
                  bool wrap_vertical,
                  int32_t x1,
                  int32_t x2)
        : m_input(input)
        , m_output(output)
        , m_x1(x1)
        , m_x2(x2)
        , m_wrap_vertical(wrap_vertical)
        , m_blur_size(blur_size)
    {
    }
    virtual ~VerticalJob()
    {
    }

    virtual void Process(void)
    {
        const uint32_t width(m_input->get_width());
        const uint32_t height(m_input->get_height());
        const uint32_t total_size((m_blur_size * 2) + 1);
        const uint32_t bottom(height - (m_blur_size + 1));
        const double weight(1.0/double(total_size));

        for(int32_t x=m_x1;x<m_x2;x++)
        {
            double total;
            float *pixel; 
            const float *front, *back;
            int32_t y;

            // first set up the line.
            pixel = m_output->get_buffer() + x;
            back = m_input->get_buffer() + x;
            front = back + (m_blur_size * width);

            total = 0.0;

            for(y=(0 - (int32_t)m_blur_size);y<=(int32_t)m_blur_size;y++)
            {
                total+= *(m_input->get_constrained_pixel(x, y, false, m_wrap_vertical));
            }

            if(0)
            {
                for(y=0;y<m_blur_size;y++)
                {
                    *pixel = float(total * weight);
                    pixel+= width;

                    total-= *(m_input->get_constrained_pixel(x-m_blur_size, y, false, m_wrap_vertical));

                    front+= width;
                    total+= *front;
                }
    
                for(y = m_blur_size;y<int(bottom);y++)
                {
                    *pixel = float(total * weight);
                    pixel+= width;

                    total-= *back;
                    back+= width;

                    front+= width;
                    total+= *front;
                }
    
                for(y = bottom;y<int32_t(height);y++)
                {
                    *pixel = float(total * weight);
                    pixel+= width;

                    total-= *back;
                    back+= width;

                    total+= *(m_input->get_constrained_pixel(x+m_blur_size+1, y, false, m_wrap_vertical));
                }
            }
            else
            { 
                for(y=0;y<int32_t(m_input->get_height());y++)
                {
                    *pixel = float(total*weight);
                    pixel+= width;
                    // move forward.
                    total-= *(m_input->get_constrained_pixel(x, y-m_blur_size, false, m_wrap_vertical));
                    total+= *(m_input->get_constrained_pixel(x, y+m_blur_size+1, false, m_wrap_vertical));
                }
            }
        }
/*
        const double weight((m_blur_size * 2) + 1);
        for(int32_t x=m_x1;x<m_x2;x++)
        {
            double total;
            float *p;
            total = 0.0;
            int32_t y;
            // first set up the line.
            for(y=(0 - (int32_t)m_blur_size);y<=(int32_t)m_blur_size;y++)
            {
                total+= *(m_input->get_constrained_pixel(x, y, false, m_wrap_vertical));
            }
            for(y=0;y<m_input->get_height();y++)
            {
                float value;
                // set pixel.
                value = total / weight;
                m_output->set_pixel(x, y, &value);
                // move forward.
                total-= *(m_input->get_constrained_pixel(x, y-m_blur_size, false, m_wrap_vertical));
                total+= *(m_input->get_constrained_pixel(x, y+m_blur_size+1, false, m_wrap_vertical));
            }
        }
  */
  }

    spectral::Image *m_input;
    spectral::Image *m_output;
    int32_t m_x1, m_x2;
    bool m_wrap_vertical;
    int32_t m_blur_size;
};

////////////////////////////////////////////////////////////////////////////////
//
// Job for calculating weight.
//
////////////////////////////////////////////////////////////////////////////////
class CalculateWeightsJob : public thread::Job
{
public:
    CalculateWeightsJob(const spectral::Image *reference,
                 const spectral::Image *image,
                 spectral::Image *result,
                 spectral::Image *weight,
                 size_t begin,
                 size_t end,
                 float tolerance)
         : m_reference(reference)
         , m_image(image)
         , m_result(result)
         , m_weight(weight)
         , m_begin(begin)
         , m_end(end)
         , m_tolerance(tolerance)
    {
    }
    virtual ~CalculateWeightsJob()
    {
    }

    virtual void Process(void)
    {
        const float *image_buf = m_image->get_buffer();
        const float *reference_buf = m_reference->get_buffer();
        float *result_buf = m_result->get_buffer();
        float *weight_buf = m_weight->get_buffer();

        for(size_t j=m_begin;j<m_end;j++)
        {
            float tmp;
            tmp = 1.0 - (fabs(image_buf[j] - reference_buf[j])/m_tolerance);

            if(tmp > 0.0)
            {
                weight_buf[j]+= tmp;
                result_buf[j]+= (image_buf[j] * tmp); 
            }
        }
    }
private:
    const spectral::Image *m_reference;
    const spectral::Image *m_image;
    spectral::Image *m_result;
    spectral::Image *m_weight;
    uint32_t m_begin;
    uint32_t m_end;
    float m_tolerance;
};

////////////////////////////////////////////////////////////////////////////////
//
// Job for applying weight.
//
////////////////////////////////////////////////////////////////////////////////
class ApplyWeightsJob : public thread::Job
{
public:
    ApplyWeightsJob(spectral::Image *result,
                    spectral::Image *weight,
                    size_t begin,
                    size_t end)
         : m_result(result)
         , m_weight(weight)
         , m_begin(begin)
         , m_end(end)
    {
    }
    virtual ~ApplyWeightsJob()
    {
    }

    virtual void Process(void)
    {
        float *result_buf = m_result->get_buffer();
        float *weight_buf = m_weight->get_buffer();

        for(uint32_t j=m_begin;j<m_end;j++)
        {
            result_buf[j]/= weight_buf[j];
        }
    }
private:
    spectral::Image *m_result;
    spectral::Image *m_weight;
    size_t m_begin;
    size_t m_end;
};

////////////////////////////////////////////////////////////////////////////////
//
// MlFilter2 processor implementation.
//
////////////////////////////////////////////////////////////////////////////////
MlFilter2::MlFilter2()
    : spectral::Processor()
    , m_lumImage(NULL)
    , m_tolerance(0.3)
    , m_num_layers(10)
    , m_passes(5)
    , m_layer_scale(1)
    , m_wrap_horizontal(false)
    , m_wrap_vertical(false)
{
}

MlFilter2::~MlFilter2()
{
}

// Get slots & output type.  Returns output colour space.
// pairs are <slot id, colour space>
uint32_t
MlFilter2::GetInterface(
    std::list<std::pair<uint32_t, uint32_t> > &mandatory,
    std::list<std::pair<uint32_t, uint32_t> > &optional) const
{
    mandatory.clear();
    optional.clear();

    mandatory.push_back(
        std::pair<uint32_t, uint32_t>(IMG_LUMINANCE, CLS_LUMINANCE));

    return CLS_LUMINANCE;
}


bool
MlFilter2::SetImage(uint32_t slot, const spectral::Image *img)
{
    if(slot == IMG_LUMINANCE)
    {
        m_lumImage = NULL;

        if(img)
        {
            if(img->get_colour_space() == CLS_LUMINANCE)
            {
                m_lumImage = img;

                return true;
            }
        }
        else
        {
            return true;
        }
    }
    return false;
}

void
MlFilter2::GetParameters(spectral::Parameters *params) const
{
    if(params)
    {
        std::string tmp;

        tmp = "number of layers";
        params->AddParameter(tmp, 1, 100, m_num_layers);

        tmp = "scale";
        params->AddParameter(tmp, 1, 50, m_layer_scale);

        tmp = "threshold";
        params->AddParameter(tmp, 0.1, 10.0, m_tolerance);
        
        tmp = "passes";
        params->AddParameter(tmp, 1, 10, m_passes);
        
        tmp = "wrap horizontal";
        params->AddParameter(tmp, m_wrap_horizontal);

        tmp = "wrap vertical";
        params->AddParameter(tmp, m_wrap_vertical);
    }
}

void
MlFilter2::SetParameters(const spectral::Parameters *params)
{
    if(params)
    {
        std::string tmp;

        tmp = "number of layers";
        params->GetParameter(tmp, m_num_layers);

        tmp = "scale";
        params->GetParameter(tmp, m_layer_scale);

        tmp = "threshold";
        params->GetParameter(tmp, m_tolerance);
        
        tmp = "passes";
        params->GetParameter(tmp, m_passes);
        
        tmp = "wrap horizontal";
        params->GetParameter(tmp, m_wrap_horizontal);

        tmp = "wrap vertical";
        params->GetParameter(tmp, m_wrap_vertical);
    }
}

spectral::Image *
MlFilter2::Process(void)
{
    spectral::Image *result(NULL);

    {
        spectral::Image *input(NULL);
        spectral::Image *weight(NULL);
        spectral::Image *reference(NULL);
        spectral::Image *tmp(NULL);

        input = new spectral::Image(*m_lumImage);
	weight = new spectral::Image(input->get_width(), input->get_height(), CLS_LUMINANCE);
	tmp = new spectral::Image(input->get_width(), input->get_height(), CLS_LUMINANCE);

        size_t size = size_t(input->get_width()) *
                      size_t(input->get_height());

        buf_luminance_to_density(input->get_buffer(),
                                 input->get_width() * 
                                 input->get_height(),
                                 1.0e-10);

        reference = new spectral::Image(*input);
        result = new spectral::Image(*reference);

        for(size_t j=0;j<size;j++)
        {
            weight->get_buffer()[j] = 1;
        }

        if(result && input && weight && reference)
        {
            //CalculateWeights(reference, reference, result, weight, m_tolerance);

            for(uint32_t i=0;i<uint32_t((m_num_layers>0)?m_num_layers:0);i++)
            {
                Progress(((100.0 * float(i))/float(m_num_layers)));

                {
                    input = FastBlur(input, &tmp, (i+1) * m_layer_scale, m_passes);
                }

                // now apply weights.
                CalculateWeights(reference, input, result, weight, m_tolerance);
            }
            
            ApplyWeights(result, weight);

            delete input;
            delete weight;
            delete reference;
            delete tmp;
        }

        buf_density_to_luminance(result->get_buffer(),
                                 result->get_width()*result->get_height());
    }
    Progress(100.0);
    return result;
}

void
MlFilter2::ClearImage(uint32_t slot)
{
    if(slot == IMG_LUMINANCE)
    {
        m_lumImage = NULL;
    }
}


spectral::Image *
MlFilter2::FastBlur(spectral::Image *img, spectral::Image **tmp, uint32_t blur_size, uint32_t passes)
{
    spectral::Image *result(NULL);

    if(img)
    {
        spectral::Image *tmp1 = img;
        spectral::Image *tmp2 = *tmp;

        if(tmp1 && tmp2)
        {
            // First do horizontal scan lines.
            for(uint32_t pass = 0;pass<passes; pass++)
            {
                uint32_t y;
                {
                    thread::Farmer farmer;
                    y = 0;
                    do
                    {
                        uint32_t next_y;
                        thread::Job *job;
                        next_y = y + STRIP_SIZE;
                        if(next_y > img->get_height())
                        {
                            next_y = img->get_height();
                        }
                        // Make job
                        job = new HorizontalJob(tmp1, tmp2, blur_size, m_wrap_horizontal, y, next_y);
                        farmer.Push(job);

                        y = next_y;
                    } while(y < img->get_height());
                    farmer.Process();
                }

                // swap buffers
                {
                    spectral::Image *tmp3;
                    tmp3 = tmp1;
                    tmp1 = tmp2;
                    tmp2 = tmp3;
                }
            }

            // Now vertical
            for(uint32_t pass = 0;pass<passes; pass++)
            {
                uint32_t x;
                {
                    thread::Farmer farmer;
                    x = 0;
                    do
                    {
                        uint32_t next_x;
                        thread::Job *job;
                        next_x = x + STRIP_SIZE;
                        if(next_x > img->get_width())
                        {
                            next_x = img->get_width();
                        }
                        // Make job
                        job = new VerticalJob(tmp1, tmp2, blur_size, m_wrap_vertical, x, next_x);
                        farmer.Push(job);
                        x = next_x;
                    } while(x < img->get_width());
                    farmer.Process();
                }
                // swap buffers
                {
                    spectral::Image *tmp3;
                    tmp3 = tmp1;
                    tmp1 = tmp2;
                    tmp2 = tmp3;
                }
            }
            
            *tmp = tmp2;
            result = tmp1;
        }
    }
    return result;
}

void 
MlFilter2::CalculateWeights(const spectral::Image *reference, 
                            const spectral::Image *image,
                            spectral::Image *result,
                            spectral::Image *weight,
                            float tolerance)
{
    if(reference && image && result && weight)
    {
        thread::Farmer farmer;
        float use_tol;
        use_tol = tolerance*log(2.0);

        size_t size = size_t(image->get_width()) *
                      size_t(image->get_height());
        size_t i = 0;

        do
        {
            thread::Job *job;
            size_t next_i;
            next_i = i + JOB_SIZE;

            if(next_i > size)
            {
                next_i = size;
            }
            job = new CalculateWeightsJob(reference, image, result, weight, i, next_i, use_tol);

            farmer.Push(job);

            i = next_i;
        } while(i < size);

        farmer.Process();
    }
}

void 
MlFilter2::ApplyWeights(spectral::Image *result, spectral::Image *weight)
{
    if(result && weight)
    {
        thread::Farmer farmer;

        size_t size = size_t(result->get_width()) *
                      size_t(result->get_height());

        size_t i = 0;

        do
        {
            thread::Job *job;
            size_t next_i;
            next_i = i + JOB_SIZE;

            if(next_i > size)
            {
                next_i = size;
            }
            job = new ApplyWeightsJob(result, weight, i, next_i);

            farmer.Push(job);

            i = next_i;
        } while(i < size);

        farmer.Process();
    }
}

}

