#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "utils.h"
#include "thread.h"
#include "image.h"
#include "parameters.h"

#include "photographic.h"

const uint32_t IMG_PICTURE = 0;
const uint32_t IMG_LAL = 1;
const uint32_t JOB_SIZE = 160000;// 400 * 400

static long sp_lroundf(float val)
{
#ifdef _WIN32
    float fv = floor(val);
    float cv = ceil(val);
    if((cv - val) <= (val - fv))
    {
        return cv;
    }
    else
    {
        return fv;
    }
#else
    return lroundf(val);
#endif
}
 

namespace photographic {

class PhotoJob : public thread::Job
{
public:
    PhotoJob(spectral::Image *img, 
             const spectral::Image *lal,
             float key,
             float contrast,
             float adaptation,
             size_t start,
             size_t end)
        : m_img(img)
        , m_lal(lal)
        , m_key(key)
        , m_contrast(contrast)
        , m_adaptation(adaptation)
        , m_start(start)
        , m_end(end)
    {
    }

    virtual void Process(void);
private:
    inline float L(float adaptation, float key, float l)
    {
        return (l * key)/adaptation;
    }

    spectral::Image *m_img;
    const spectral::Image *m_lal;
    float m_key, m_contrast, m_adaptation;
    size_t m_start, m_end;
};


void 
PhotoJob::Process(void)
{
    if(m_img)
    {
        float *buf = m_img->get_buffer();
        size_t index = (m_start * 3) + 2;

        if(m_lal)
        {
            float *lal = m_lal->get_buffer();

            for(size_t i=m_start;i<m_end;i++)
            {
                float scaled_lum, scaled_lal, result, v, diff;
                scaled_lum = L(m_adaptation, m_key, buf[index]);
                scaled_lal = L(m_adaptation, m_key, lal[i]);
                // boost contrast.
                diff = scaled_lum - scaled_lal;

                diff *= m_contrast;
                v = scaled_lum + diff;
                
                result = v / (1.0 + scaled_lum);
                if(result > 0.0)
                {
                    buf[index] = result;
                }
                else
                {
                    buf[index] = 0.0;
                }

                index+=3;
            }
        }
        else
        {
            for(size_t i=m_start;i<m_end;i++)
            {
                float scaled_lum;
                scaled_lum = L(m_adaptation, m_key, buf[index]);
                buf[index] = scaled_lum / (1.0 + scaled_lum);
                index+=3;
            }
        }
    }
}

class PhotoFarmer : public thread::Farmer
{
public:
    PhotoFarmer(Photographic *owner)
    : m_owner(owner)
    {}

    virtual void ProcessCompletedJob(thread::Job *);
private:
    Photographic *m_owner;
};


void
PhotoFarmer::ProcessCompletedJob(thread::Job *)
{
    float progress = GetNumCompleted() + 1;
    progress/= GetNumSubmitted();
    progress*= 100.0;
    if(m_owner)
    {
        m_owner->Progress(progress);
    }
}

Photographic::Photographic()
    : spectral::Processor()
    , m_picture(NULL)
    , m_lal(NULL)
    , m_key(0.18)
    , m_contrast(1.0)
    , m_samples(2000)
{
}

Photographic::~Photographic()
{
}

// Get slots & output type.  Returns output colour space.
// pairs are <slot id, colour space>
uint32_t
Photographic::GetInterface(
    std::list<std::pair<uint32_t, uint32_t> > &mandatory,
    std::list<std::pair<uint32_t, uint32_t> > &optional) const
{   
    mandatory.clear();
    optional.clear();
        
    mandatory.push_back(     
        std::pair<uint32_t, uint32_t>(IMG_PICTURE, CLS_CIE_XYY));
            
    optional.push_back(     
        std::pair<uint32_t, uint32_t>(IMG_LAL, CLS_LUMINANCE));
    return CLS_CIE_XYY;
}  

bool
Photographic::SetImage(uint32_t type, const spectral::Image *img)
{
    bool result = false;

    if(type == IMG_PICTURE)
    {
        m_picture = NULL;
        if(img)
        {
            if(img->get_colour_space() == CLS_CIE_XYY)
            {
                m_picture = img;
                result = true;
            }
        }
        else
        {
            result = true;
        }
    }

    if(type == IMG_LAL)
    {
        m_lal = NULL;
        if(img)
        {
            if(img->get_colour_space() == CLS_LUMINANCE)
            {
                m_lal = img;
                result = true;
            }
        }
        else
        {
            result = true;
        }
    }

    return result;
}

void
Photographic::ClearImage(uint32_t type)
{
    if(type == IMG_PICTURE)
    {
        m_picture = NULL;
    }
}

void 
Photographic::GetParameters(spectral::Parameters *params) const 
{
    if(params)
    {
        params->AddParameter(std::string("key"),
                             0.0, 1.0, m_key);
        params->AddParameter(std::string("contrast"),
                             0.0, 10.0, m_contrast);
        params->AddParameter(std::string("samples"),
                             100, 10000000, m_samples);
    }
}

// Apply parameters to a processor.
void
Photographic::SetParameters(const spectral::Parameters *params)
{
    if(params)
    {
        params->GetParameter(std::string("key"), m_key);
        params->GetParameter(std::string("contrast"), m_contrast);
        params->GetParameter(std::string("samples"), m_samples);
    }
}

spectral::Image *
Photographic::Process(void)
{
    spectral::Image *result = NULL;

    Progress(0);
    if(m_picture)
    {
        result = new spectral::Image(*m_picture);
    }

    if(result)
    {
        size_t size = size_t(result->get_width()) *
                      size_t(result->get_height());
        size_t start, end;
        float adaptation = GetAdaptation(result, m_samples);
        const spectral::Image *lal= NULL;

        fprintf(stderr, "adaptation = %f\n", adaptation);

        PhotoFarmer farmer(this);

        start = end = size;
        if(m_lal &&
           (m_lal->get_width() == result->get_width()) &&
           (m_lal->get_height() == result->get_height()))
        {
            // lal image must match source image.
            lal = m_lal;
        } 

        while(end)
        {
            if(end >= JOB_SIZE)
            {
                start = end - JOB_SIZE;
            }
            else
            {
                start = 0;
            }
            farmer.Push(new PhotoJob(result,  
                                     lal,
                                     m_key,
                                     m_contrast,
                                     adaptation, 
                                     start,
                                     end));

            end = start;
        }
        farmer.Process();    
    }
    return result;
}

// estimate log average luminance of the image
float
Photographic::GetAdaptation(spectral::Image *img, uint32_t num_samples)
{
    const float sigma = 1.0e-6;
    float result = sigma;
    if(!num_samples)
    {
        num_samples = 1;
    }
    if(img)
    {
        uint32_t width, height;
        float total(0.0);

        width = img->get_width();
        height = img->get_height();
        
        for(uint32_t i=0;i<num_samples;i++)
        {
            float fx, fy;
            int32_t x, y;

            fx = halton(3, i);
            fy = halton(5, i);
            fx*= width;
            fy*= height;

            x = sp_lroundf(fx);
            y = sp_lroundf(fy);

            if(x<0)x=0;
            if(y<0)y=0;
            if((uint32_t)x>=width)x=(int32_t)(width-1);
            if((uint32_t)y>=height)y=(int32_t)(height-1);

            {
                const float *pixel = img->get_pixel(x, y);
                total+= log(sigma + pixel[2]);
            }
        }
        total/= num_samples;
        result = exp(total);
    }
    return result;
}

}
