#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <unistd.h>

#include "utils.h"
#include "bufutils.h"
#include "thread.h"
#include "image.h"
#include "parameters.h"

#include "photoreceptor.h"

const uint32_t IMG_PICTURE = 0;
const uint32_t JOB_SIZE = 160000;// 400 * 400

namespace photoreceptor {

////////////////////////////////////////////////////////////////////////////////

class PrJob : public thread::Job
{
public:
    PrJob(const spectral::Image *xyy,
          spectral::Image *rgb,
          float m,
          float f,
          float c,
          float a,
          float av,
          float *cav,
          size_t start,
          size_t end)
        : m_xyy(xyy)  // reference luminance
        , m_rgb(rgb)  // rgb image being tonemapped
        , m_m(m)      // contrast
        , m_f(f)      // intensity
        , m_c(c)      // chromatic adaptation
        , m_a(a)      // light adaptation
        , m_av(av)
        , m_cav(cav)
        , m_start(start)
        , m_end(end)
        , m_max(0.0)
    {
    }

    virtual void Process(void);

    float GetMax(void) { return m_max; }
private:
    const spectral::Image *m_xyy;
    spectral::Image *m_rgb;
    float m_m, m_f, m_c, m_a, m_av, *m_cav;
    size_t m_start, m_end;
    float m_max;
};

void
PrJob::Process(void)
{
    if(m_xyy && m_rgb)
    {
        float *rgb_buf = m_rgb->get_buffer();
        const float *lum_buf = m_xyy->get_buffer();

        uint32_t lum_idx = 2 + (m_start * 3);
        uint32_t rgb_idx = (m_start * 3);

        for(size_t i=m_start;i<m_end;i++)
        {
            for(size_t j=0;j<3;j++)
            {
                float Il, Ig, Ia, val;
                Il = (m_c * rgb_buf[rgb_idx]) + ((1.0 - m_c) * lum_buf[lum_idx]);
                Ig = (m_c * m_cav[j]) + ((1.0 - m_c) * m_av);
                Ia = (m_a * Il) + ((1.0 - m_a) * Ig);
                val = rgb_buf[rgb_idx];
                val/= val + pow(m_f * Ia, m_m);

                if(val > m_max) 
                {
                    m_max = val;
                }

                rgb_buf[rgb_idx] = val;
 
                rgb_idx++;
            }
            lum_idx+= 3;
        }
    }
}

////////////////////////////////////////////////////////////////////////////////

class PrFarmer : public thread::Farmer
{
public:
    PrFarmer(Photoreceptor *owner)
    : m_owner(owner)
    , m_max(0.0)
    {}

    virtual void ProcessCompletedJob(thread::Job *);

    float GetMax(void) { return m_max; }

private:
    Photoreceptor *m_owner;
    float m_max;
};


void
PrFarmer::ProcessCompletedJob(thread::Job *job)
{
    float progress = GetNumCompleted() + 1;
    progress/= GetNumSubmitted();

    progress*= 60.0;
    progress+= 20.0;

    PrJob *pjob = static_cast<PrJob*>(job);

    if(m_max < pjob->GetMax())
    {
        m_max = pjob->GetMax();
    }

    if(m_owner)
    {
        m_owner->Progress(progress);
    }
}

////////////////////////////////////////////////////////////////////////////////

Photoreceptor::Photoreceptor()
    : spectral::Processor()
    , m_picture(NULL)
    , m_samples(2000)
    , m_contrast(0.5)
    , m_saturation(0.5)
    , m_adaptation(1.0)
{
}

Photoreceptor::~Photoreceptor()
{
}

// Get slots & output type.  Returns output colour space.
// pairs are <slot id, colour space>
uint32_t
Photoreceptor::GetInterface(
    std::list<std::pair<uint32_t, uint32_t> > &mandatory,
    std::list<std::pair<uint32_t, uint32_t> > &optional) const
{   
    mandatory.clear();
    optional.clear();
        
    mandatory.push_back(     
        std::pair<uint32_t, uint32_t>(IMG_PICTURE, CLS_CIE_XYY));
            
    return CLS_CIE_XYY;
}  

bool
Photoreceptor::SetImage(uint32_t type, const spectral::Image *img)
{
    bool result = false;

    if(type == IMG_PICTURE)
    {
        m_picture = NULL;
        if(img)
        {
            if(img->get_colour_space() == CLS_CIE_XYY)
            {
                m_picture = img;
                result = true;
            }
        }
        else
        {
            result = true;
        }
    }

    return result;
}

void
Photoreceptor::ClearImage(uint32_t type)
{
    if(type == IMG_PICTURE)
    {
        m_picture = NULL;
    }
}

void 
Photoreceptor::GetParameters(spectral::Parameters *params) const 
{
    if(params)
    {
        params->AddParameter(std::string("samples"),
                             100, 10000000, m_samples);

        params->AddParameter(std::string("contrast"),
                             0.0, 1.0, m_contrast);

        params->AddParameter(std::string("intensity"),
                             -15.0, 15.0, m_intensity);

        params->AddParameter(std::string("adaptation"),
                             0.0, 1.0, m_adaptation);

        params->AddParameter(std::string("saturation"),
                             0.0, 1.0, m_saturation);

    }
}

// Apply parameters to a processor.
void
Photoreceptor::SetParameters(const spectral::Parameters *params)
{
    if(params)
    {
        params->GetParameter(std::string("samples"), m_samples);
        
        params->GetParameter(std::string("contrast"), m_contrast);

        params->GetParameter(std::string("intensity"), m_intensity);

        params->GetParameter(std::string("adaptation"), m_adaptation);

        params->GetParameter(std::string("saturation"), m_saturation);
    }
}

spectral::Image *
Photoreceptor::Process(void)
{
    spectral::Image *result = NULL;

    Progress(0);
    if(m_picture)
    {
        result = new spectral::Image(*m_picture);
    }

    if(result)
    {
        size_t size = size_t(result->get_width()) *
                      size_t(result->get_height());
        size_t start, end;
        float logmin(0), logmax(0), logav(0), av(0);
        GetAdaptation(result, m_samples, logmin, logmax, logav, av);

        if(logmax != logmin)
        {
            //float k = (logmax - logav)/(logmax - logmin);
            float m = 0.5 - (m_contrast * 2);
            float f = exp((-m_intensity)/log(2)); // intensity parameter  f = exp(-f') -8<=f'<=8
            float c = m_saturation; // chromatic adaptation.
            float a = m_adaptation;

            float cav[3], logcav[3];

            buf_xyy_to_linear_srgb(result->get_buffer(), size);
           
            Progress(20); 
            GetAverageRGB(result, m_samples, cav, logcav);
  
            PrFarmer farmer(this);
    
            start = end = size;
    
            while(end)
            {
                if(end >= JOB_SIZE)
                {
                    start-= JOB_SIZE;
                }
                else
                {
                    start = 0;
                }
                farmer.Push(new PrJob(m_picture, result, 
                                      m, f, c, a, av, cav, start, end));
                //farmer.Push(new PrJob(m_picture, result, 
                //                      m, f, c, a, logav, logcav, start, end));

                end = start;
            }

            farmer.Process();

            if(farmer.GetMax() > 1.0e-10)
            {
                buf_scale(result->get_buffer(), size*3, 1.0/farmer.GetMax());
            }

            buf_linear_srgb_to_xyy(result->get_buffer(), size);
            Progress(100);
        }
    }
    return result;
}

// estimate log average luminance of the image
void
Photoreceptor::GetAdaptation(spectral::Image *img,
                             uint32_t num_samples,
                             float &logmin,
                             float &logmax,
                             float &logav,
                             float &av)
{
    const float sigma = 1.0e-6;
    //float average = sigma;

    logmin = logmax = logav = av = 0;

    if(!num_samples)
    {
        num_samples = 1;
    }
    if(img)
    {
        uint32_t width, height;
        float total(0.0), logtotal(0.0);

        width = img->get_width();
        height = img->get_height();
        
        for(uint32_t i=0;i<num_samples;i++)
        {
            float fx, fy;
            int32_t x, y;

            fx = halton(3, i);
            fy = halton(5, i);
            fx*= width;
            fy*= height;

            x = lroundf(fx);
            y = lroundf(fy);
            
            if(x<0)x=0;
            if(y<0)y=0;
            if(x>=int32_t(width))x=(width-1);
            if(y>=int32_t(height))y=(height-1);
            {
                float logval;
                const float *pixel = img->get_pixel(x, y);
                logval = log(sigma + pixel[2]);
                logtotal+= logval;
                total+= pixel[2];

                if((logval < logmin) || (i == 0))
                {
                    logmin = logval;
                }
                if((logval > logmax) || (i == 0))
                {
                    logmax = logval;
                }
            }
        }
        logav = logtotal/num_samples;
        av = total/num_samples;
    }
}

void
Photoreceptor::GetAverageRGB(spectral::Image *img, uint32_t num_samples, float *result, float *logresult)
{
    const float sigma = 1.0e-6;
    if(!num_samples)
    {
        num_samples = 1;
    }

    if(img && result && logresult)
    {
        uint32_t width, height;
        result[0] = result[1] = result[2] = 0.0;

        width = img->get_width();
        height = img->get_height();
        
        for(uint32_t i=0;i<num_samples;i++)
        {
            float fx, fy;
            int32_t x, y;

            fx = halton(3, i);
            fy = halton(5, i);
            fx*= width;
            fy*= height;

            x = lroundf(fx);
            y = lroundf(fy);
            
            if(x<0)x=0;
            if(y<0)y=0;
            if(x>=int32_t(width))x=(width-1);
            if(y>=int32_t(height))y=(height-1);

            {
                const float *pixel = img->get_pixel(x, y);
                for(uint32_t c=0;c<3;c++)
                {
                    float logval;
                    result[c]+= pixel[c];
                    logval = log(sigma + pixel[c]);
                    logresult[c]+= logval;
                }
            }
        }
        for(uint32_t c=0;c<3;c++)
        {
            result[c]/= num_samples;
            logresult[c]/= num_samples;
        }
    }
}

}
