#include "mapping.h"
#include <math.h>

namespace remap {

////////////////////////////////////////////////////////////////////////////////
bool
Mapping::LatLongToVector(float lat, float lon, float *vector) const
{
    if(vector)
    {
        vector[0] = sin(lon) * cos(lat); // X - Horizontal
        vector[1] = cos(lon) * cos(lat); // Y - Horizontal
        vector[2] = sin(lat);            // Z is to zenith.
        return true;
    }
    else
    {
        fprintf(stderr, "LatLongToVector failed\n");
        return false;
    }
}

bool 
Mapping::VectorToLatLong(const float *vector, float &lat, float &lon)  const
{
    if(vector)
    {
        float vec[3];
        int i;
        float len(0);
        for(i=0;i<3;i++)
        {
            vec[i] = vector[i];
            len+=(vec[i] * vec[i]);
        }

        len = sqrt(len);

        if(len > 1.0e-5)
        {
            // Normalise
            for(i=0;i<3;i++)
            {
                vec[i]/= len;
            }
            lat = asin(vec[2]);
            lon = atan2(vec[0], vec[1]);
            return true;
        }
        else
        {
            fprintf(stderr, "VectorToLatLong failed\n");
            return false;
        }
    }
    else
    {
        fprintf(stderr, "VectorToLatLong failed\n");
        return false;
    }
}

////////////////////////////////////////////////////////////////////////////////
Equirectangular::Equirectangular(int xres, int yres, float phi1)
    : m_xres(xres)
    , m_yres(yres)
    , m_phi1(phi1)
{
}

Equirectangular::~Equirectangular()
{
}

bool
Equirectangular::MapTo(float lat, float lon, float &x, float &y) const
{
    bool valid(true);
    
    if(fabs(lat) > PI_OVER_TWO)
    {
        valid = false;
    }
    else
    {
//        lon = fmod(lon, 2.0 * PI);
//
//        if(lon < 0.0)
//        {
//            lon+= 2.0 * PI;
//        }

        float centre_x = (float)m_xres / 2.0;
        float centre_y = (float)m_yres / 2.0;

        float xx, yy;

        xx = lon * cos(m_phi1);
        yy = lat;

        xx/= PI;
        yy/= PI_OVER_TWO;

        xx*= centre_x;
        yy*= centre_y;

        x = xx + centre_x;
        y = yy + centre_y;
    }
    return valid;
}

bool 
Equirectangular::MapFrom(float x, float y, float &lat, float &lon) const
{
    bool valid(true);

    if((x > m_xres) || 
       (x < 0) || 
       (y > m_yres) || 
       (y < 0))
    {
        valid = false;
    }
    else
    {
        float xx, yy;
        float centre_x = (float)m_xres / 2.0;
        float centre_y = (float)m_yres / 2.0;
        xx = ((x - centre_x)/centre_x) * PI;
        yy = ((y - centre_y)/centre_y) * PI_OVER_TWO;

        lat = yy;
        lon = xx / cos(m_phi1);
    }

    return valid;
}       

////////////////////////////////////////////////////////////////////////////////
AngularFisheye::AngularFisheye(float centre_x, float centre_y, 
                               float radius)
    : m_centre_x(centre_x)
    , m_centre_y(centre_y)
    , m_zenith_x(centre_x)
    , m_zenith_y(centre_y)
    , m_radius(radius)
{
    matrix_identity(&m_forward_transform);
    matrix_identity(&m_inverse_transform);
}

AngularFisheye::~AngularFisheye()
{
}

bool 
AngularFisheye::MapTo(float lat, float lon, float &x, float &y) const
{
    float vec[3];

    if(LatLongToVector(lat, lon, vec))
    {
        float angle;

        transform_vector(m_forward_transform, vec);

        angle = acos(vec[2]);
        
        //vector[0] = sin(lon) * cos(lat); // X is right/left
        //vector[1] = cos(lon) * cos(lat); // Y up/down
        //vector[2] = sin(lat);            // Z is to zenith.

        if(angle <= PI_OVER_TWO)
        {
            float xx = vec[0];
            float yy = vec[1];

            float distance = (angle / PI_OVER_TWO) * m_radius;

            {
                float len = sqrt((xx*xx) + (yy*yy));
                xx/= len;
                yy/= len;
            }

            x = m_centre_x + (xx * distance);
            y = m_centre_y + (yy * distance);

            return true;
        }
        else
        {
            //fprintf(stderr, "AngularFisheye::MapTo - invalid angle (%f > %f)\n",
            //      angle, PI_OVER_TWO);
            return false;
        }
    }
    else
    {
        fprintf(stderr, "AngularFisheye::MapTo - invalid lat lon(%f > %f)\n",
                lat, lon);
        return false;
    }
}

bool 
AngularFisheye::MapFrom(float x, float y, float &lat, float &lon) const
{
    float r;
    
    float xx, yy;

    xx = (x-m_centre_x);
    yy = (y-m_centre_y);

    xx/= m_radius;
    yy/= m_radius;

    r = sqrt((xx * xx) + (yy * yy));
    
    // normalise
    xx/= r;
    yy/= r;

    if(r <= 1.0)
    {
        float angle;

        angle = r * PI_OVER_TWO;

        float vec[3];

        if(angle > 1.0e-8)
        {
            // rotate (angle) in the direction xx, yy
            vec[0] = xx * sin(angle);
            vec[1] = yy * sin(angle);
            vec[2] = cos(angle);
        }
        else
        {
            // Straight ahead.
            vec[0] = vec[1] = 0.0;
            vec[2] = 1.0;
        }

        transform_vector(m_inverse_transform, vec);

        return VectorToLatLong(vec, lat, lon);
    }
    else
    {
        return false;
    }
}


bool
AngularFisheye::SetZenith(float x, float y)
{
    float xd, yd, r;

    xd = x - m_centre_x;
    yd = y - m_centre_y;

    r = sqrt((xd*xd) + (yd*yd));

    if(r < m_radius)
    {
        float lat, lon;
        m_zenith_x = x;
        m_zenith_y = y;

        matrix_identity(&m_forward_transform);
        matrix_identity(&m_inverse_transform);

        if(MapFrom(x, y, lat, lon))
        {
            float k[3];
            if(LatLongToVector(lat, lon, k))
            {
                float i[3] = {0.0, 0.0, 0.0};
                float j[3] = {0.0, 0.0, 0.0};
    
                {
                    int min_idx(0);
                    float min(k[0]);
    
                    for(int n=1;n<3;n++)
                    {
                        if(k[n] < min)
                        {
                            min_idx = n;
                            min = k[n];
                        }
                    }
                    j[min_idx] = 1.0;
                }
    
                // i^j = k
                // j^k = i
                // k^i = j
    
                vector_product(j, k, i);
                vector_product(k, i, j);
    
                vector_normalise(i);
                vector_normalise(j);
                vector_normalise(k);
    
                for(int n=0;n<3;n++)
                {
                    m_inverse_transform[0][n] = i[n];
                    m_inverse_transform[1][n] = j[n];
                    m_inverse_transform[2][n] = k[n];
                }
                matrix_inverse(m_inverse_transform, &m_forward_transform);
    
                return true;
            }
            else
            {
                return false;
            }
        }
        else
        {
            return false;
        }
    }
    else
    {
        return false;
    }
}

void 
AngularFisheye::GetCentre(float &x, float &y) const
{
    x = m_centre_x;
    y = m_centre_y;
}

void
AngularFisheye::GetZenith(float &x, float &y) const
{
    x = m_zenith_x;
    y = m_zenith_y;
}

////////////////////////////////////////////////////////////////////////////////

float dp2(float *v1, float *v2)
{
    return (v1[0] * v2[0]) + (v1[1] * v2[1]);
}

void setvec(float x1, float y1, float x2, float y2, float *vec)
{
    vec[0] = x2 - x1;
    vec[1] = y2 - y1;
}

void
circumcircle(float x1, float y1, float x2, float y2, float x3, float y3,
             float &cx, float &cy, float &radius)
{
    // From graphics gems.
    float d1, d2, d3, c1, c2, c3, c;
    {
        float vec1[2], vec2[2];
        setvec(x1, y1, x3, y3, vec1);
        setvec(x1, y1, x2, y2, vec2);
        d1 = dp2(vec1, vec2);

        setvec(x2, y2, x3, y3, vec1);
        setvec(x2, y2, x1, y1, vec2);
        d2 = dp2(vec1, vec2);

        setvec(x3, y3, x1, y1, vec1);
        setvec(x3, y3, x2, y2, vec2);
        d3 = dp2(vec1, vec2);
    }
    c1 = d2 * d3;
    c2 = d3 * d1;
    c3 = d1 * d2;
    c = c1 + c2 + c3;

    radius = sqrt(((d1+d2)*(d2+d3)*(d3+d1))/c)/2.0;
    cx = ((x1*(c2+c3)) + (x2*(c3+c1)) + (x3*(c1+c2)))/(c*2.0);
    cy = ((y1*(c2+c3)) + (y2*(c3+c1)) + (y3*(c1+c2)))/(c*2.0);
}

}

