#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "utils.h"
#include "thread.h"
#include "bufutils.h"
#include "image.h"
#include "sampler.h"
#include "parameters.h"
#include "mapping.h"

#include "remap.h"

const uint32_t IMG_PICTURE = 0;

#define ONE_EIGHTY_OVER_PI 57.2957795
#ifndef PI
#define PI 3.14159265
#endif

const uint32_t SAMPLE_NEAREST = 0;
const uint32_t SAMPLE_LANCZOS_1 = 1;
const uint32_t SAMPLE_LANCZOS_2 = 2;
const uint32_t SAMPLE_LANCZOS_3 = 3;

namespace remap {

class RemapJob : public thread::Job
{
public:
    RemapJob(const spectral::Image *src,
             const Mapping &src_map,
             const Mapping &dst_map,
             uint32_t samples,
             const spectral::ImageSampler *sampler,
             const float *background,
             uint32_t y1,
             uint32_t y2,
             spectral::Image *result)
        : m_src(src)
        , m_src_map(src_map)
        , m_dst_map(dst_map)
        , m_samples(samples)
        , m_sampler(sampler)
        , m_background(background)
        , m_y1(y1)
        , m_y2(y2)
        , m_result(result)
    {
    }

    virtual void Process(void);
private:
    const spectral::Image        *m_src;
    const Mapping      &m_src_map;
    const Mapping      &m_dst_map;
    uint32_t            m_samples;
    const spectral::ImageSampler *m_sampler;
    const float        *m_background;
    uint32_t            m_y1,
                        m_y2;
    spectral::Image              *m_result;
};


void
RemapJob::Process(void)
{
    uint32_t width = m_result->get_width();
    uint32_t height = m_result->get_height();
    //uint32_t channels = m_src->get_channels();

    for(uint32_t y=m_y1;y<m_y2;y++)
    {
        for(uint32_t x=0;x<width;x++)
        {
            float val[3];

            val[0] = val[1] = val[2] = 0.0;

            for(uint32_t i=0;i<m_samples;i++)
            {
                float x1, y1;
                float lat, lon;

                // TODO: Replace random sampling with something 
                // more sensible.  e.g n rooks
                if(m_samples > 1)
                {
                    x1 = rand();
                    x1/= RAND_MAX;
                    y1 = rand();
                    y1/= RAND_MAX;
                    x1+= x;
                    y1+= y;
                    y1 = (float)height - y1;
                }
                else
                {
                    // Centre sampling.
                    x1 = 0.5 + (float)x;
                    y1 = 0.5 + (float)y;
                }

                if(m_dst_map.MapFrom(x1, y1, lat, lon))
                {
                    //printf("%f, %f -> %f, %f\n", x1, y1, lat, lon);
                    float x2, y2;

                    if(m_src_map.MapTo(lat, lon, x2, y2))
                    {
                        float p[3];
                        m_sampler->SampleImage(m_src, x2, y2, m_background, p);
                    
                        val[0]+= p[0];
                        val[1]+= p[1];
                        val[2]+= p[2];
                    }
                    else
                    {
                        // Outside fisheye.
                        val[0]+= m_background[0];
                        val[1]+= m_background[1];
                        val[2]+= m_background[2];
                    }
                }
                else
                {
                    // Some sort of mapping screwup?
                    val[0]+= m_background[0];
                    val[1]+= m_background[1];
                    val[2]+= m_background[2];
                }
            }
            // Result!
            val[0]/=(float)m_samples;
            val[1]/=(float)m_samples;
            val[2]/=(float)m_samples;
            m_result->set_pixel(x, y, val);
        }
    }
}

class RemapFarmer : public thread::Farmer
{
public:
    RemapFarmer(Remap *owner)
    : m_owner(owner)
    {
    }

    virtual ~RemapFarmer()
    {
    }

    virtual void ProcessCompletedJob(thread::Job *);
private:
    Remap *m_owner;
};

void
RemapFarmer::ProcessCompletedJob(thread::Job *)
{
    float progress = GetNumCompleted() + 1;
    progress/= GetNumSubmitted();
    progress*= 100.0;
    if(m_owner)
    {
        m_owner->Progress(progress);
    }
}



Remap::Remap()
    : spectral::Processor()
    , m_picture(NULL)
    , m_x1(0.0)
    , m_y1(0.0)
    , m_x2(1.0)
    , m_y2(0.0)
    , m_x3(0.0)
    , m_y3(1.0)
    , m_zenith_x(0.0)
    , m_zenith_y(0.0)
    , m_phi1(0.0)
    , m_auto_width(false)
    , m_move_zenith(false)
    , m_width(1024)
    , m_samples(10)
    , m_bg_cover(2)
    , m_filter(SAMPLE_LANCZOS_1)
{
}

Remap::~Remap()
{
}

// Get slots & output type.  Returns output colour space.
// pairs are <slot id, colour space>
uint32_t
Remap::GetInterface(
    std::list<std::pair<uint32_t, uint32_t> > &mandatory,
    std::list<std::pair<uint32_t, uint32_t> > &optional) const
{
    mandatory.clear();
    optional.clear();

    mandatory.push_back(
        std::pair<uint32_t, uint32_t>(IMG_PICTURE, CLS_CIE_XYY));

    return CLS_CIE_XYY;
}

bool
Remap::SetImage(uint32_t type, const spectral::Image *img)
{
    bool result = false;

    if(type == IMG_PICTURE)
    {
        m_picture = NULL;
        if(img)
        {
            if(img->get_colour_space() == CLS_CIE_XYY)
            {
                m_picture = img;
                result = true;
            }
        }
        else
        {
            result = true;
        }
    }

    return result;
}

void
Remap::ClearImage(uint32_t type)
{
    if(type == IMG_PICTURE)
    {
        m_picture = NULL;
    }
}

void
Remap::GetParameters(spectral::Parameters *params) const
{
    if(params)
    {
        params->AddParameter(std::string("x1"), 0.0, 1e5, m_x1);
        params->AddParameter(std::string("y1"), 0.0, 1e5, m_y1);
        params->AddParameter(std::string("x2"), 0.0, 1e5, m_x2);
        params->AddParameter(std::string("y2"), 0.0, 1e5, m_y2);
        params->AddParameter(std::string("x3"), 0.0, 1e5, m_x3);
        params->AddParameter(std::string("y3"), 0.0, 1e5, m_y3);

        params->AddParameter(std::string("auto size"),  m_auto_width);
        params->AddParameter(std::string("width"), 256, 4096, m_width);

        params->AddParameter(std::string("move zenith"), m_move_zenith);
        params->AddParameter(std::string("zenith x"), 0.0, 1e5, m_zenith_x);
        params->AddParameter(std::string("zenith y"), 0.0, 1e5, m_zenith_y);

        params->AddParameter(std::string("samples"), 5, 100, m_samples);
        {
            std::list<std::pair<uint32_t, std::string> > options;

            options.push_back(
                std::pair<uint32_t, std::string>
                    (SAMPLE_NEAREST, "nearest"));
            options.push_back(
                std::pair<uint32_t, std::string>
                    (SAMPLE_LANCZOS_1, "lanczos 1"));
            options.push_back(
                std::pair<uint32_t, std::string>
                    (SAMPLE_LANCZOS_2, "lanczos 2"));
            options.push_back(
                std::pair<uint32_t, std::string>
                    (SAMPLE_LANCZOS_3, "lanczos 3"));

            params->AddParameter(std::string("filter"), options, m_filter);
        }

        params->AddParameter(std::string("background range"), 1, 10, m_bg_cover);
    }
}

// Apply parameters to a processor.
void
Remap::SetParameters(const spectral::Parameters *params)
{
    if(params)
    {
        std::string tmp;

        params->GetParameter(std::string("x1"), m_x1);
        params->GetParameter(std::string("y1"), m_y1);
        params->GetParameter(std::string("x2"), m_x2);
        params->GetParameter(std::string("y2"), m_y2);
        params->GetParameter(std::string("x3"), m_x3);
        params->GetParameter(std::string("y3"), m_y3);

        params->GetParameter(std::string("auto size"), m_auto_width);
        params->GetParameter(std::string("width"), m_width);

        params->GetParameter(std::string("move zenith"), m_move_zenith);
        params->GetParameter(std::string("zenith x"), m_zenith_x);
        params->GetParameter(std::string("zenith y"), m_zenith_y);

        params->GetParameter(std::string("samples"), m_samples);
        params->GetParameter(std::string("filter"), m_filter, tmp);
        
        params->GetParameter(std::string("background range"), m_bg_cover);
    }
}

spectral::Image *
Remap::Process(void)
{
    spectral::Image *result = NULL;

    if(m_picture)
    {
        float centre_x, centre_y, radius;
        uint32_t width, height;

        // Construct circle for fisheye.
        circumcircle(m_x1, m_y1, m_x2, m_y2, m_x3, m_y3,
                     centre_x, centre_y, radius);

        // Calculate size.
        if(m_auto_width)
        {
            height = ceil(radius) * 2;
            width = height * 2;
        }
        else
        {
            width = m_width;
            height = width / 2;
            if(height & 1)
            {
                height++;
                width = height * 2;
            }
        }

        // Temporary XYZ src image, as xyY is unsuitable for averaging.
        spectral::Image *tmp = new spectral::Image(*m_picture);

        if(tmp)
        {
            result = new spectral::Image(width, height, CLS_CIE_XYY);

            if(result)
            {
                float background[3];

                buf_xyy_to_xyz(tmp->get_buffer(),
                               tmp->get_width() * tmp->get_height());

                fprintf(stderr, "fish eye centred on (%f, %f), radius %f\n",
                        centre_x, centre_y, radius);

                // Set up transformations.
                AngularFisheye  src_map(centre_x, centre_y, radius);
                Equirectangular dst_map(width, height, m_phi1/ONE_EIGHTY_OVER_PI);

                if(m_move_zenith)
                {
                    if(!src_map.SetZenith(m_zenith_x, m_zenith_y))
                    {
                        fprintf(stderr, "Unable to set up zenith mapping\n");
                    }
                }

                // Get background colour.
                {
                    float cv = m_bg_cover;
                    cv/= 100;
                    GetBackground(centre_x, centre_y, radius, cv, 500, tmp, background);
                }

                // Set up image sampling.
                spectral::ImageSampler *sampler(NULL);
                switch(m_filter)
                {
                case SAMPLE_LANCZOS_1:
                    sampler = new spectral::LanczosSampler(
                        spectral::ImageSampler::BACKGROUND,
                        spectral::ImageSampler::BACKGROUND,
                        1.0);
                    break;
                case SAMPLE_LANCZOS_2:
                    sampler = new spectral::LanczosSampler(
                        spectral::ImageSampler::BACKGROUND,
                        spectral::ImageSampler::BACKGROUND,
                        2.0);
                    break;
                case SAMPLE_LANCZOS_3:
                    sampler = new spectral::LanczosSampler(
                        spectral::ImageSampler::BACKGROUND,
                        spectral::ImageSampler::BACKGROUND,
                        3.0);
                    break;
                default:
                    sampler = new spectral::NearestNeighbourSampler(
                        spectral::ImageSampler::BACKGROUND,
                        spectral::ImageSampler::BACKGROUND);
                }

                // Perform remapping.
                RemapFarmer farmer(this);

                {
                    uint32_t y = 0;

                    while(y < result->get_height())
                    {
                        uint32_t next_y;
                        thread::Job *job;
                        next_y = y + 16;
                        if(next_y > result->get_height())
                        {
                            next_y = result->get_height();
                        }
                        job = new RemapJob(tmp,
                                           src_map,
                                           dst_map,
                                           m_samples,
                                           sampler,
                                           background,
                                           y,
                                           next_y,
                                           result);
                        farmer.Push(job);
                        y = next_y;
                    }
                }
                farmer.Process();

                delete sampler;

                // Convert back to xyY
                buf_xyz_to_xyy(result->get_buffer(), width * height);
            }
            delete tmp;
        }
    }
    return result;
}

//
// Calculate a background colour by averaging the outside of the
// fisheye circle.
//
void
Remap::GetBackground(float x, float y, float r,
                     float coverage,
                     uint32_t samples,
                     const spectral::Image *img,
                     float *bg)
{
    uint32_t i, j;
    float minr, dr;
    float samples_taken(0);

    for(j=0;j<3;j++)
    {
        bg[j] = 0.0;
    }
    if(coverage > 1.0)
    {
       coverage = 1.0;
    }
    if(coverage < 0.0)
    {
        coverage = 0.0;
    }
    minr = r * (1.0 - coverage);
    dr = r - minr;

    for(i=0;i<samples;i++)
    {
        // This is a non uniform mapping that biases strongly toward the centre
        // of the disk.  This is OK though, as we are only sampling a thin band
        // on the outside where it should be roughly uniform.
        float angle;
        float radius;
        float xx, yy;
        int ix, iy;

        angle = halton(3, i) * 2.0 * PI;

        radius = halton(5, i);
        radius*= dr;
        radius+= minr;

        xx = x + (sin(angle) * radius);
        yy = y + (cos(angle) * radius);

        // TODO: Replace this with interpolation.
        ix = round_float(xx);
        iy = round_float(yy);

        if((ix < (int)img->get_width()) &&
           (iy < (int)img->get_height()))
        {
            const float *p = img->get_pixel(ix, iy);
            for(j=0;j<3;j++)
            {
                bg[j]+=p[j];
                //bg[j] += log(1.0e-6 + p[j]);
                samples_taken++;
            }
        }
    }
    if(samples_taken)
    {
        for(j=0;j<3;j++)
        {
            bg[j]/= samples_taken;
        }
    }
}


}
