#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include <utility>
#include <queue>
#include <vector>
#include <string>

#include "utils.h"
#include "parameters.h"
#include "image.h"
#include "thread.h"

#include "yee.h"
namespace segment {

// spectral::Image slots
const uint32_t IMG_LUMINANCE = 0;

// Fill search patterns
const uint32_t SEGMENT_SCAN_SMALL   = 0;
const uint32_t SEGMENT_SCAN_MEDIUM  = 1;
const uint32_t SEGMENT_SCAN_LARGE   = 2;

////////////////////////////////////////////////////////////////////////////////
//
// Fill utilities
//
////////////////////////////////////////////////////////////////////////////////


// 4th order rotational symmetry transformations for building fill patterns.
static void sym4(int32_t x, int32_t y,
                 std::vector<std::pair<int32_t, int32_t> > &out)
{
    out.push_back(std::pair<int32_t,int32_t>( x, y));
    out.push_back(std::pair<int32_t,int32_t>( y,-x));
    out.push_back(std::pair<int32_t,int32_t>(-x, y));
    out.push_back(std::pair<int32_t,int32_t>(-y, x));
}

// Defines a pending fill point.
class FillFrame
{
public:
    FillFrame(int32_t x, int32_t y)
    : m_x(x)
    , m_y(y)
    {
    }
    int32_t m_x, m_y;
};

////////////////////////////////////////////////////////////////////////////////
//
// Thread support
//
////////////////////////////////////////////////////////////////////////////////

class LayerJob : public thread::Job
{
    class GroupInfo
    {
    public:
       GroupInfo() : m_numPixels(0), m_totalValue(0), m_meanValue(0)
       {
       }
       uint32_t m_numPixels;
       float    m_totalValue;
       float    m_meanValue;
    };

public:

    LayerJob(const spectral::Image *density, 
             float binSize, 
             uint32_t pattern,
             bool wrap_horizontal,
             bool wrap_vertical);
    ~LayerJob();
    virtual void Process(void);

    spectral::Image *GetResult(void);

private:
    void SegmentImage(const spectral::Image *density, spectral::Image *out);
    int get_bin_id(float density);
    void flood_fill(const spectral::Image &density,
                    const spectral::image<int> &bins,
                    int bin_id,
                    int group_id,
                    uint32_t x,
                    uint32_t y,
                    std::vector<std::pair<int32_t, int32_t> > &,
                    spectral::image<int> &group_image,
                    GroupInfo &new_group,
                    std::vector<int> *neighbours = NULL);
    uint32_t m_pattern;
    const spectral::Image *m_density;
    spectral::Image *m_result;
    float m_binSize;
    bool m_wrap_horizontal, m_wrap_vertical;
};

////////////////////////////////////////////////////////////////////////////////

class LayerFarmer : public thread::Farmer
{
public:
    LayerFarmer(spectral::Image *result, Yee *owner);
    virtual ~LayerFarmer();
    virtual void ProcessCompletedJob(thread::Job *);

    void Progress(float);
private:
    spectral::Image *m_result;
    Yee *m_owner;
};

////////////////////////////////////////////////////////////////////////////////
//
// LayerJob implementation
//
////////////////////////////////////////////////////////////////////////////////
LayerJob::LayerJob(const spectral::Image *density,
                   float binSize,
                   uint32_t pattern,
                   bool wrap_horizontal,
                   bool wrap_vertical)
    : m_pattern(pattern)
    , m_density(density)
    , m_result(NULL)
    , m_binSize(binSize)
    , m_wrap_horizontal(wrap_horizontal)
    , m_wrap_vertical(wrap_vertical)
{
}

LayerJob::~LayerJob()
{
    if(m_result)
    {
        delete m_result;
    }
}

spectral::Image *
LayerJob::GetResult(void)
{
    return m_result;
}

// Job entry point.
void
LayerJob::Process(void)
{
    if(m_density)
    {
        m_result = new spectral::Image(m_density->get_width(),
                             m_density->get_height(),
                             m_density->get_colour_space());
    }
    if(m_result)
    {
        SegmentImage(m_density, m_result);
    }
}

// Create a density layer for this image.
void
LayerJob::SegmentImage(const spectral::Image *density, spectral::Image *out)
{
    if(density && out)
    {
        uint32_t width, height, size;
        uint32_t x, y, group_id;
        std::vector<std::pair<int32_t, int32_t> > scanRange;
        std::vector<GroupInfo> groups;
   
        width = density->get_width();
        height = density->get_height();
        size = size_t(width) * size_t(height);

        spectral::image<int> group_image(width, height, 1);
        spectral::image<int> bin_image(width, height, 1);

        // set up flood fill search pattern.
        switch(m_pattern)
        {
        case 2:
            sym4(0, 5, scanRange);
            sym4(2, 4, scanRange);
            sym4(4, 2, scanRange);
        case 1:
            sym4(0, 3, scanRange);
            sym4(2, 2, scanRange);
        default:
            sym4(0, 1, scanRange);
        }

        // Set up bin image.
        {
            int *bin_buf = bin_image.get_buffer();
            const float *density_buf = density->get_buffer();
            for(uint32_t i=0;i<size;i++)
            {
                bin_buf[i] = get_bin_id(density_buf[i]);
            }
        }

        // clear the group set.
        {
            int *buf(group_image.get_buffer());
            for(uint32_t i=0;i<size;i++)
            {
                buf[i] = -1;
            }
        }
        // build groups

        group_id = 0;

        for(y=0;y<height;y++)
        {
            for(x=0;x<width;x++)
            {
                if((*group_image.get_pixel(x, y)) == -1)
                {
                    GroupInfo new_group;
                    int bin_id;

    		    bin_id = *(bin_image.get_pixel(x,y));

                    flood_fill((*density),
                               bin_image,
                               bin_id,
                               group_id,
                               x, y, scanRange,
                               group_image,
                               new_group);

                    groups.push_back(new_group);

                    group_id++;
                }
            }
        }

        // TODO: Move into GroupInfo class
        for(uint32_t i=0;i<groups.size();i++)
        {
            groups[i].m_meanValue = 
                groups[i].m_totalValue / (float)(groups[i].m_numPixels);
        }
        // TODO: Assimilation step.

        {
            float *buf = out->get_buffer();
            int *grp = group_image.get_buffer();
            for(uint32_t i = 0;i<size;i++)
            {
                buf[i] = groups[grp[i]].m_meanValue;
            }
        }
    }
}

// Queue based flood fill.  Recursive fills fail horribly on large images.
void
LayerJob::flood_fill(const spectral::Image &density,
                     const spectral::image<int> &bins,
                     int bin_id,
                     int group_id,
                     uint32_t x,
                     uint32_t y,
                     std::vector<std::pair<int32_t, int32_t> > &scanRange,
                     spectral::image<int> &group_image,
                     GroupInfo &new_group,
                     std::vector<int> * /*neighbours*/)
{
    uint32_t width = group_image.get_width();
    uint32_t height = group_image.get_height();
    std::queue<FillFrame> queue;

    // get image buffers.
    float *density_buf = density.get_buffer();
    int *group_buf = group_image.get_buffer();
    const int *bin_buf = bins.get_buffer();

    if((x < width) && (y < height) && 
       (*(group_image.get_pixel(x, y)) == -1) &&
       (*(bins.get_pixel(x, y)) == bin_id))
    {
        // Set first pixel
        new_group.m_numPixels++;
        new_group.m_totalValue+= *(density.get_pixel(x, y));

        group_image.set_pixel(x, y, &group_id);

        // push neighbouring pixels
        std::vector<std::pair<int32_t, int32_t> >::const_iterator it;
        for(it=scanRange.begin();it<scanRange.end();it++)
        {
            int32_t nx, ny;

            nx = ((int32_t)x) + (it->first);
            ny = ((int32_t)y) + (it->second);

            if(m_wrap_horizontal)
            {
                if(nx < 0)
                {
                    nx+= width;
                }
                else if((uint32_t)nx >= width)
                {
                    nx-= width;
                }
            }
            if(m_wrap_vertical)
            {
                if(ny < 0)
                {
                    ny+= height;
                }
                else if((uint32_t)ny >= height)
                {
                    ny-= height;
                }
            }

            if((nx >= 0) && (ny >= 0) && 
               ((uint32_t)nx < width) && ((uint32_t)ny < height))
            {
                int32_t index;
                index = (nx + (ny * width));

                if((group_buf[index] == -1) &&
                   (bin_buf[index] == bin_id))
                {
                    new_group.m_numPixels++;
                    new_group.m_totalValue+= density_buf[index];
                    group_buf[index] = group_id;

                    queue.push(FillFrame(nx, ny));
                }
            }
        }
    }

    while(queue.size())
    {
        int32_t cur_x, cur_y;    

        {
            FillFrame pixel(queue.front());
            queue.pop();
            cur_x = pixel.m_x;
            cur_y = pixel.m_y;
        }

        // Proceed
        std::vector<std::pair<int32_t, int32_t> >::const_iterator it;
        for(it=scanRange.begin();it<scanRange.end();it++)
        {
            int32_t nx, ny;
            int32_t index;
            
            nx = cur_x + (it->first);
            ny = cur_y + (it->second);

            index = (nx + (ny * width));

            if(m_wrap_horizontal)
            {
                if(nx < 0)
                {
                    nx+= width;
                }
                else if((uint32_t)nx >= width)
                {
                    nx-= width;
                }
            }
            if(m_wrap_vertical)
            {
                if(ny < 0)
                {
                    ny+= height;
                }
                else if((uint32_t)ny >= height)
                {
                    ny-= height;
                }
            }

            if((nx >= 0) && (ny >= 0) && 
               ((uint32_t)nx < width) && ((uint32_t)ny < height))
            {
                if((group_buf[index] == -1) &&
                   (bin_buf[index] == bin_id))
                {
                    new_group.m_numPixels++;
                    new_group.m_totalValue+= density_buf[index];
                    group_buf[index] = group_id;

                    queue.push(FillFrame(nx, ny));
                }
                /*
                else if(neighbours)
                {
                    neighbours->push_back(current);
                }*/
            }
        }
    }
}

int
LayerJob::get_bin_id(float density)
{
    return round_float(density / m_binSize);
}

////////////////////////////////////////////////////////////////////////////////
//
// LayerFarmer implementation
//
////////////////////////////////////////////////////////////////////////////////

LayerFarmer::LayerFarmer(spectral::Image *result, Yee *owner)
    : thread::Farmer()
    , m_result(result)
    , m_owner(owner)
{
}

LayerFarmer::~LayerFarmer()
{
}

void LayerFarmer::Progress(float progress)
{
    if(m_owner)
    {
        m_owner->Progress(progress);
    }
}

void LayerFarmer::ProcessCompletedJob(thread::Job *p)
{
    LayerJob *job = dynamic_cast<LayerJob *>(p);

    if(job && m_result)
    {
        spectral::Image *img = job->GetResult();

        if(img)
        {
            float *job_buf = img->get_buffer();
            float *out_buf = m_result->get_buffer();
            size_t size = size_t(img->get_width()) * size_t(img->get_height());

            // Add the new layer to the result.
            for(size_t i=0;i<size;i++)
            {
                out_buf[i]+=job_buf[i];
            }
        }
        {
            float progress = (float)GetNumCompleted() + 1.0;
            progress/= (float)GetNumSubmitted();
            progress*= 100.0;
            Progress(progress);
        }
    }
}

////////////////////////////////////////////////////////////////////////////////
//
// Yee processor implementation.
//
////////////////////////////////////////////////////////////////////////////////
Yee::Yee() 
    : spectral::Processor()
    , m_lumImage(NULL)
    , m_min(0.1)
    , m_max(2.0)
    , m_first_layer_id(0)
    , m_last_layer_id(50)
    , m_pattern(SEGMENT_SCAN_SMALL)
    , m_wrap_horizontal(false)
    , m_wrap_vertical(false)
{
}

Yee::~Yee()
{
}

// Get slots & output type.  Returns output colour space.
// pairs are <slot id, colour space>
uint32_t
Yee::GetInterface(
    std::list<std::pair<uint32_t, uint32_t> > &mandatory,
    std::list<std::pair<uint32_t, uint32_t> > &optional) const
{
    mandatory.clear();
    optional.clear();

    mandatory.push_back(
        std::pair<uint32_t, uint32_t>(IMG_LUMINANCE, CLS_LUMINANCE));
        
    return CLS_LUMINANCE;
}   

//Populate a spectral::Parameters object.
void
Yee::GetParameters(spectral::Parameters *params) const
{
    if(params)
    {
        params->AddParameter(std::string("min bin size"), 0.1, 8.0, m_min); 
        params->AddParameter(std::string("max bin size"), 0.1, 8.0, m_max); 
        params->AddParameter(std::string("number of layers"), 
                             1, 1000, m_last_layer_id);
        {
            std::list<std::pair<uint32_t, std::string> >range;
            range.push_back(std::pair<uint32_t, std::string>
                (SEGMENT_SCAN_SMALL, std::string("small")));
            range.push_back(std::pair<uint32_t, std::string>
                (SEGMENT_SCAN_MEDIUM, std::string("medium")));
            range.push_back(std::pair<uint32_t, std::string>
                (SEGMENT_SCAN_LARGE, std::string("large")));

            params->AddParameter(std::string("search pattern"), range,
                                 m_pattern);
        }
        params->AddParameter(std::string("wrap horizontal"),m_wrap_horizontal); 
        params->AddParameter(std::string("wrap vertical"), m_wrap_vertical); 
    }
}
    
// Apply parameters to a processor.
void 
Yee::SetParameters(const spectral::Parameters *params)
{
    if(params)
    {
        std::string tmp;
        params->GetParameter(std::string("min bin size"), m_min);
        params->GetParameter(std::string("max bin size"), m_max);
        params->GetParameter(std::string("number of layers"), m_last_layer_id);
        params->GetParameter(std::string("pattern"), m_pattern, tmp);
        params->GetParameter(std::string("wrap horizontal"),m_wrap_horizontal); 
        params->GetParameter(std::string("wrap vertical"), m_wrap_vertical); 
    }
}

bool
Yee::SetImage(uint32_t slot, const spectral::Image *img)
{
    if(slot == IMG_LUMINANCE)
    {
        m_lumImage = NULL;
        if(img)
        {
            if(img->get_colour_space() == CLS_LUMINANCE)
            {
                m_lumImage = new spectral::Image(*img);
    
                return true;
            }
        }
        else
        {
            return true;
        }
    }
    return false;
}

spectral::Image *
Yee::Process(void)
{
    spectral::Image *result(NULL);

    if(m_lumImage)
    {
        spectral::Image *density(NULL);
        uint32_t width, height, size;
        width = m_lumImage->get_width();
        height = m_lumImage->get_height();
        size = size_t(width) * size_t(height);

        float min, max;
        min = max = *(m_lumImage->get_pixel(0, 0));

        density = new spectral::Image(width, height, CLS_DENSITY);

        if(density)
        {
            // Generate density image.
            uint32_t x, y;
            result = new spectral::Image(width, height, CLS_LUMINANCE);

            for(y=0;y<height;y++)
            {
                for(x=0;x<width;x++)
                {
                    float d, tmp(0);
                    d = luminance_to_density(*(m_lumImage->get_pixel(x, y)));
                    {
                        float lum;
                        lum = *(m_lumImage->get_pixel(x, y));
                        if(lum < min) min = lum;
                        if(lum > max) max = lum;
                    }
                    density->set_pixel(x, y, &d);
                    result->set_pixel(x, y, &tmp);
                }
            }

            // Binning, segmentation etc.
            if(result)
            {
                LayerFarmer farmer(result, this);

                // Create layer jobs.
                for(int i=m_first_layer_id;i<=m_last_layer_id;i++)
                {
                    farmer.Push(new LayerJob(density, GetBinSize(i),
                                m_pattern, m_wrap_horizontal, m_wrap_vertical));
                }

                // Do the work.
                farmer.Process();

                {
                    float *buf = result->get_buffer();
                    float num_layers;
                    num_layers = (float)((m_last_layer_id-m_first_layer_id)+1);

                    for(uint32_t i=0;i<size;i++)
                    {
                        // calculate mean density.
                        buf[i]/= num_layers;

                        // Convert back to luminance.
                        buf[i] = exp(buf[i]);
                    }
                }
            }
        }
    }
    else
    {
    }
    return result;
}

float
Yee::GetBinSize(int id)
{
    float result;

    if(id == 0)
    {
        result = m_min;
    }
    else if(id == 1)
    {
        result = m_max;
    }
    else
    {
        result = halton(3, (id - 1));
        result*= (m_max - m_min);
        result+= m_min;
    }
    return result;
}

void
Yee::ClearImage(uint32_t slot)
{
    if(slot == IMG_LUMINANCE)
    {
        m_lumImage = NULL;
    }
}

}

