#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "utils.h"
#include "thread.h"
#include "image.h"
#include "parameters.h"

#include "contrast.h"

const uint32_t IMG_PICTURE = 0;
const uint32_t IMG_LAL = 1;
const uint32_t JOB_SIZE = 160000;// 400 * 400
const uint32_t JOB_HEIGHT = 64;
const float MIN_INTENSITY = 1.0e-10;

#define MAX2(_a,_b) ((_a>_b)?_a:_b)

const uint8_t GRAD_PX(0x01);
const uint8_t GRAD_PY(0x02);
const uint8_t GRAD_MX(0x04);
const uint8_t GRAD_MY(0x08);
const uint8_t GRAD_PXPY(0x10);
const uint8_t GRAD_PXMY(0x20);
const uint8_t GRAD_MXPY(0x40);
const uint8_t GRAD_MXMY(0x80);

namespace smartcontrast {

class ApplyContrastJob : public thread::Job
{
public:
    ApplyContrastJob(const spectral::Image *img,
                const spectral::Image *lal,
                const spectral::Image *contrast_map,
                spectral::Image *result, 
                size_t start,
                size_t end)
        : m_img(img)
        , m_result(result)
        , m_lal(lal)
        , m_start(start)
        , m_end(end)
        , m_contrast_map(contrast_map)
    {
    }

    virtual void Process(void)
    {
        if(m_img && m_lal)
        {
            const float *in_buf = m_img->get_buffer();
            float *out_buf = m_result->get_buffer();
            float *lal = m_lal->get_buffer();
            float *contrast = m_contrast_map->get_buffer();
    
            size_t index = (m_start * 3) + 2;
    
            {
                for(size_t i=m_start;i<m_end;i++)
                {
                    float L, La, difference;
                    L = log(MAX2(in_buf[index], MIN_INTENSITY));
                    La = log(MAX2(lal[i], MIN_INTENSITY));
                    difference = L - La;
                    L+= (difference * contrast[i]);
                    out_buf[index] = exp(L);
        
                    index+=3;
                }
            }
        }
    }
private:

    const spectral::Image *m_img;
    spectral::Image *m_result;
    const spectral::Image *m_lal;
    size_t m_start, m_end;
    const spectral::Image *m_contrast_map;
};

class PopulateGradientMapJob : public thread::Job
{
public:
    PopulateGradientMapJob(const spectral::Image *xyY, 
                           bool wrap_horizontal, 
                           bool wrap_vertical, 
                           spectral::image<uint8_t> *map, 
                           uint32_t y1, 
                           uint32_t y2)
        : m_xyY(xyY)
        , m_wrap_horizontal(wrap_horizontal)
        , m_wrap_vertical(wrap_vertical)
        , m_map(map)
        , m_y1(y1)
        , m_y2(y2)
    {
    }

    virtual void Process()
    {
        uint32_t x, y;
    
        for(y=m_y1;y<m_y2;y++)
        {
            for(x=0;x<m_xyY->get_width();x++)
            {
                uint8_t gradient;
                float cur_value;
                cur_value = m_xyY->get_pixel(x, y)[2];
                gradient = 0;
    
                if(m_xyY->get_constrained_pixel(x+1, y, m_wrap_horizontal, m_wrap_vertical)[2] > cur_value)
                    gradient |= GRAD_PX;
    
                if(m_xyY->get_constrained_pixel(x-1, y, m_wrap_horizontal, m_wrap_vertical)[2] > cur_value)
                    gradient |= GRAD_MX;
    
                if(m_xyY->get_constrained_pixel(x, y+1, m_wrap_horizontal, m_wrap_vertical)[2] > cur_value)
                    gradient |= GRAD_PY;
    
                if(m_xyY->get_constrained_pixel(x, y-1, m_wrap_horizontal, m_wrap_vertical)[2] > cur_value)
                    gradient |= GRAD_MY;
    
                if(m_xyY->get_constrained_pixel(x+1, y+1, m_wrap_horizontal, m_wrap_vertical)[2] > cur_value)
                    gradient |= GRAD_PXPY;
    
                if(m_xyY->get_constrained_pixel(x+1, y-1, m_wrap_horizontal, m_wrap_vertical)[2] > cur_value)
                    gradient |= GRAD_PXMY;
    
                if(m_xyY->get_constrained_pixel(x-1, y+1, m_wrap_horizontal, m_wrap_vertical)[2] > cur_value)
                    gradient |= GRAD_MXPY;
    
                if(m_xyY->get_constrained_pixel(x-1, y-1, m_wrap_horizontal, m_wrap_vertical)[2] > cur_value)
                    gradient |= GRAD_MXMY;
    
                m_map->set_pixel(x, y, &gradient);
            }
        } 
    }
private:
    const spectral::Image *m_xyY; 
    bool m_wrap_horizontal, m_wrap_vertical; 
    spectral::image<uint8_t> *m_map; 
    uint32_t m_y1, m_y2;
};

class UpdateContrastMapJob : public thread::Job
{
public:
    UpdateContrastMapJob(const spectral::image<uint8_t> *reference_gradient_map,
                         const spectral::image<uint8_t> *result_gradient_map,
                         float max_contrast,
                         spectral::Image *contrast_map,
                         size_t start,
                         size_t end)
        : m_reference_gradient_map(reference_gradient_map)
        , m_result_gradient_map(result_gradient_map)
        , m_max_contrast(max_contrast)
        , m_contrast_map(contrast_map)
        , m_start(start)
        , m_end(end)
        , m_clean(true)
    {
    }

    virtual void Process()
    {
        uint8_t *ref = m_reference_gradient_map->get_buffer();
        uint8_t *res = m_result_gradient_map->get_buffer();
        float *cnt = m_contrast_map->get_buffer();
    
        for(size_t i=m_start;i<m_end;i++)
        {
            if(ref[i] != res[i])
            {
                m_clean = false;
                cnt[i]*= 0.75; // reduce contrast by arbitrary amount.
            }
            else if(cnt[i] < m_max_contrast)
            {
                m_clean = false;
                cnt[i]*= 1.125; // increase contrast by smaller arbitrary amount.
            }
        }
    }

private:
    const spectral::image<uint8_t> *m_reference_gradient_map;
    const spectral::image<uint8_t> *m_result_gradient_map;
    float m_max_contrast;
    spectral::Image *m_contrast_map;
    size_t m_start, m_end;
    bool m_clean;
};

////////////////////////////////////////////////////////////////////////////////
Contrast::Contrast()
    : spectral::Processor()
    , m_picture(NULL)
    , m_lal(NULL)
    , m_contrast(1.0)
    , m_wrap_horizontal(false)
    , m_wrap_vertical(false)
    , m_max_iterations(5)
    , m_smooth_contrast(true)
    , m_output_contrast(false)
{
}

Contrast::~Contrast()
{
}

// Get slots & output type.  Returns output colour space.
// pairs are <slot id, colour space>
uint32_t
Contrast::GetInterface(
    std::list<std::pair<uint32_t, uint32_t> > &mandatory,
    std::list<std::pair<uint32_t, uint32_t> > &optional) const
{   
    mandatory.clear();
    optional.clear();

    mandatory.push_back(     
        std::pair<uint32_t, uint32_t>(IMG_PICTURE, CLS_CIE_XYY));
            
    mandatory.push_back(     
        std::pair<uint32_t, uint32_t>(IMG_LAL, CLS_LUMINANCE));
    return CLS_CIE_XYY;
}  

bool
Contrast::SetImage(uint32_t type, const spectral::Image *img)
{
    bool result = false;

    if(type == IMG_PICTURE)
    {
        m_picture = NULL;
        if(img)
        {
            if(img->get_colour_space() == CLS_CIE_XYY)
            {
                m_picture = img;
                result = true;
            }
        }
        else
        {
            result = true;
        }
    }

    if(type == IMG_LAL)
    {
        m_lal = NULL;
        if(img)
        {
            if(img->get_colour_space() == CLS_LUMINANCE)
            {
                m_lal = img;
                result = true;
            }
        }
        else
        {
            result = true;
        }
    }

    return result;
}

void
Contrast::ClearImage(uint32_t type)
{
    if(type == IMG_PICTURE)
    {
        m_picture = NULL;
    }
}

void 
Contrast::GetParameters(spectral::Parameters *params) const 
{
    if(params)
    {
        params->AddParameter(std::string("contrast"),
                             0.0, 10.0, m_contrast);
        params->AddParameter(std::string("passes"),
                             1, 10, m_max_iterations);
        params->AddParameter(std::string("reduce noise"), 
                             m_smooth_contrast);
        params->AddParameter(std::string("output contrast"), 
                             m_output_contrast);
    }
}

// Apply parameters to a processor.
void
Contrast::SetParameters(const spectral::Parameters *params)
{
    if(params)
    {
        params->GetParameter(std::string("contrast"), m_contrast);
        params->GetParameter(std::string("passes"), m_max_iterations);
        params->GetParameter(std::string("reduce noise"), m_smooth_contrast);
        params->GetParameter(std::string("output contrast"), m_output_contrast);
    }
}

spectral::Image *
Contrast::Process(void)
{
    spectral::Image *result = NULL;

    Progress(0);

    if(m_picture && m_lal)
    {
        const size_t size = size_t(m_picture->get_width()) *
                            size_t(m_picture->get_height());
        uint32_t iterations;
        spectral::image<uint8_t> *reference_gradient_map, *result_gradient_map;
        spectral::Image *contrast_map;
        spectral::Image *temp_contrast_map;
        const spectral::Image *lal = NULL;

        result = new spectral::Image(*m_picture);

        reference_gradient_map = new spectral::image<uint8_t>(m_picture->get_width(), m_picture->get_height(), 1);

        result_gradient_map = new spectral::image<uint8_t>(m_picture->get_width(), m_picture->get_height(), 1);

        contrast_map = new spectral::Image(m_picture->get_width(), m_picture->get_height(), CLS_NUMBER);

        temp_contrast_map = new spectral::Image(m_picture->get_width(), m_picture->get_height(), CLS_NUMBER);

        // Set up reference gradient map.
        PopulateGradientMap(m_picture, m_wrap_horizontal, m_wrap_vertical, reference_gradient_map);

        {
            // Initialise contrast map.
            float *p = contrast_map->get_buffer();

            for(size_t i=0;i<size;i++)
            {
                p[i] = m_contrast;
            }
        }

        // Set up iterations.

        iterations = m_max_iterations + 1;

        bool clean(true);

        do
        {
            iterations--;

            clean = true;
        
            if(m_lal &&
               (m_lal->get_width() == result->get_width()) &&
               (m_lal->get_height() == result->get_height()))
            {
                // lal image must match source image.
                lal = m_lal;
            } 

            ApplyContrast(m_picture,
                          lal,
                          contrast_map,
                          result);

            if(iterations)
            {
                // Create a new gradient map and compare with the reference map.
                PopulateGradientMap(result,
                                    m_wrap_horizontal, 
                                    m_wrap_vertical,
                                    result_gradient_map);

                // Now modify contrast.
                clean = UpdateContrastMap(reference_gradient_map, 
                                          result_gradient_map, 
                                          m_contrast, 
                                          contrast_map);

                // Do we have noise control switched on?
                if(m_smooth_contrast)
                {
                    SmoothContrastMap(contrast_map, 
                                      temp_contrast_map,
                                      m_wrap_horizontal,
                                      m_wrap_vertical);
                    spectral::Image *tmp = contrast_map;
                    contrast_map = temp_contrast_map;
                    temp_contrast_map = tmp;
                }

                {
                    float progress = m_max_iterations + 1 - iterations;
                    progress/= (m_max_iterations + 1);
                    progress*= 100;
                    Progress(progress);
                }

            }
        } while(iterations && (m_smooth_contrast || !clean));

        if(m_output_contrast)
        {
            // Overwrite the result with the final state of the contrast map.
            size_t size, idx;
            size = size_t(result->get_width()) *
                   size_t(result->get_height());
            idx = 0;

            for(size_t i=0;i<size;i++)
            {
                result->get_buffer()[idx++] = 0.31271;
                result->get_buffer()[idx++] = 0.32902;
                result->get_buffer()[idx++] = contrast_map->get_buffer()[i];
            }

        }

        delete reference_gradient_map;
        delete result_gradient_map;
        delete contrast_map;
        delete temp_contrast_map;
        Progress(100.0);
    }
    return result;
}

void
Contrast::ApplyContrast(const spectral::Image *img,
                        const spectral::Image *lal,
                        const spectral::Image *contrast_map,
                        spectral::Image *result)
{
    size_t start, end, size;
    size = size_t(img->get_width()) *
           size_t(img->get_height());
    start = end = size;

    thread::Farmer farmer;
    while(end)
    {
        if(end >= JOB_SIZE)
        {
            start-= JOB_SIZE;
        }
        else
        {
            start = 0;
        }

        farmer.Push(new ApplyContrastJob(img,
                                        lal,
                                        contrast_map,
                                        result,
                                        start,
                                        end));
    
        end = start;
    }
    farmer.Process();
}


void 
Contrast::PopulateGradientMap(const spectral::Image *xyY, bool wrap_horizontal, bool wrap_vertical, spectral::image<uint8_t> *map)
{
    uint32_t start, end;

    start = end = xyY->get_height();


    thread::Farmer farmer;
    while(end)
    { 
        if(end >= JOB_HEIGHT)
        {
            start-= JOB_HEIGHT;
        }
        else
        {
            start = 0;
        }
        farmer.Push(new PopulateGradientMapJob(xyY,
                                               wrap_horizontal,
                                               wrap_vertical,
                                               map,
                                               start,
                                               end));
        end = start;
    }
    farmer.Process();
}

bool 
Contrast::UpdateContrastMap(const spectral::image<uint8_t> *reference_gradient_map,
                            const spectral::image<uint8_t> *result_gradient_map,
                            float max_contrast,
                            spectral::Image *contrast_map)
{
    size_t start, end, size;
    size = size_t(reference_gradient_map->get_width()) *
           size_t(reference_gradient_map->get_height());
    start = end = size;

    thread::Farmer farmer;
    while(end)
    {
        if(end >= JOB_SIZE)
        {
            start-= JOB_SIZE;
        }
        else
        {
            start = 0;
        }

        farmer.Push(new UpdateContrastMapJob(reference_gradient_map,
                                             result_gradient_map,
                                             max_contrast,
                                             contrast_map,
                                             start,
                                             end));
    
        end = start;
    }
    farmer.Process();

    return false;
}

void
Contrast::SmoothContrastMap(spectral::Image *src, 
                            spectral::Image *dst,
                            bool wrap_horizontal, bool wrap_vertical)
{
    int32_t x, y;
    for(y=0;y<int32_t(src->get_height());y++)
    {
        for(x=0;x<int32_t(src->get_width());x++)
        {
#if 0
            // blend value with lowest neighbouring value.
            float cur, min, val;
            cur = *src->get_pixel(x, y);
            min = cur;
            {
                int32_t xx, yy;
                for(yy = (y-1);yy<=(y + 1);yy++)
                {
                    for(xx = (x-1);xx<=(x + 1);xx++)
                    {
                        if((*src->get_constrained_pixel(xx, yy)) < min)
                        {
                            min = (*src->get_constrained_pixel(xx, yy));
                        }
                    }
                }
            }
            //val = ((cur * .75) + (min * 0.25));// / 2.0;
            val = (cur + min ) / 2.0;
            dst->set_pixel(x, y, &val);
#else
            // blend value with average of lower neighbouring values.
            float cur, val, total, weight;
            cur = *src->get_pixel(x, y);
            total = weight = 0.0;
            {
                int32_t xx, yy;
                for(yy = (y-1);yy<=(y + 1);yy++)
                {
                    for(xx = (x-1);xx<=(x + 1);xx++)
                    {
                        if((*src->get_constrained_pixel(xx, yy)) <= cur)
                        {
                            weight++;
                            total+= (*src->get_constrained_pixel(xx, yy));
                        }
                    }
                }
            }
            val = (total/weight);
            dst->set_pixel(x, y, &val);
#endif
        }
    }
}

}
