#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "utils.h"
#include "image.h"
#include "parameters.h"

#include "whitepoint.h"

const uint32_t IMG_PICTURE = 0;

namespace whitepoint {

Whitepoint::Whitepoint()
    : spectral::Processor()
    , m_picture(NULL)
    , m_xs(0.312713)
    , m_ys(0.329016)
    , m_xd(0.312713)
    , m_yd(0.329016)
{
}

Whitepoint::~Whitepoint()
{
}

// Get slots & output type.  Returns output colour space.
// pairs are <slot id, colour space>
uint32_t
Whitepoint::GetInterface(
    std::list<std::pair<uint32_t, uint32_t> > &mandatory,
    std::list<std::pair<uint32_t, uint32_t> > &optional) const
{   
    mandatory.clear();
    optional.clear();
        
    mandatory.push_back(     
        std::pair<uint32_t, uint32_t>(IMG_PICTURE, CLS_CIE_XYY));
            
    return CLS_CIE_XYY;
}  

bool
Whitepoint::SetImage(uint32_t type, const spectral::Image *img)
{
    bool result = false;

    if(type == IMG_PICTURE)
    {
        m_picture = NULL;
        if(img)
        {
            if(img->get_colour_space() == CLS_CIE_XYY)
            {
                m_picture = img;
                result = true;
            }
        }
        else
        {
            result = true;
        }
    }

    return result;
}

void
Whitepoint::ClearImage(uint32_t type)
{
    if(type == IMG_PICTURE)
    {
        m_picture = NULL;
    }
}

void 
Whitepoint::GetParameters(spectral::Parameters *params) const 
{
    if(params)
    {
        params->AddParameter(std::string("from white"),
                             spectral::Chromaticity(m_xs, m_ys));
        params->AddParameter(std::string("to white"),
                             spectral::Chromaticity(m_xd, m_yd));
    }
}

// Apply parameters to a processor.
void
Whitepoint::SetParameters(const spectral::Parameters *params)
{
    if(params)
    {
        spectral::Chromaticity tmp;

        if(params->GetParameter(std::string("from white"), tmp))
        {
            tmp.GetXY(m_xs, m_ys);
        }
        if(params->GetParameter(std::string("to white"), tmp))
        {
            tmp.GetXY(m_xd, m_yd);
        }

    }
}

spectral::Image *
Whitepoint::Process(void)
{
    spectral::Image *result = NULL;
    //send_progress(0.0);

    if(m_picture)
    {
        result = new spectral::Image(*m_picture);
    }

    if(result)
    {
        float ws[3], wd[3];
        Matrix33 matrix;

        size_t size = size_t(result->get_width()) * size_t(result->get_height());
        float *buf = result->get_buffer();
        size_t i;

        ws[0] = m_xs;
        ws[1] = m_ys;
        ws[2] = 1;

        wd[0] = m_xd;
        wd[1] = m_yd;
        wd[2] = 1;

        xyy_to_xyz(ws); 
        xyy_to_xyz(wd);
      
        chromatic_adaptation(ws, wd, &matrix);

        //send_progress(0);

        for(i=0;i<size;i++)
        {
            float *p;
            p = buf + (i * 3);
            xyy_to_xyz(p);
            transform_colour(matrix, p);
            xyz_to_xyy(p);
        }

        //send_progress(100);
    }

    return result;
}

}
